from libc.stdlib cimport malloc, free

import numpy as np
cimport numpy as np

from compmech.fem.utils cimport invJ2, detJ2
from compmech.fem.model import Shell2DFSDT

cdef extern from "math.h":
    double cos(double t) nogil
    double sin(double t) nogil


class FSDTDkquad8(Shell2DFSDT):
    els = []
    nnodes = 8
    ptypenum = 1 # this number must be unique
    nk0 = 1600
    nk0L = None
    nkLL = None
    nkG = None

    def __init__(self, stack, plyts, laminaprops):
        super(Property, self).__init__(stack, plyts, laminaprops)

    def calc_el_attr(self, coords):
        cdef int i
        cdef int nel = len(self.els)
        cdef int nnodes = self.nnodes
        cdef np.ndarray[np.float64_t, ndim=3] T
        cdef np.ndarray[np.float64_t, ndim=2] p0, lcoords
        cdef double *Tflat

        # we need one T for each element
        T = np.empty((nel, 3, 4), dtype=np.float64)
        Tflat = &T[0,0,0]
        p0 = np.empty((nel, 3), dtype=np.float64)

        for i in range(nel):
            # define vectorized Ts...
            a
            b
            g
            p0[i, 0] = x0
            p0[i, 1] = y0
            p0[i, 2] = z0

            Tflat[nnodes*i*9+0] = cos(b)*cos(g)
            Tflat[nnodes*i*9+1] = (sin(a)*sin(b)*cos(g) + cos(a)*sin(g))
            Tflat[nnodes*i*9+2] = (sin(a)*sin(g) - cos(a)*sin(b)*cos(g))
            Tflat[nnodes*i*9+3] = -cos(b)*sin(g)
            Tflat[nnodes*i*9+4] = (cos(a)*cos(g) - sin(a)*sin(b)*sin(g))
            Tflat[nnodes*i*9+5] = (sin(a)*cos(g) + cos(a)*sin(b)*sin(g))
            Tflat[nnodes*i*9+6] = sin(b)
            Tflat[nnodes*i*9+7] = -sin(a)*cos(b)
            Tflat[nnodes*i*9+8] = cos(a)*cos(b)

        lcoords = np.empty_like(coords)
        np.einsum('ijk, ik -> ij', T, coords, out=lcoords)
        lcoords += p0

        np.ein

        p0 =
        p1 =
        p2 =

        J
        rs
        arad


        pass


cdef void calc_k(int nel,
                 double *k0, double *k0L, double *kLL, double *kG,
                 double *Cs, int *Cind,
                 double *J, double *rs, double *arads) nogil:
    cdef int i, j, pt, pos
    cdef int npts=9
    cdef int ndim=2
    cdef int nk0=1600
    cdef double *invJi
    cdef double h1, h2, h3, h4, h5, h6, h7, h8
    cdef double h1x, h2x, h3x, h4x, h5x, h6x, h7x, h8x
    cdef double h1y, h2y, h3y, h4y, h5y, h6y, h7y, h8y
    cdef double xis[8]
    cdef double etas[8]
    cdef double wis[8]
    cdef double xi, eta, wi
    cdef double j11, j12, j21, j22, detJ
    cdef double A11, A12, A16, A22, A26, A66, A44, A45, A55
    cdef double B11, B12, B16, B22, B26, B66
    cdef double D11, D12, D16, D22, D26, D66
    cdef double r1, r2, s, cosa, sina
    cdef double *C

    xis[0] = +0.774596669241483
    xis[1] = -0.774596669241483
    xis[2] = -0.774596669241483
    xis[3] = +0.774596669241483
    xis[4] = 0
    xis[5] = -0.774596669241483
    xis[6] = 0
    xis[7] = +0.774596669241483
    xis[8] = 0

    etas[0] = +0.774596669241483
    etas[1] = +0.774596669241483
    etas[2] = -0.774596669241483
    etas[3] = -0.774596669241483
    etas[4] = +0.774596669241483
    etas[5] = 0
    etas[6] = -0.774596669241483
    etas[7] = 0
    etas[8] = 0

    wis[0] = 0.308641975308643
    wis[1] = 0.308641975308643
    wis[2] = 0.308641975308643
    wis[3] = 0.308641975308643
    wis[4] = 0.493827160549383
    wis[5] = 0.493827160549383
    wis[6] = 0.493827160549383
    wis[7] = 0.493827160549383
    wis[8] = 0.790123456967901

    cdef double *invJ = <double *>malloc(ndim*ndim*nel*npts*sizeof(double))
    cdef double *detJs = <double *>malloc(nel*npts*sizeof(double))

    invJ2(nel, npts, J, invJ)
    detJ2(nel, npts, J, detJs)

    for i in range(nel):
        pos = nk0*i
        r1 = rs[2*i+0]
        r2 = rs[2*i+1]
        cosa = cos(arads[i])
        sina = sin(arads[i])
        C = &Cs[Cind[i]]
        A11 = C[0]
        A12 = C[1]
        A16 = C[2]
        A22 = C[3]
        A26 = C[4]
        A66 = C[5]
        B11 = C[6]
        B12 = C[7]
        B16 = C[8]
        B22 = C[9]
        B26 = C[10]
        B66 = C[11]
        D11 = C[12]
        D12 = C[13]
        D16 = C[14]
        D22 = C[15]
        D26 = C[16]
        D66 = C[17]
        A44 = C[18]
        A45 = C[19]
        A55 = C[20]

        for j in range(nk0):
            k0[pos+j] = 0.

        for pt in range(npts):
            xi = xis[pt]
            eta = etas[pt]
            wi = wis[pt]
            s = 1./(1/2.*(r1*(1-xi)+r2*(1+xi)))
            invJi = &invJ[ndim*ndim*npts*i + ndim*pt]
            j11 = invJi[0]
            j12 = invJi[1]
            j21 = invJi[2]
            j22 = invJi[3]
            detJ = detJs[npts*i + pt]

            h8 = 1/2.*(1+xi)*(1-eta*eta)
            h7 = 1/2.*(1-xi*xi)*(1-eta)
            h6 = 1/2.*(1-xi)*(1-eta*eta)
            h5 = 1/2.*(1-xi*xi)*(1+eta)
            h1 = 1/4.*(1+xi)*(1+eta) -1/2.*h5 - 1/2.*h8
            h2 = 1/4.*(1-xi)*(1+eta) -1/2.*h5 - 1/2.*h6
            h3 = 1/4.*(1-xi)*(1-eta) -1/2.*h6 - 1/2.*h7
            h4 = 1/4.*(1+xi)*(1-eta) -1/2.*h7 - 1/2.*h8

            h1x = (eta*eta*j11 + xi*(2*j11 + j12 + j12*xi) + eta*(j11 + 2*j12 + 2*(j11 + j12)*xi))/4.
            h2x = (-(eta*(j11 + eta*j11 - 2*j12)) + (2*(1 + eta)*j11 - (1 + 2*eta)*j12)*xi + j12*xi*xi)/4.
            h3x = (-(eta*eta*j11) + xi*(2*j11 + j12 - j12*xi) + eta*(j11 + 2*j12 - 2*(j11 + j12)*xi))/4.
            h4x = (eta*eta*j11 - xi*(-2*j11 + j12 + j12*xi) - eta*(j11 + 2*j11*xi - 2*j12*(1 + xi)))/4.
            h5x = -((1 + eta)*j11*xi) - (j12*(-1 + xi*xi))/2.
            h6x = ((-1 + eta*eta)*j11)/2. + eta*j12*(-1 + xi)
            h7x = (-1 + eta)*j11*xi + (j12*(-1 + xi*xi))/2.
            h8x = -((-1 + eta*eta)*j11)/2. - eta*j12*(1 + xi)

            h1y = (eta*eta*j21 + xi*(2*j21 + j22 + j22*xi) + eta*(j21 + 2*j22 + 2*(j21 + j22)*xi))/4.
            h2y = (-(eta*(j21 + eta*j21 - 2*j22)) + (2*(1 + eta)*j21 - (1 + 2*eta)*j22)*xi + j22*xi*xi)/4.
            h3y = (-(eta*eta*j21) + xi*(2*j21 + j22 - j22*xi) + eta*(j21 + 2*j22 - 2*(j21 + j22)*xi))/4.
            h4y = (eta*eta*j21 - xi*(-2*j21 + j22 + j22*xi) - eta*(j21 + 2*j21*xi - 2*j22*(1 + xi)))/4.
            h5y = -((1 + eta)*j21*xi) - (j22*(-1 + xi*xi))/2.
            h6y = ((-1 + eta*eta)*j21)/2. + eta*j22*(-1 + xi)
            h7y = (-1 + eta)*j21*xi + (j22*(-1 + xi*xi))/2.
            h8y = -((-1 + eta*eta)*j21)/2. - eta*j22*(1 + xi)

            k0[pos+0] += detJ*wi*(A11*(h1x*h1x) + 2*A12*h1*h1x*s*sina + 2*A16*h1x*h1y + A22*(h1*h1)*(s*s)*(sina*sina) + 2*A26*h1*h1y*s*sina + A66*(h1y*h1y))
            k0[pos+1] += detJ*wi*(h2*s*sina*(A12*h1x + A22*h1*s*sina + A26*h1y) + h2x*(A11*h1x + A12*h1*s*sina + A16*h1y) + h2y*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+2] += detJ*wi*(h3*s*sina*(A12*h1x + A22*h1*s*sina + A26*h1y) + h3x*(A11*h1x + A12*h1*s*sina + A16*h1y) + h3y*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+3] += detJ*wi*(h4*s*sina*(A12*h1x + A22*h1*s*sina + A26*h1y) + h4x*(A11*h1x + A12*h1*s*sina + A16*h1y) + h4y*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+4] += detJ*wi*(h5*s*sina*(A12*h1x + A22*h1*s*sina + A26*h1y) + h5x*(A11*h1x + A12*h1*s*sina + A16*h1y) + h5y*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+5] += detJ*wi*(h6*s*sina*(A12*h1x + A22*h1*s*sina + A26*h1y) + h6x*(A11*h1x + A12*h1*s*sina + A16*h1y) + h6y*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+6] += detJ*wi*(h7*s*sina*(A12*h1x + A22*h1*s*sina + A26*h1y) + h7x*(A11*h1x + A12*h1*s*sina + A16*h1y) + h7y*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+7] += detJ*wi*(h8*s*sina*(A12*h1x + A22*h1*s*sina + A26*h1y) + h8x*(A11*h1x + A12*h1*s*sina + A16*h1y) + h8y*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+8] += detJ*wi*(h1y*(A12*h1x + A22*h1*s*sina + A26*h1y) + (-h1*s*sina + h1x)*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+9] += detJ*wi*(h2y*(A12*h1x + A22*h1*s*sina + A26*h1y) + (-h2*s*sina + h2x)*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+10] += detJ*wi*(h3y*(A12*h1x + A22*h1*s*sina + A26*h1y) + (-h3*s*sina + h3x)*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+11] += detJ*wi*(h4y*(A12*h1x + A22*h1*s*sina + A26*h1y) + (-h4*s*sina + h4x)*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+12] += detJ*wi*(h5y*(A12*h1x + A22*h1*s*sina + A26*h1y) + (-h5*s*sina + h5x)*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+13] += detJ*wi*(h6y*(A12*h1x + A22*h1*s*sina + A26*h1y) + (-h6*s*sina + h6x)*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+14] += detJ*wi*(h7y*(A12*h1x + A22*h1*s*sina + A26*h1y) + (-h7*s*sina + h7x)*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+15] += detJ*wi*(h8y*(A12*h1x + A22*h1*s*sina + A26*h1y) + (-h8*s*sina + h8x)*(A16*h1x + A26*h1*s*sina + A66*h1y))
            k0[pos+16] += cosa*detJ*h1*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+17] += cosa*detJ*h2*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+18] += cosa*detJ*h3*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+19] += cosa*detJ*h4*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+20] += cosa*detJ*h5*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+21] += cosa*detJ*h6*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+22] += cosa*detJ*h7*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+23] += cosa*detJ*h8*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+24] += detJ*wi*(B11*(h1x*h1x) + 2*B12*h1*h1x*s*sina + 2*B16*h1x*h1y + B22*(h1*h1)*(s*s)*(sina*sina) + 2*B26*h1*h1y*s*sina + B66*(h1y*h1y))
            k0[pos+25] += detJ*wi*(h2*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h2x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h2y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+26] += detJ*wi*(h3*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h3x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h3y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+27] += detJ*wi*(h4*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h4x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h4y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+28] += detJ*wi*(h5*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h5x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h5y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+29] += detJ*wi*(h6*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h6x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h6y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+30] += detJ*wi*(h7*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h7x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h7y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+31] += detJ*wi*(h8*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h8x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h8y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+32] += detJ*wi*(h1y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h1*s*sina + h1x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+33] += detJ*wi*(h2y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h2*s*sina + h2x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+34] += detJ*wi*(h3y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h3*s*sina + h3x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+35] += detJ*wi*(h4y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h4*s*sina + h4x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+36] += detJ*wi*(h5y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h5*s*sina + h5x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+37] += detJ*wi*(h6y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h6*s*sina + h6x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+38] += detJ*wi*(h7y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h7*s*sina + h7x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+39] += detJ*wi*(h8y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h8*s*sina + h8x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+40] += detJ*wi*(h1*s*sina*(A12*h2x + A22*h2*s*sina + A26*h2y) + h1x*(A11*h2x + A12*h2*s*sina + A16*h2y) + h1y*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+41] += detJ*wi*(A11*(h2x*h2x) + 2*A12*h2*h2x*s*sina + 2*A16*h2x*h2y + A22*(h2*h2)*(s*s)*(sina*sina) + 2*A26*h2*h2y*s*sina + A66*(h2y*h2y))
            k0[pos+42] += detJ*wi*(h3*s*sina*(A12*h2x + A22*h2*s*sina + A26*h2y) + h3x*(A11*h2x + A12*h2*s*sina + A16*h2y) + h3y*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+43] += detJ*wi*(h4*s*sina*(A12*h2x + A22*h2*s*sina + A26*h2y) + h4x*(A11*h2x + A12*h2*s*sina + A16*h2y) + h4y*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+44] += detJ*wi*(h5*s*sina*(A12*h2x + A22*h2*s*sina + A26*h2y) + h5x*(A11*h2x + A12*h2*s*sina + A16*h2y) + h5y*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+45] += detJ*wi*(h6*s*sina*(A12*h2x + A22*h2*s*sina + A26*h2y) + h6x*(A11*h2x + A12*h2*s*sina + A16*h2y) + h6y*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+46] += detJ*wi*(h7*s*sina*(A12*h2x + A22*h2*s*sina + A26*h2y) + h7x*(A11*h2x + A12*h2*s*sina + A16*h2y) + h7y*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+47] += detJ*wi*(h8*s*sina*(A12*h2x + A22*h2*s*sina + A26*h2y) + h8x*(A11*h2x + A12*h2*s*sina + A16*h2y) + h8y*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+48] += detJ*wi*(h1y*(A12*h2x + A22*h2*s*sina + A26*h2y) + (-h1*s*sina + h1x)*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+49] += detJ*wi*(h2y*(A12*h2x + A22*h2*s*sina + A26*h2y) + (-h2*s*sina + h2x)*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+50] += detJ*wi*(h3y*(A12*h2x + A22*h2*s*sina + A26*h2y) + (-h3*s*sina + h3x)*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+51] += detJ*wi*(h4y*(A12*h2x + A22*h2*s*sina + A26*h2y) + (-h4*s*sina + h4x)*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+52] += detJ*wi*(h5y*(A12*h2x + A22*h2*s*sina + A26*h2y) + (-h5*s*sina + h5x)*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+53] += detJ*wi*(h6y*(A12*h2x + A22*h2*s*sina + A26*h2y) + (-h6*s*sina + h6x)*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+54] += detJ*wi*(h7y*(A12*h2x + A22*h2*s*sina + A26*h2y) + (-h7*s*sina + h7x)*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+55] += detJ*wi*(h8y*(A12*h2x + A22*h2*s*sina + A26*h2y) + (-h8*s*sina + h8x)*(A16*h2x + A26*h2*s*sina + A66*h2y))
            k0[pos+56] += cosa*detJ*h1*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+57] += cosa*detJ*h2*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+58] += cosa*detJ*h3*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+59] += cosa*detJ*h4*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+60] += cosa*detJ*h5*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+61] += cosa*detJ*h6*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+62] += cosa*detJ*h7*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+63] += cosa*detJ*h8*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+64] += detJ*wi*(h1*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h1x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h1y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+65] += detJ*wi*(B11*(h2x*h2x) + 2*B12*h2*h2x*s*sina + 2*B16*h2x*h2y + B22*(h2*h2)*(s*s)*(sina*sina) + 2*B26*h2*h2y*s*sina + B66*(h2y*h2y))
            k0[pos+66] += detJ*wi*(h3*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h3x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h3y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+67] += detJ*wi*(h4*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h4x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h4y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+68] += detJ*wi*(h5*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h5x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h5y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+69] += detJ*wi*(h6*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h6x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h6y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+70] += detJ*wi*(h7*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h7x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h7y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+71] += detJ*wi*(h8*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h8x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h8y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+72] += detJ*wi*(h1y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h1*s*sina + h1x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+73] += detJ*wi*(h2y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h2*s*sina + h2x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+74] += detJ*wi*(h3y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h3*s*sina + h3x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+75] += detJ*wi*(h4y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h4*s*sina + h4x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+76] += detJ*wi*(h5y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h5*s*sina + h5x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+77] += detJ*wi*(h6y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h6*s*sina + h6x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+78] += detJ*wi*(h7y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h7*s*sina + h7x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+79] += detJ*wi*(h8y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h8*s*sina + h8x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+80] += detJ*wi*(h1*s*sina*(A12*h3x + A22*h3*s*sina + A26*h3y) + h1x*(A11*h3x + A12*h3*s*sina + A16*h3y) + h1y*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+81] += detJ*wi*(h2*s*sina*(A12*h3x + A22*h3*s*sina + A26*h3y) + h2x*(A11*h3x + A12*h3*s*sina + A16*h3y) + h2y*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+82] += detJ*wi*(A11*(h3x*h3x) + 2*A12*h3*h3x*s*sina + 2*A16*h3x*h3y + A22*(h3*h3)*(s*s)*(sina*sina) + 2*A26*h3*h3y*s*sina + A66*(h3y*h3y))
            k0[pos+83] += detJ*wi*(h4*s*sina*(A12*h3x + A22*h3*s*sina + A26*h3y) + h4x*(A11*h3x + A12*h3*s*sina + A16*h3y) + h4y*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+84] += detJ*wi*(h5*s*sina*(A12*h3x + A22*h3*s*sina + A26*h3y) + h5x*(A11*h3x + A12*h3*s*sina + A16*h3y) + h5y*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+85] += detJ*wi*(h6*s*sina*(A12*h3x + A22*h3*s*sina + A26*h3y) + h6x*(A11*h3x + A12*h3*s*sina + A16*h3y) + h6y*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+86] += detJ*wi*(h7*s*sina*(A12*h3x + A22*h3*s*sina + A26*h3y) + h7x*(A11*h3x + A12*h3*s*sina + A16*h3y) + h7y*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+87] += detJ*wi*(h8*s*sina*(A12*h3x + A22*h3*s*sina + A26*h3y) + h8x*(A11*h3x + A12*h3*s*sina + A16*h3y) + h8y*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+88] += detJ*wi*(h1y*(A12*h3x + A22*h3*s*sina + A26*h3y) + (-h1*s*sina + h1x)*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+89] += detJ*wi*(h2y*(A12*h3x + A22*h3*s*sina + A26*h3y) + (-h2*s*sina + h2x)*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+90] += detJ*wi*(h3y*(A12*h3x + A22*h3*s*sina + A26*h3y) + (-h3*s*sina + h3x)*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+91] += detJ*wi*(h4y*(A12*h3x + A22*h3*s*sina + A26*h3y) + (-h4*s*sina + h4x)*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+92] += detJ*wi*(h5y*(A12*h3x + A22*h3*s*sina + A26*h3y) + (-h5*s*sina + h5x)*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+93] += detJ*wi*(h6y*(A12*h3x + A22*h3*s*sina + A26*h3y) + (-h6*s*sina + h6x)*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+94] += detJ*wi*(h7y*(A12*h3x + A22*h3*s*sina + A26*h3y) + (-h7*s*sina + h7x)*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+95] += detJ*wi*(h8y*(A12*h3x + A22*h3*s*sina + A26*h3y) + (-h8*s*sina + h8x)*(A16*h3x + A26*h3*s*sina + A66*h3y))
            k0[pos+96] += cosa*detJ*h1*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+97] += cosa*detJ*h2*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+98] += cosa*detJ*h3*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+99] += cosa*detJ*h4*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+100] += cosa*detJ*h5*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+101] += cosa*detJ*h6*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+102] += cosa*detJ*h7*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+103] += cosa*detJ*h8*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+104] += detJ*wi*(h1*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h1x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h1y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+105] += detJ*wi*(h2*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h2x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h2y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+106] += detJ*wi*(B11*(h3x*h3x) + 2*B12*h3*h3x*s*sina + 2*B16*h3x*h3y + B22*(h3*h3)*(s*s)*(sina*sina) + 2*B26*h3*h3y*s*sina + B66*(h3y*h3y))
            k0[pos+107] += detJ*wi*(h4*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h4x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h4y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+108] += detJ*wi*(h5*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h5x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h5y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+109] += detJ*wi*(h6*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h6x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h6y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+110] += detJ*wi*(h7*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h7x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h7y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+111] += detJ*wi*(h8*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h8x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h8y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+112] += detJ*wi*(h1y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h1*s*sina + h1x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+113] += detJ*wi*(h2y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h2*s*sina + h2x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+114] += detJ*wi*(h3y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h3*s*sina + h3x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+115] += detJ*wi*(h4y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h4*s*sina + h4x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+116] += detJ*wi*(h5y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h5*s*sina + h5x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+117] += detJ*wi*(h6y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h6*s*sina + h6x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+118] += detJ*wi*(h7y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h7*s*sina + h7x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+119] += detJ*wi*(h8y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h8*s*sina + h8x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+120] += detJ*wi*(h1*s*sina*(A12*h4x + A22*h4*s*sina + A26*h4y) + h1x*(A11*h4x + A12*h4*s*sina + A16*h4y) + h1y*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+121] += detJ*wi*(h2*s*sina*(A12*h4x + A22*h4*s*sina + A26*h4y) + h2x*(A11*h4x + A12*h4*s*sina + A16*h4y) + h2y*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+122] += detJ*wi*(h3*s*sina*(A12*h4x + A22*h4*s*sina + A26*h4y) + h3x*(A11*h4x + A12*h4*s*sina + A16*h4y) + h3y*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+123] += detJ*wi*(A11*(h4x*h4x) + 2*A12*h4*h4x*s*sina + 2*A16*h4x*h4y + A22*(h4*h4)*(s*s)*(sina*sina) + 2*A26*h4*h4y*s*sina + A66*(h4y*h4y))
            k0[pos+124] += detJ*wi*(h5*s*sina*(A12*h4x + A22*h4*s*sina + A26*h4y) + h5x*(A11*h4x + A12*h4*s*sina + A16*h4y) + h5y*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+125] += detJ*wi*(h6*s*sina*(A12*h4x + A22*h4*s*sina + A26*h4y) + h6x*(A11*h4x + A12*h4*s*sina + A16*h4y) + h6y*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+126] += detJ*wi*(h7*s*sina*(A12*h4x + A22*h4*s*sina + A26*h4y) + h7x*(A11*h4x + A12*h4*s*sina + A16*h4y) + h7y*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+127] += detJ*wi*(h8*s*sina*(A12*h4x + A22*h4*s*sina + A26*h4y) + h8x*(A11*h4x + A12*h4*s*sina + A16*h4y) + h8y*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+128] += detJ*wi*(h1y*(A12*h4x + A22*h4*s*sina + A26*h4y) + (-h1*s*sina + h1x)*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+129] += detJ*wi*(h2y*(A12*h4x + A22*h4*s*sina + A26*h4y) + (-h2*s*sina + h2x)*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+130] += detJ*wi*(h3y*(A12*h4x + A22*h4*s*sina + A26*h4y) + (-h3*s*sina + h3x)*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+131] += detJ*wi*(h4y*(A12*h4x + A22*h4*s*sina + A26*h4y) + (-h4*s*sina + h4x)*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+132] += detJ*wi*(h5y*(A12*h4x + A22*h4*s*sina + A26*h4y) + (-h5*s*sina + h5x)*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+133] += detJ*wi*(h6y*(A12*h4x + A22*h4*s*sina + A26*h4y) + (-h6*s*sina + h6x)*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+134] += detJ*wi*(h7y*(A12*h4x + A22*h4*s*sina + A26*h4y) + (-h7*s*sina + h7x)*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+135] += detJ*wi*(h8y*(A12*h4x + A22*h4*s*sina + A26*h4y) + (-h8*s*sina + h8x)*(A16*h4x + A26*h4*s*sina + A66*h4y))
            k0[pos+136] += cosa*detJ*h1*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+137] += cosa*detJ*h2*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+138] += cosa*detJ*h3*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+139] += cosa*detJ*h4*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+140] += cosa*detJ*h5*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+141] += cosa*detJ*h6*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+142] += cosa*detJ*h7*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+143] += cosa*detJ*h8*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+144] += detJ*wi*(h1*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h1x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h1y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+145] += detJ*wi*(h2*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h2x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h2y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+146] += detJ*wi*(h3*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h3x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h3y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+147] += detJ*wi*(B11*(h4x*h4x) + 2*B12*h4*h4x*s*sina + 2*B16*h4x*h4y + B22*(h4*h4)*(s*s)*(sina*sina) + 2*B26*h4*h4y*s*sina + B66*(h4y*h4y))
            k0[pos+148] += detJ*wi*(h5*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h5x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h5y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+149] += detJ*wi*(h6*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h6x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h6y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+150] += detJ*wi*(h7*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h7x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h7y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+151] += detJ*wi*(h8*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h8x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h8y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+152] += detJ*wi*(h1y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h1*s*sina + h1x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+153] += detJ*wi*(h2y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h2*s*sina + h2x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+154] += detJ*wi*(h3y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h3*s*sina + h3x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+155] += detJ*wi*(h4y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h4*s*sina + h4x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+156] += detJ*wi*(h5y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h5*s*sina + h5x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+157] += detJ*wi*(h6y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h6*s*sina + h6x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+158] += detJ*wi*(h7y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h7*s*sina + h7x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+159] += detJ*wi*(h8y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h8*s*sina + h8x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+160] += detJ*wi*(h1*s*sina*(A12*h5x + A22*h5*s*sina + A26*h5y) + h1x*(A11*h5x + A12*h5*s*sina + A16*h5y) + h1y*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+161] += detJ*wi*(h2*s*sina*(A12*h5x + A22*h5*s*sina + A26*h5y) + h2x*(A11*h5x + A12*h5*s*sina + A16*h5y) + h2y*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+162] += detJ*wi*(h3*s*sina*(A12*h5x + A22*h5*s*sina + A26*h5y) + h3x*(A11*h5x + A12*h5*s*sina + A16*h5y) + h3y*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+163] += detJ*wi*(h4*s*sina*(A12*h5x + A22*h5*s*sina + A26*h5y) + h4x*(A11*h5x + A12*h5*s*sina + A16*h5y) + h4y*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+164] += detJ*wi*(A11*(h5x*h5x) + 2*A12*h5*h5x*s*sina + 2*A16*h5x*h5y + A22*(h5*h5)*(s*s)*(sina*sina) + 2*A26*h5*h5y*s*sina + A66*(h5y*h5y))
            k0[pos+165] += detJ*wi*(h6*s*sina*(A12*h5x + A22*h5*s*sina + A26*h5y) + h6x*(A11*h5x + A12*h5*s*sina + A16*h5y) + h6y*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+166] += detJ*wi*(h7*s*sina*(A12*h5x + A22*h5*s*sina + A26*h5y) + h7x*(A11*h5x + A12*h5*s*sina + A16*h5y) + h7y*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+167] += detJ*wi*(h8*s*sina*(A12*h5x + A22*h5*s*sina + A26*h5y) + h8x*(A11*h5x + A12*h5*s*sina + A16*h5y) + h8y*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+168] += detJ*wi*(h1y*(A12*h5x + A22*h5*s*sina + A26*h5y) + (-h1*s*sina + h1x)*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+169] += detJ*wi*(h2y*(A12*h5x + A22*h5*s*sina + A26*h5y) + (-h2*s*sina + h2x)*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+170] += detJ*wi*(h3y*(A12*h5x + A22*h5*s*sina + A26*h5y) + (-h3*s*sina + h3x)*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+171] += detJ*wi*(h4y*(A12*h5x + A22*h5*s*sina + A26*h5y) + (-h4*s*sina + h4x)*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+172] += detJ*wi*(h5y*(A12*h5x + A22*h5*s*sina + A26*h5y) + (-h5*s*sina + h5x)*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+173] += detJ*wi*(h6y*(A12*h5x + A22*h5*s*sina + A26*h5y) + (-h6*s*sina + h6x)*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+174] += detJ*wi*(h7y*(A12*h5x + A22*h5*s*sina + A26*h5y) + (-h7*s*sina + h7x)*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+175] += detJ*wi*(h8y*(A12*h5x + A22*h5*s*sina + A26*h5y) + (-h8*s*sina + h8x)*(A16*h5x + A26*h5*s*sina + A66*h5y))
            k0[pos+176] += cosa*detJ*h1*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+177] += cosa*detJ*h2*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+178] += cosa*detJ*h3*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+179] += cosa*detJ*h4*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+180] += cosa*detJ*h5*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+181] += cosa*detJ*h6*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+182] += cosa*detJ*h7*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+183] += cosa*detJ*h8*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+184] += detJ*wi*(h1*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h1x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h1y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+185] += detJ*wi*(h2*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h2x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h2y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+186] += detJ*wi*(h3*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h3x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h3y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+187] += detJ*wi*(h4*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h4x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h4y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+188] += detJ*wi*(B11*(h5x*h5x) + 2*B12*h5*h5x*s*sina + 2*B16*h5x*h5y + B22*(h5*h5)*(s*s)*(sina*sina) + 2*B26*h5*h5y*s*sina + B66*(h5y*h5y))
            k0[pos+189] += detJ*wi*(h6*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h6x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h6y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+190] += detJ*wi*(h7*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h7x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h7y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+191] += detJ*wi*(h8*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h8x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h8y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+192] += detJ*wi*(h1y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h1*s*sina + h1x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+193] += detJ*wi*(h2y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h2*s*sina + h2x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+194] += detJ*wi*(h3y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h3*s*sina + h3x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+195] += detJ*wi*(h4y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h4*s*sina + h4x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+196] += detJ*wi*(h5y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h5*s*sina + h5x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+197] += detJ*wi*(h6y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h6*s*sina + h6x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+198] += detJ*wi*(h7y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h7*s*sina + h7x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+199] += detJ*wi*(h8y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h8*s*sina + h8x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+200] += detJ*wi*(h1*s*sina*(A12*h6x + A22*h6*s*sina + A26*h6y) + h1x*(A11*h6x + A12*h6*s*sina + A16*h6y) + h1y*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+201] += detJ*wi*(h2*s*sina*(A12*h6x + A22*h6*s*sina + A26*h6y) + h2x*(A11*h6x + A12*h6*s*sina + A16*h6y) + h2y*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+202] += detJ*wi*(h3*s*sina*(A12*h6x + A22*h6*s*sina + A26*h6y) + h3x*(A11*h6x + A12*h6*s*sina + A16*h6y) + h3y*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+203] += detJ*wi*(h4*s*sina*(A12*h6x + A22*h6*s*sina + A26*h6y) + h4x*(A11*h6x + A12*h6*s*sina + A16*h6y) + h4y*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+204] += detJ*wi*(h5*s*sina*(A12*h6x + A22*h6*s*sina + A26*h6y) + h5x*(A11*h6x + A12*h6*s*sina + A16*h6y) + h5y*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+205] += detJ*wi*(A11*(h6x*h6x) + 2*A12*h6*h6x*s*sina + 2*A16*h6x*h6y + A22*(h6*h6)*(s*s)*(sina*sina) + 2*A26*h6*h6y*s*sina + A66*(h6y*h6y))
            k0[pos+206] += detJ*wi*(h7*s*sina*(A12*h6x + A22*h6*s*sina + A26*h6y) + h7x*(A11*h6x + A12*h6*s*sina + A16*h6y) + h7y*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+207] += detJ*wi*(h8*s*sina*(A12*h6x + A22*h6*s*sina + A26*h6y) + h8x*(A11*h6x + A12*h6*s*sina + A16*h6y) + h8y*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+208] += detJ*wi*(h1y*(A12*h6x + A22*h6*s*sina + A26*h6y) + (-h1*s*sina + h1x)*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+209] += detJ*wi*(h2y*(A12*h6x + A22*h6*s*sina + A26*h6y) + (-h2*s*sina + h2x)*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+210] += detJ*wi*(h3y*(A12*h6x + A22*h6*s*sina + A26*h6y) + (-h3*s*sina + h3x)*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+211] += detJ*wi*(h4y*(A12*h6x + A22*h6*s*sina + A26*h6y) + (-h4*s*sina + h4x)*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+212] += detJ*wi*(h5y*(A12*h6x + A22*h6*s*sina + A26*h6y) + (-h5*s*sina + h5x)*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+213] += detJ*wi*(h6y*(A12*h6x + A22*h6*s*sina + A26*h6y) + (-h6*s*sina + h6x)*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+214] += detJ*wi*(h7y*(A12*h6x + A22*h6*s*sina + A26*h6y) + (-h7*s*sina + h7x)*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+215] += detJ*wi*(h8y*(A12*h6x + A22*h6*s*sina + A26*h6y) + (-h8*s*sina + h8x)*(A16*h6x + A26*h6*s*sina + A66*h6y))
            k0[pos+216] += cosa*detJ*h1*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+217] += cosa*detJ*h2*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+218] += cosa*detJ*h3*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+219] += cosa*detJ*h4*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+220] += cosa*detJ*h5*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+221] += cosa*detJ*h6*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+222] += cosa*detJ*h7*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+223] += cosa*detJ*h8*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+224] += detJ*wi*(h1*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h1x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h1y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+225] += detJ*wi*(h2*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h2x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h2y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+226] += detJ*wi*(h3*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h3x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h3y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+227] += detJ*wi*(h4*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h4x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h4y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+228] += detJ*wi*(h5*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h5x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h5y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+229] += detJ*wi*(B11*(h6x*h6x) + 2*B12*h6*h6x*s*sina + 2*B16*h6x*h6y + B22*(h6*h6)*(s*s)*(sina*sina) + 2*B26*h6*h6y*s*sina + B66*(h6y*h6y))
            k0[pos+230] += detJ*wi*(h7*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h7x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h7y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+231] += detJ*wi*(h8*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h8x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h8y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+232] += detJ*wi*(h1y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h1*s*sina + h1x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+233] += detJ*wi*(h2y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h2*s*sina + h2x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+234] += detJ*wi*(h3y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h3*s*sina + h3x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+235] += detJ*wi*(h4y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h4*s*sina + h4x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+236] += detJ*wi*(h5y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h5*s*sina + h5x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+237] += detJ*wi*(h6y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h6*s*sina + h6x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+238] += detJ*wi*(h7y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h7*s*sina + h7x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+239] += detJ*wi*(h8y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h8*s*sina + h8x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+240] += detJ*wi*(h1*s*sina*(A12*h7x + A22*h7*s*sina + A26*h7y) + h1x*(A11*h7x + A12*h7*s*sina + A16*h7y) + h1y*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+241] += detJ*wi*(h2*s*sina*(A12*h7x + A22*h7*s*sina + A26*h7y) + h2x*(A11*h7x + A12*h7*s*sina + A16*h7y) + h2y*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+242] += detJ*wi*(h3*s*sina*(A12*h7x + A22*h7*s*sina + A26*h7y) + h3x*(A11*h7x + A12*h7*s*sina + A16*h7y) + h3y*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+243] += detJ*wi*(h4*s*sina*(A12*h7x + A22*h7*s*sina + A26*h7y) + h4x*(A11*h7x + A12*h7*s*sina + A16*h7y) + h4y*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+244] += detJ*wi*(h5*s*sina*(A12*h7x + A22*h7*s*sina + A26*h7y) + h5x*(A11*h7x + A12*h7*s*sina + A16*h7y) + h5y*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+245] += detJ*wi*(h6*s*sina*(A12*h7x + A22*h7*s*sina + A26*h7y) + h6x*(A11*h7x + A12*h7*s*sina + A16*h7y) + h6y*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+246] += detJ*wi*(A11*(h7x*h7x) + 2*A12*h7*h7x*s*sina + 2*A16*h7x*h7y + A22*(h7*h7)*(s*s)*(sina*sina) + 2*A26*h7*h7y*s*sina + A66*(h7y*h7y))
            k0[pos+247] += detJ*wi*(h8*s*sina*(A12*h7x + A22*h7*s*sina + A26*h7y) + h8x*(A11*h7x + A12*h7*s*sina + A16*h7y) + h8y*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+248] += detJ*wi*(h1y*(A12*h7x + A22*h7*s*sina + A26*h7y) + (-h1*s*sina + h1x)*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+249] += detJ*wi*(h2y*(A12*h7x + A22*h7*s*sina + A26*h7y) + (-h2*s*sina + h2x)*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+250] += detJ*wi*(h3y*(A12*h7x + A22*h7*s*sina + A26*h7y) + (-h3*s*sina + h3x)*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+251] += detJ*wi*(h4y*(A12*h7x + A22*h7*s*sina + A26*h7y) + (-h4*s*sina + h4x)*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+252] += detJ*wi*(h5y*(A12*h7x + A22*h7*s*sina + A26*h7y) + (-h5*s*sina + h5x)*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+253] += detJ*wi*(h6y*(A12*h7x + A22*h7*s*sina + A26*h7y) + (-h6*s*sina + h6x)*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+254] += detJ*wi*(h7y*(A12*h7x + A22*h7*s*sina + A26*h7y) + (-h7*s*sina + h7x)*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+255] += detJ*wi*(h8y*(A12*h7x + A22*h7*s*sina + A26*h7y) + (-h8*s*sina + h8x)*(A16*h7x + A26*h7*s*sina + A66*h7y))
            k0[pos+256] += cosa*detJ*h1*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+257] += cosa*detJ*h2*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+258] += cosa*detJ*h3*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+259] += cosa*detJ*h4*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+260] += cosa*detJ*h5*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+261] += cosa*detJ*h6*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+262] += cosa*detJ*h7*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+263] += cosa*detJ*h8*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+264] += detJ*wi*(h1*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h1x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h1y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+265] += detJ*wi*(h2*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h2x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h2y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+266] += detJ*wi*(h3*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h3x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h3y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+267] += detJ*wi*(h4*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h4x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h4y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+268] += detJ*wi*(h5*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h5x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h5y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+269] += detJ*wi*(h6*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h6x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h6y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+270] += detJ*wi*(B11*(h7x*h7x) + 2*B12*h7*h7x*s*sina + 2*B16*h7x*h7y + B22*(h7*h7)*(s*s)*(sina*sina) + 2*B26*h7*h7y*s*sina + B66*(h7y*h7y))
            k0[pos+271] += detJ*wi*(h8*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h8x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h8y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+272] += detJ*wi*(h1y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h1*s*sina + h1x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+273] += detJ*wi*(h2y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h2*s*sina + h2x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+274] += detJ*wi*(h3y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h3*s*sina + h3x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+275] += detJ*wi*(h4y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h4*s*sina + h4x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+276] += detJ*wi*(h5y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h5*s*sina + h5x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+277] += detJ*wi*(h6y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h6*s*sina + h6x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+278] += detJ*wi*(h7y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h7*s*sina + h7x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+279] += detJ*wi*(h8y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h8*s*sina + h8x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+280] += detJ*wi*(h1*s*sina*(A12*h8x + A22*h8*s*sina + A26*h8y) + h1x*(A11*h8x + A12*h8*s*sina + A16*h8y) + h1y*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+281] += detJ*wi*(h2*s*sina*(A12*h8x + A22*h8*s*sina + A26*h8y) + h2x*(A11*h8x + A12*h8*s*sina + A16*h8y) + h2y*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+282] += detJ*wi*(h3*s*sina*(A12*h8x + A22*h8*s*sina + A26*h8y) + h3x*(A11*h8x + A12*h8*s*sina + A16*h8y) + h3y*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+283] += detJ*wi*(h4*s*sina*(A12*h8x + A22*h8*s*sina + A26*h8y) + h4x*(A11*h8x + A12*h8*s*sina + A16*h8y) + h4y*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+284] += detJ*wi*(h5*s*sina*(A12*h8x + A22*h8*s*sina + A26*h8y) + h5x*(A11*h8x + A12*h8*s*sina + A16*h8y) + h5y*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+285] += detJ*wi*(h6*s*sina*(A12*h8x + A22*h8*s*sina + A26*h8y) + h6x*(A11*h8x + A12*h8*s*sina + A16*h8y) + h6y*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+286] += detJ*wi*(h7*s*sina*(A12*h8x + A22*h8*s*sina + A26*h8y) + h7x*(A11*h8x + A12*h8*s*sina + A16*h8y) + h7y*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+287] += detJ*wi*(A11*(h8x*h8x) + 2*A12*h8*h8x*s*sina + 2*A16*h8x*h8y + A22*(h8*h8)*(s*s)*(sina*sina) + 2*A26*h8*h8y*s*sina + A66*(h8y*h8y))
            k0[pos+288] += detJ*wi*(h1y*(A12*h8x + A22*h8*s*sina + A26*h8y) + (-h1*s*sina + h1x)*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+289] += detJ*wi*(h2y*(A12*h8x + A22*h8*s*sina + A26*h8y) + (-h2*s*sina + h2x)*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+290] += detJ*wi*(h3y*(A12*h8x + A22*h8*s*sina + A26*h8y) + (-h3*s*sina + h3x)*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+291] += detJ*wi*(h4y*(A12*h8x + A22*h8*s*sina + A26*h8y) + (-h4*s*sina + h4x)*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+292] += detJ*wi*(h5y*(A12*h8x + A22*h8*s*sina + A26*h8y) + (-h5*s*sina + h5x)*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+293] += detJ*wi*(h6y*(A12*h8x + A22*h8*s*sina + A26*h8y) + (-h6*s*sina + h6x)*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+294] += detJ*wi*(h7y*(A12*h8x + A22*h8*s*sina + A26*h8y) + (-h7*s*sina + h7x)*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+295] += detJ*wi*(h8y*(A12*h8x + A22*h8*s*sina + A26*h8y) + (-h8*s*sina + h8x)*(A16*h8x + A26*h8*s*sina + A66*h8y))
            k0[pos+296] += cosa*detJ*h1*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+297] += cosa*detJ*h2*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+298] += cosa*detJ*h3*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+299] += cosa*detJ*h4*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+300] += cosa*detJ*h5*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+301] += cosa*detJ*h6*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+302] += cosa*detJ*h7*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+303] += cosa*detJ*h8*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+304] += detJ*wi*(h1*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h1x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h1y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+305] += detJ*wi*(h2*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h2x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h2y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+306] += detJ*wi*(h3*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h3x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h3y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+307] += detJ*wi*(h4*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h4x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h4y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+308] += detJ*wi*(h5*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h5x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h5y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+309] += detJ*wi*(h6*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h6x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h6y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+310] += detJ*wi*(h7*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h7x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h7y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+311] += detJ*wi*(B11*(h8x*h8x) + 2*B12*h8*h8x*s*sina + 2*B16*h8x*h8y + B22*(h8*h8)*(s*s)*(sina*sina) + 2*B26*h8*h8y*s*sina + B66*(h8y*h8y))
            k0[pos+312] += detJ*wi*(h1y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h1*s*sina + h1x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+313] += detJ*wi*(h2y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h2*s*sina + h2x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+314] += detJ*wi*(h3y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h3*s*sina + h3x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+315] += detJ*wi*(h4y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h4*s*sina + h4x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+316] += detJ*wi*(h5y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h5*s*sina + h5x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+317] += detJ*wi*(h6y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h6*s*sina + h6x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+318] += detJ*wi*(h7y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h7*s*sina + h7x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+319] += detJ*wi*(h8y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h8*s*sina + h8x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+320] += detJ*wi*(h1*s*sina*(A22*h1y + A26*(-h1*s*sina + h1x)) + h1x*(A12*h1y + A16*(-h1*s*sina + h1x)) + h1y*(A26*h1y + A66*(-h1*s*sina + h1x)))
            k0[pos+321] += detJ*wi*(h2*s*sina*(A22*h1y + A26*(-h1*s*sina + h1x)) + h2x*(A12*h1y + A16*(-h1*s*sina + h1x)) + h2y*(A26*h1y + A66*(-h1*s*sina + h1x)))
            k0[pos+322] += detJ*wi*(h3*s*sina*(A22*h1y + A26*(-h1*s*sina + h1x)) + h3x*(A12*h1y + A16*(-h1*s*sina + h1x)) + h3y*(A26*h1y + A66*(-h1*s*sina + h1x)))
            k0[pos+323] += detJ*wi*(h4*s*sina*(A22*h1y + A26*(-h1*s*sina + h1x)) + h4x*(A12*h1y + A16*(-h1*s*sina + h1x)) + h4y*(A26*h1y + A66*(-h1*s*sina + h1x)))
            k0[pos+324] += detJ*wi*(h5*s*sina*(A22*h1y + A26*(-h1*s*sina + h1x)) + h5x*(A12*h1y + A16*(-h1*s*sina + h1x)) + h5y*(A26*h1y + A66*(-h1*s*sina + h1x)))
            k0[pos+325] += detJ*wi*(h6*s*sina*(A22*h1y + A26*(-h1*s*sina + h1x)) + h6x*(A12*h1y + A16*(-h1*s*sina + h1x)) + h6y*(A26*h1y + A66*(-h1*s*sina + h1x)))
            k0[pos+326] += detJ*wi*(h7*s*sina*(A22*h1y + A26*(-h1*s*sina + h1x)) + h7x*(A12*h1y + A16*(-h1*s*sina + h1x)) + h7y*(A26*h1y + A66*(-h1*s*sina + h1x)))
            k0[pos+327] += detJ*wi*(h8*s*sina*(A22*h1y + A26*(-h1*s*sina + h1x)) + h8x*(A12*h1y + A16*(-h1*s*sina + h1x)) + h8y*(A26*h1y + A66*(-h1*s*sina + h1x)))
            k0[pos+328] += detJ*wi*(A22*(h1y*h1y) + 2*A26*h1y*(-h1*s*sina + h1x) + A44*(cosa*cosa)*(h1*h1)*(s*s) + A66*(-h1*s*sina + h1x)**2)
            k0[pos+329] += detJ*wi*(A44*(cosa*cosa)*h1*h2*(s*s) + h2y*(A22*h1y + A26*(-h1*s*sina + h1x)) + (A26*h1y + A66*(-h1*s*sina + h1x))*(-h2*s*sina + h2x))
            k0[pos+330] += detJ*wi*(A44*(cosa*cosa)*h1*h3*(s*s) + h3y*(A22*h1y + A26*(-h1*s*sina + h1x)) + (A26*h1y + A66*(-h1*s*sina + h1x))*(-h3*s*sina + h3x))
            k0[pos+331] += detJ*wi*(A44*(cosa*cosa)*h1*h4*(s*s) + h4y*(A22*h1y + A26*(-h1*s*sina + h1x)) + (A26*h1y + A66*(-h1*s*sina + h1x))*(-h4*s*sina + h4x))
            k0[pos+332] += detJ*wi*(A44*(cosa*cosa)*h1*h5*(s*s) + h5y*(A22*h1y + A26*(-h1*s*sina + h1x)) + (A26*h1y + A66*(-h1*s*sina + h1x))*(-h5*s*sina + h5x))
            k0[pos+333] += detJ*wi*(A44*(cosa*cosa)*h1*h6*(s*s) + h6y*(A22*h1y + A26*(-h1*s*sina + h1x)) + (A26*h1y + A66*(-h1*s*sina + h1x))*(-h6*s*sina + h6x))
            k0[pos+334] += detJ*wi*(A44*(cosa*cosa)*h1*h7*(s*s) + h7y*(A22*h1y + A26*(-h1*s*sina + h1x)) + (A26*h1y + A66*(-h1*s*sina + h1x))*(-h7*s*sina + h7x))
            k0[pos+335] += detJ*wi*(A44*(cosa*cosa)*h1*h8*(s*s) + h8y*(A22*h1y + A26*(-h1*s*sina + h1x)) + (A26*h1y + A66*(-h1*s*sina + h1x))*(-h8*s*sina + h8x))
            k0[pos+336] += -cosa*detJ*h1*s*wi*(-A22*h1y + A26*h1*s*sina - A26*h1x + A44*h1y + A45*h1x)
            k0[pos+337] += -cosa*detJ*s*wi*(-A22*h1y*h2 + A26*h1*h2*s*sina - A26*h1x*h2 + A44*h1*h2y + A45*h1*h2x)
            k0[pos+338] += -cosa*detJ*s*wi*(-A22*h1y*h3 + A26*h1*h3*s*sina - A26*h1x*h3 + A44*h1*h3y + A45*h1*h3x)
            k0[pos+339] += -cosa*detJ*s*wi*(-A22*h1y*h4 + A26*h1*h4*s*sina - A26*h1x*h4 + A44*h1*h4y + A45*h1*h4x)
            k0[pos+340] += -cosa*detJ*s*wi*(-A22*h1y*h5 + A26*h1*h5*s*sina - A26*h1x*h5 + A44*h1*h5y + A45*h1*h5x)
            k0[pos+341] += -cosa*detJ*s*wi*(-A22*h1y*h6 + A26*h1*h6*s*sina - A26*h1x*h6 + A44*h1*h6y + A45*h1*h6x)
            k0[pos+342] += -cosa*detJ*s*wi*(-A22*h1y*h7 + A26*h1*h7*s*sina - A26*h1x*h7 + A44*h1*h7y + A45*h1*h7x)
            k0[pos+343] += -cosa*detJ*s*wi*(-A22*h1y*h8 + A26*h1*h8*s*sina - A26*h1x*h8 + A44*h1*h8y + A45*h1*h8x)
            k0[pos+344] += detJ*wi*(-A45*cosa*(h1*h1)*s + h1*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h1x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h1y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+345] += detJ*wi*(-A45*cosa*h1*h2*s + h2*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h2x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h2y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+346] += detJ*wi*(-A45*cosa*h1*h3*s + h3*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h3x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h3y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+347] += detJ*wi*(-A45*cosa*h1*h4*s + h4*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h4x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h4y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+348] += detJ*wi*(-A45*cosa*h1*h5*s + h5*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h5x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h5y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+349] += detJ*wi*(-A45*cosa*h1*h6*s + h6*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h6x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h6y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+350] += detJ*wi*(-A45*cosa*h1*h7*s + h7*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h7x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h7y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+351] += detJ*wi*(-A45*cosa*h1*h8*s + h8*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h8x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h8y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+352] += detJ*wi*(-A44*cosa*(h1*h1)*s + B22*(h1y*h1y) + 2*B26*h1y*(-h1*s*sina + h1x) + B66*(-h1*s*sina + h1x)**2)
            k0[pos+353] += detJ*wi*(-A44*cosa*h1*h2*s + h2y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h2*s*sina + h2x))
            k0[pos+354] += detJ*wi*(-A44*cosa*h1*h3*s + h3y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h3*s*sina + h3x))
            k0[pos+355] += detJ*wi*(-A44*cosa*h1*h4*s + h4y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h4*s*sina + h4x))
            k0[pos+356] += detJ*wi*(-A44*cosa*h1*h5*s + h5y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h5*s*sina + h5x))
            k0[pos+357] += detJ*wi*(-A44*cosa*h1*h6*s + h6y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h6*s*sina + h6x))
            k0[pos+358] += detJ*wi*(-A44*cosa*h1*h7*s + h7y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h7*s*sina + h7x))
            k0[pos+359] += detJ*wi*(-A44*cosa*h1*h8*s + h8y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h8*s*sina + h8x))
            k0[pos+360] += detJ*wi*(h1*s*sina*(A22*h2y + A26*(-h2*s*sina + h2x)) + h1x*(A12*h2y + A16*(-h2*s*sina + h2x)) + h1y*(A26*h2y + A66*(-h2*s*sina + h2x)))
            k0[pos+361] += detJ*wi*(h2*s*sina*(A22*h2y + A26*(-h2*s*sina + h2x)) + h2x*(A12*h2y + A16*(-h2*s*sina + h2x)) + h2y*(A26*h2y + A66*(-h2*s*sina + h2x)))
            k0[pos+362] += detJ*wi*(h3*s*sina*(A22*h2y + A26*(-h2*s*sina + h2x)) + h3x*(A12*h2y + A16*(-h2*s*sina + h2x)) + h3y*(A26*h2y + A66*(-h2*s*sina + h2x)))
            k0[pos+363] += detJ*wi*(h4*s*sina*(A22*h2y + A26*(-h2*s*sina + h2x)) + h4x*(A12*h2y + A16*(-h2*s*sina + h2x)) + h4y*(A26*h2y + A66*(-h2*s*sina + h2x)))
            k0[pos+364] += detJ*wi*(h5*s*sina*(A22*h2y + A26*(-h2*s*sina + h2x)) + h5x*(A12*h2y + A16*(-h2*s*sina + h2x)) + h5y*(A26*h2y + A66*(-h2*s*sina + h2x)))
            k0[pos+365] += detJ*wi*(h6*s*sina*(A22*h2y + A26*(-h2*s*sina + h2x)) + h6x*(A12*h2y + A16*(-h2*s*sina + h2x)) + h6y*(A26*h2y + A66*(-h2*s*sina + h2x)))
            k0[pos+366] += detJ*wi*(h7*s*sina*(A22*h2y + A26*(-h2*s*sina + h2x)) + h7x*(A12*h2y + A16*(-h2*s*sina + h2x)) + h7y*(A26*h2y + A66*(-h2*s*sina + h2x)))
            k0[pos+367] += detJ*wi*(h8*s*sina*(A22*h2y + A26*(-h2*s*sina + h2x)) + h8x*(A12*h2y + A16*(-h2*s*sina + h2x)) + h8y*(A26*h2y + A66*(-h2*s*sina + h2x)))
            k0[pos+368] += detJ*wi*(A44*(cosa*cosa)*h1*h2*(s*s) + h1y*(A22*h2y + A26*(-h2*s*sina + h2x)) + (A26*h2y + A66*(-h2*s*sina + h2x))*(-h1*s*sina + h1x))
            k0[pos+369] += detJ*wi*(A22*(h2y*h2y) + 2*A26*h2y*(-h2*s*sina + h2x) + A44*(cosa*cosa)*(h2*h2)*(s*s) + A66*(-h2*s*sina + h2x)**2)
            k0[pos+370] += detJ*wi*(A44*(cosa*cosa)*h2*h3*(s*s) + h3y*(A22*h2y + A26*(-h2*s*sina + h2x)) + (A26*h2y + A66*(-h2*s*sina + h2x))*(-h3*s*sina + h3x))
            k0[pos+371] += detJ*wi*(A44*(cosa*cosa)*h2*h4*(s*s) + h4y*(A22*h2y + A26*(-h2*s*sina + h2x)) + (A26*h2y + A66*(-h2*s*sina + h2x))*(-h4*s*sina + h4x))
            k0[pos+372] += detJ*wi*(A44*(cosa*cosa)*h2*h5*(s*s) + h5y*(A22*h2y + A26*(-h2*s*sina + h2x)) + (A26*h2y + A66*(-h2*s*sina + h2x))*(-h5*s*sina + h5x))
            k0[pos+373] += detJ*wi*(A44*(cosa*cosa)*h2*h6*(s*s) + h6y*(A22*h2y + A26*(-h2*s*sina + h2x)) + (A26*h2y + A66*(-h2*s*sina + h2x))*(-h6*s*sina + h6x))
            k0[pos+374] += detJ*wi*(A44*(cosa*cosa)*h2*h7*(s*s) + h7y*(A22*h2y + A26*(-h2*s*sina + h2x)) + (A26*h2y + A66*(-h2*s*sina + h2x))*(-h7*s*sina + h7x))
            k0[pos+375] += detJ*wi*(A44*(cosa*cosa)*h2*h8*(s*s) + h8y*(A22*h2y + A26*(-h2*s*sina + h2x)) + (A26*h2y + A66*(-h2*s*sina + h2x))*(-h8*s*sina + h8x))
            k0[pos+376] += -cosa*detJ*s*wi*(A44*h1y*h2 + A45*h1x*h2 - h1*(A22*h2y - A26*h2*s*sina + A26*h2x))
            k0[pos+377] += -cosa*detJ*h2*s*wi*(-A22*h2y + A26*h2*s*sina - A26*h2x + A44*h2y + A45*h2x)
            k0[pos+378] += -cosa*detJ*s*wi*(-A22*h2y*h3 + A26*h2*h3*s*sina - A26*h2x*h3 + A44*h2*h3y + A45*h2*h3x)
            k0[pos+379] += -cosa*detJ*s*wi*(-A22*h2y*h4 + A26*h2*h4*s*sina - A26*h2x*h4 + A44*h2*h4y + A45*h2*h4x)
            k0[pos+380] += -cosa*detJ*s*wi*(-A22*h2y*h5 + A26*h2*h5*s*sina - A26*h2x*h5 + A44*h2*h5y + A45*h2*h5x)
            k0[pos+381] += -cosa*detJ*s*wi*(-A22*h2y*h6 + A26*h2*h6*s*sina - A26*h2x*h6 + A44*h2*h6y + A45*h2*h6x)
            k0[pos+382] += -cosa*detJ*s*wi*(-A22*h2y*h7 + A26*h2*h7*s*sina - A26*h2x*h7 + A44*h2*h7y + A45*h2*h7x)
            k0[pos+383] += -cosa*detJ*s*wi*(-A22*h2y*h8 + A26*h2*h8*s*sina - A26*h2x*h8 + A44*h2*h8y + A45*h2*h8x)
            k0[pos+384] += detJ*wi*(-A45*cosa*h1*h2*s + h1*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h1x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h1y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+385] += detJ*wi*(-A45*cosa*(h2*h2)*s + h2*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h2x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h2y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+386] += detJ*wi*(-A45*cosa*h2*h3*s + h3*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h3x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h3y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+387] += detJ*wi*(-A45*cosa*h2*h4*s + h4*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h4x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h4y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+388] += detJ*wi*(-A45*cosa*h2*h5*s + h5*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h5x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h5y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+389] += detJ*wi*(-A45*cosa*h2*h6*s + h6*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h6x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h6y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+390] += detJ*wi*(-A45*cosa*h2*h7*s + h7*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h7x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h7y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+391] += detJ*wi*(-A45*cosa*h2*h8*s + h8*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h8x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h8y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+392] += detJ*wi*(-A44*cosa*h1*h2*s + h1y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h1*s*sina + h1x))
            k0[pos+393] += detJ*wi*(-A44*cosa*(h2*h2)*s + B22*(h2y*h2y) + 2*B26*h2y*(-h2*s*sina + h2x) + B66*(-h2*s*sina + h2x)**2)
            k0[pos+394] += detJ*wi*(-A44*cosa*h2*h3*s + h3y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h3*s*sina + h3x))
            k0[pos+395] += detJ*wi*(-A44*cosa*h2*h4*s + h4y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h4*s*sina + h4x))
            k0[pos+396] += detJ*wi*(-A44*cosa*h2*h5*s + h5y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h5*s*sina + h5x))
            k0[pos+397] += detJ*wi*(-A44*cosa*h2*h6*s + h6y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h6*s*sina + h6x))
            k0[pos+398] += detJ*wi*(-A44*cosa*h2*h7*s + h7y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h7*s*sina + h7x))
            k0[pos+399] += detJ*wi*(-A44*cosa*h2*h8*s + h8y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h8*s*sina + h8x))
            k0[pos+400] += detJ*wi*(h1*s*sina*(A22*h3y + A26*(-h3*s*sina + h3x)) + h1x*(A12*h3y + A16*(-h3*s*sina + h3x)) + h1y*(A26*h3y + A66*(-h3*s*sina + h3x)))
            k0[pos+401] += detJ*wi*(h2*s*sina*(A22*h3y + A26*(-h3*s*sina + h3x)) + h2x*(A12*h3y + A16*(-h3*s*sina + h3x)) + h2y*(A26*h3y + A66*(-h3*s*sina + h3x)))
            k0[pos+402] += detJ*wi*(h3*s*sina*(A22*h3y + A26*(-h3*s*sina + h3x)) + h3x*(A12*h3y + A16*(-h3*s*sina + h3x)) + h3y*(A26*h3y + A66*(-h3*s*sina + h3x)))
            k0[pos+403] += detJ*wi*(h4*s*sina*(A22*h3y + A26*(-h3*s*sina + h3x)) + h4x*(A12*h3y + A16*(-h3*s*sina + h3x)) + h4y*(A26*h3y + A66*(-h3*s*sina + h3x)))
            k0[pos+404] += detJ*wi*(h5*s*sina*(A22*h3y + A26*(-h3*s*sina + h3x)) + h5x*(A12*h3y + A16*(-h3*s*sina + h3x)) + h5y*(A26*h3y + A66*(-h3*s*sina + h3x)))
            k0[pos+405] += detJ*wi*(h6*s*sina*(A22*h3y + A26*(-h3*s*sina + h3x)) + h6x*(A12*h3y + A16*(-h3*s*sina + h3x)) + h6y*(A26*h3y + A66*(-h3*s*sina + h3x)))
            k0[pos+406] += detJ*wi*(h7*s*sina*(A22*h3y + A26*(-h3*s*sina + h3x)) + h7x*(A12*h3y + A16*(-h3*s*sina + h3x)) + h7y*(A26*h3y + A66*(-h3*s*sina + h3x)))
            k0[pos+407] += detJ*wi*(h8*s*sina*(A22*h3y + A26*(-h3*s*sina + h3x)) + h8x*(A12*h3y + A16*(-h3*s*sina + h3x)) + h8y*(A26*h3y + A66*(-h3*s*sina + h3x)))
            k0[pos+408] += detJ*wi*(A44*(cosa*cosa)*h1*h3*(s*s) + h1y*(A22*h3y + A26*(-h3*s*sina + h3x)) + (A26*h3y + A66*(-h3*s*sina + h3x))*(-h1*s*sina + h1x))
            k0[pos+409] += detJ*wi*(A44*(cosa*cosa)*h2*h3*(s*s) + h2y*(A22*h3y + A26*(-h3*s*sina + h3x)) + (A26*h3y + A66*(-h3*s*sina + h3x))*(-h2*s*sina + h2x))
            k0[pos+410] += detJ*wi*(A22*(h3y*h3y) + 2*A26*h3y*(-h3*s*sina + h3x) + A44*(cosa*cosa)*(h3*h3)*(s*s) + A66*(-h3*s*sina + h3x)**2)
            k0[pos+411] += detJ*wi*(A44*(cosa*cosa)*h3*h4*(s*s) + h4y*(A22*h3y + A26*(-h3*s*sina + h3x)) + (A26*h3y + A66*(-h3*s*sina + h3x))*(-h4*s*sina + h4x))
            k0[pos+412] += detJ*wi*(A44*(cosa*cosa)*h3*h5*(s*s) + h5y*(A22*h3y + A26*(-h3*s*sina + h3x)) + (A26*h3y + A66*(-h3*s*sina + h3x))*(-h5*s*sina + h5x))
            k0[pos+413] += detJ*wi*(A44*(cosa*cosa)*h3*h6*(s*s) + h6y*(A22*h3y + A26*(-h3*s*sina + h3x)) + (A26*h3y + A66*(-h3*s*sina + h3x))*(-h6*s*sina + h6x))
            k0[pos+414] += detJ*wi*(A44*(cosa*cosa)*h3*h7*(s*s) + h7y*(A22*h3y + A26*(-h3*s*sina + h3x)) + (A26*h3y + A66*(-h3*s*sina + h3x))*(-h7*s*sina + h7x))
            k0[pos+415] += detJ*wi*(A44*(cosa*cosa)*h3*h8*(s*s) + h8y*(A22*h3y + A26*(-h3*s*sina + h3x)) + (A26*h3y + A66*(-h3*s*sina + h3x))*(-h8*s*sina + h8x))
            k0[pos+416] += -cosa*detJ*s*wi*(A44*h1y*h3 + A45*h1x*h3 - h1*(A22*h3y - A26*h3*s*sina + A26*h3x))
            k0[pos+417] += -cosa*detJ*s*wi*(A44*h2y*h3 + A45*h2x*h3 - h2*(A22*h3y - A26*h3*s*sina + A26*h3x))
            k0[pos+418] += -cosa*detJ*h3*s*wi*(-A22*h3y + A26*h3*s*sina - A26*h3x + A44*h3y + A45*h3x)
            k0[pos+419] += -cosa*detJ*s*wi*(-A22*h3y*h4 + A26*h3*h4*s*sina - A26*h3x*h4 + A44*h3*h4y + A45*h3*h4x)
            k0[pos+420] += -cosa*detJ*s*wi*(-A22*h3y*h5 + A26*h3*h5*s*sina - A26*h3x*h5 + A44*h3*h5y + A45*h3*h5x)
            k0[pos+421] += -cosa*detJ*s*wi*(-A22*h3y*h6 + A26*h3*h6*s*sina - A26*h3x*h6 + A44*h3*h6y + A45*h3*h6x)
            k0[pos+422] += -cosa*detJ*s*wi*(-A22*h3y*h7 + A26*h3*h7*s*sina - A26*h3x*h7 + A44*h3*h7y + A45*h3*h7x)
            k0[pos+423] += -cosa*detJ*s*wi*(-A22*h3y*h8 + A26*h3*h8*s*sina - A26*h3x*h8 + A44*h3*h8y + A45*h3*h8x)
            k0[pos+424] += detJ*wi*(-A45*cosa*h1*h3*s + h1*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h1x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h1y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+425] += detJ*wi*(-A45*cosa*h2*h3*s + h2*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h2x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h2y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+426] += detJ*wi*(-A45*cosa*(h3*h3)*s + h3*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h3x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h3y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+427] += detJ*wi*(-A45*cosa*h3*h4*s + h4*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h4x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h4y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+428] += detJ*wi*(-A45*cosa*h3*h5*s + h5*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h5x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h5y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+429] += detJ*wi*(-A45*cosa*h3*h6*s + h6*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h6x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h6y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+430] += detJ*wi*(-A45*cosa*h3*h7*s + h7*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h7x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h7y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+431] += detJ*wi*(-A45*cosa*h3*h8*s + h8*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h8x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h8y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+432] += detJ*wi*(-A44*cosa*h1*h3*s + h1y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h1*s*sina + h1x))
            k0[pos+433] += detJ*wi*(-A44*cosa*h2*h3*s + h2y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h2*s*sina + h2x))
            k0[pos+434] += detJ*wi*(-A44*cosa*(h3*h3)*s + B22*(h3y*h3y) + 2*B26*h3y*(-h3*s*sina + h3x) + B66*(-h3*s*sina + h3x)**2)
            k0[pos+435] += detJ*wi*(-A44*cosa*h3*h4*s + h4y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h4*s*sina + h4x))
            k0[pos+436] += detJ*wi*(-A44*cosa*h3*h5*s + h5y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h5*s*sina + h5x))
            k0[pos+437] += detJ*wi*(-A44*cosa*h3*h6*s + h6y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h6*s*sina + h6x))
            k0[pos+438] += detJ*wi*(-A44*cosa*h3*h7*s + h7y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h7*s*sina + h7x))
            k0[pos+439] += detJ*wi*(-A44*cosa*h3*h8*s + h8y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h8*s*sina + h8x))
            k0[pos+440] += detJ*wi*(h1*s*sina*(A22*h4y + A26*(-h4*s*sina + h4x)) + h1x*(A12*h4y + A16*(-h4*s*sina + h4x)) + h1y*(A26*h4y + A66*(-h4*s*sina + h4x)))
            k0[pos+441] += detJ*wi*(h2*s*sina*(A22*h4y + A26*(-h4*s*sina + h4x)) + h2x*(A12*h4y + A16*(-h4*s*sina + h4x)) + h2y*(A26*h4y + A66*(-h4*s*sina + h4x)))
            k0[pos+442] += detJ*wi*(h3*s*sina*(A22*h4y + A26*(-h4*s*sina + h4x)) + h3x*(A12*h4y + A16*(-h4*s*sina + h4x)) + h3y*(A26*h4y + A66*(-h4*s*sina + h4x)))
            k0[pos+443] += detJ*wi*(h4*s*sina*(A22*h4y + A26*(-h4*s*sina + h4x)) + h4x*(A12*h4y + A16*(-h4*s*sina + h4x)) + h4y*(A26*h4y + A66*(-h4*s*sina + h4x)))
            k0[pos+444] += detJ*wi*(h5*s*sina*(A22*h4y + A26*(-h4*s*sina + h4x)) + h5x*(A12*h4y + A16*(-h4*s*sina + h4x)) + h5y*(A26*h4y + A66*(-h4*s*sina + h4x)))
            k0[pos+445] += detJ*wi*(h6*s*sina*(A22*h4y + A26*(-h4*s*sina + h4x)) + h6x*(A12*h4y + A16*(-h4*s*sina + h4x)) + h6y*(A26*h4y + A66*(-h4*s*sina + h4x)))
            k0[pos+446] += detJ*wi*(h7*s*sina*(A22*h4y + A26*(-h4*s*sina + h4x)) + h7x*(A12*h4y + A16*(-h4*s*sina + h4x)) + h7y*(A26*h4y + A66*(-h4*s*sina + h4x)))
            k0[pos+447] += detJ*wi*(h8*s*sina*(A22*h4y + A26*(-h4*s*sina + h4x)) + h8x*(A12*h4y + A16*(-h4*s*sina + h4x)) + h8y*(A26*h4y + A66*(-h4*s*sina + h4x)))
            k0[pos+448] += detJ*wi*(A44*(cosa*cosa)*h1*h4*(s*s) + h1y*(A22*h4y + A26*(-h4*s*sina + h4x)) + (A26*h4y + A66*(-h4*s*sina + h4x))*(-h1*s*sina + h1x))
            k0[pos+449] += detJ*wi*(A44*(cosa*cosa)*h2*h4*(s*s) + h2y*(A22*h4y + A26*(-h4*s*sina + h4x)) + (A26*h4y + A66*(-h4*s*sina + h4x))*(-h2*s*sina + h2x))
            k0[pos+450] += detJ*wi*(A44*(cosa*cosa)*h3*h4*(s*s) + h3y*(A22*h4y + A26*(-h4*s*sina + h4x)) + (A26*h4y + A66*(-h4*s*sina + h4x))*(-h3*s*sina + h3x))
            k0[pos+451] += detJ*wi*(A22*(h4y*h4y) + 2*A26*h4y*(-h4*s*sina + h4x) + A44*(cosa*cosa)*(h4*h4)*(s*s) + A66*(-h4*s*sina + h4x)**2)
            k0[pos+452] += detJ*wi*(A44*(cosa*cosa)*h4*h5*(s*s) + h5y*(A22*h4y + A26*(-h4*s*sina + h4x)) + (A26*h4y + A66*(-h4*s*sina + h4x))*(-h5*s*sina + h5x))
            k0[pos+453] += detJ*wi*(A44*(cosa*cosa)*h4*h6*(s*s) + h6y*(A22*h4y + A26*(-h4*s*sina + h4x)) + (A26*h4y + A66*(-h4*s*sina + h4x))*(-h6*s*sina + h6x))
            k0[pos+454] += detJ*wi*(A44*(cosa*cosa)*h4*h7*(s*s) + h7y*(A22*h4y + A26*(-h4*s*sina + h4x)) + (A26*h4y + A66*(-h4*s*sina + h4x))*(-h7*s*sina + h7x))
            k0[pos+455] += detJ*wi*(A44*(cosa*cosa)*h4*h8*(s*s) + h8y*(A22*h4y + A26*(-h4*s*sina + h4x)) + (A26*h4y + A66*(-h4*s*sina + h4x))*(-h8*s*sina + h8x))
            k0[pos+456] += -cosa*detJ*s*wi*(A44*h1y*h4 + A45*h1x*h4 - h1*(A22*h4y - A26*h4*s*sina + A26*h4x))
            k0[pos+457] += -cosa*detJ*s*wi*(A44*h2y*h4 + A45*h2x*h4 - h2*(A22*h4y - A26*h4*s*sina + A26*h4x))
            k0[pos+458] += -cosa*detJ*s*wi*(A44*h3y*h4 + A45*h3x*h4 - h3*(A22*h4y - A26*h4*s*sina + A26*h4x))
            k0[pos+459] += -cosa*detJ*h4*s*wi*(-A22*h4y + A26*h4*s*sina - A26*h4x + A44*h4y + A45*h4x)
            k0[pos+460] += -cosa*detJ*s*wi*(-A22*h4y*h5 + A26*h4*h5*s*sina - A26*h4x*h5 + A44*h4*h5y + A45*h4*h5x)
            k0[pos+461] += -cosa*detJ*s*wi*(-A22*h4y*h6 + A26*h4*h6*s*sina - A26*h4x*h6 + A44*h4*h6y + A45*h4*h6x)
            k0[pos+462] += -cosa*detJ*s*wi*(-A22*h4y*h7 + A26*h4*h7*s*sina - A26*h4x*h7 + A44*h4*h7y + A45*h4*h7x)
            k0[pos+463] += -cosa*detJ*s*wi*(-A22*h4y*h8 + A26*h4*h8*s*sina - A26*h4x*h8 + A44*h4*h8y + A45*h4*h8x)
            k0[pos+464] += detJ*wi*(-A45*cosa*h1*h4*s + h1*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h1x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h1y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+465] += detJ*wi*(-A45*cosa*h2*h4*s + h2*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h2x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h2y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+466] += detJ*wi*(-A45*cosa*h3*h4*s + h3*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h3x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h3y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+467] += detJ*wi*(-A45*cosa*(h4*h4)*s + h4*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h4x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h4y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+468] += detJ*wi*(-A45*cosa*h4*h5*s + h5*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h5x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h5y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+469] += detJ*wi*(-A45*cosa*h4*h6*s + h6*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h6x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h6y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+470] += detJ*wi*(-A45*cosa*h4*h7*s + h7*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h7x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h7y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+471] += detJ*wi*(-A45*cosa*h4*h8*s + h8*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h8x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h8y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+472] += detJ*wi*(-A44*cosa*h1*h4*s + h1y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h1*s*sina + h1x))
            k0[pos+473] += detJ*wi*(-A44*cosa*h2*h4*s + h2y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h2*s*sina + h2x))
            k0[pos+474] += detJ*wi*(-A44*cosa*h3*h4*s + h3y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h3*s*sina + h3x))
            k0[pos+475] += detJ*wi*(-A44*cosa*(h4*h4)*s + B22*(h4y*h4y) + 2*B26*h4y*(-h4*s*sina + h4x) + B66*(-h4*s*sina + h4x)**2)
            k0[pos+476] += detJ*wi*(-A44*cosa*h4*h5*s + h5y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h5*s*sina + h5x))
            k0[pos+477] += detJ*wi*(-A44*cosa*h4*h6*s + h6y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h6*s*sina + h6x))
            k0[pos+478] += detJ*wi*(-A44*cosa*h4*h7*s + h7y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h7*s*sina + h7x))
            k0[pos+479] += detJ*wi*(-A44*cosa*h4*h8*s + h8y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h8*s*sina + h8x))
            k0[pos+480] += detJ*wi*(h1*s*sina*(A22*h5y + A26*(-h5*s*sina + h5x)) + h1x*(A12*h5y + A16*(-h5*s*sina + h5x)) + h1y*(A26*h5y + A66*(-h5*s*sina + h5x)))
            k0[pos+481] += detJ*wi*(h2*s*sina*(A22*h5y + A26*(-h5*s*sina + h5x)) + h2x*(A12*h5y + A16*(-h5*s*sina + h5x)) + h2y*(A26*h5y + A66*(-h5*s*sina + h5x)))
            k0[pos+482] += detJ*wi*(h3*s*sina*(A22*h5y + A26*(-h5*s*sina + h5x)) + h3x*(A12*h5y + A16*(-h5*s*sina + h5x)) + h3y*(A26*h5y + A66*(-h5*s*sina + h5x)))
            k0[pos+483] += detJ*wi*(h4*s*sina*(A22*h5y + A26*(-h5*s*sina + h5x)) + h4x*(A12*h5y + A16*(-h5*s*sina + h5x)) + h4y*(A26*h5y + A66*(-h5*s*sina + h5x)))
            k0[pos+484] += detJ*wi*(h5*s*sina*(A22*h5y + A26*(-h5*s*sina + h5x)) + h5x*(A12*h5y + A16*(-h5*s*sina + h5x)) + h5y*(A26*h5y + A66*(-h5*s*sina + h5x)))
            k0[pos+485] += detJ*wi*(h6*s*sina*(A22*h5y + A26*(-h5*s*sina + h5x)) + h6x*(A12*h5y + A16*(-h5*s*sina + h5x)) + h6y*(A26*h5y + A66*(-h5*s*sina + h5x)))
            k0[pos+486] += detJ*wi*(h7*s*sina*(A22*h5y + A26*(-h5*s*sina + h5x)) + h7x*(A12*h5y + A16*(-h5*s*sina + h5x)) + h7y*(A26*h5y + A66*(-h5*s*sina + h5x)))
            k0[pos+487] += detJ*wi*(h8*s*sina*(A22*h5y + A26*(-h5*s*sina + h5x)) + h8x*(A12*h5y + A16*(-h5*s*sina + h5x)) + h8y*(A26*h5y + A66*(-h5*s*sina + h5x)))
            k0[pos+488] += detJ*wi*(A44*(cosa*cosa)*h1*h5*(s*s) + h1y*(A22*h5y + A26*(-h5*s*sina + h5x)) + (A26*h5y + A66*(-h5*s*sina + h5x))*(-h1*s*sina + h1x))
            k0[pos+489] += detJ*wi*(A44*(cosa*cosa)*h2*h5*(s*s) + h2y*(A22*h5y + A26*(-h5*s*sina + h5x)) + (A26*h5y + A66*(-h5*s*sina + h5x))*(-h2*s*sina + h2x))
            k0[pos+490] += detJ*wi*(A44*(cosa*cosa)*h3*h5*(s*s) + h3y*(A22*h5y + A26*(-h5*s*sina + h5x)) + (A26*h5y + A66*(-h5*s*sina + h5x))*(-h3*s*sina + h3x))
            k0[pos+491] += detJ*wi*(A44*(cosa*cosa)*h4*h5*(s*s) + h4y*(A22*h5y + A26*(-h5*s*sina + h5x)) + (A26*h5y + A66*(-h5*s*sina + h5x))*(-h4*s*sina + h4x))
            k0[pos+492] += detJ*wi*(A22*(h5y*h5y) + 2*A26*h5y*(-h5*s*sina + h5x) + A44*(cosa*cosa)*(h5*h5)*(s*s) + A66*(-h5*s*sina + h5x)**2)
            k0[pos+493] += detJ*wi*(A44*(cosa*cosa)*h5*h6*(s*s) + h6y*(A22*h5y + A26*(-h5*s*sina + h5x)) + (A26*h5y + A66*(-h5*s*sina + h5x))*(-h6*s*sina + h6x))
            k0[pos+494] += detJ*wi*(A44*(cosa*cosa)*h5*h7*(s*s) + h7y*(A22*h5y + A26*(-h5*s*sina + h5x)) + (A26*h5y + A66*(-h5*s*sina + h5x))*(-h7*s*sina + h7x))
            k0[pos+495] += detJ*wi*(A44*(cosa*cosa)*h5*h8*(s*s) + h8y*(A22*h5y + A26*(-h5*s*sina + h5x)) + (A26*h5y + A66*(-h5*s*sina + h5x))*(-h8*s*sina + h8x))
            k0[pos+496] += -cosa*detJ*s*wi*(A44*h1y*h5 + A45*h1x*h5 - h1*(A22*h5y - A26*h5*s*sina + A26*h5x))
            k0[pos+497] += -cosa*detJ*s*wi*(A44*h2y*h5 + A45*h2x*h5 - h2*(A22*h5y - A26*h5*s*sina + A26*h5x))
            k0[pos+498] += -cosa*detJ*s*wi*(A44*h3y*h5 + A45*h3x*h5 - h3*(A22*h5y - A26*h5*s*sina + A26*h5x))
            k0[pos+499] += -cosa*detJ*s*wi*(A44*h4y*h5 + A45*h4x*h5 - h4*(A22*h5y - A26*h5*s*sina + A26*h5x))
            k0[pos+500] += -cosa*detJ*h5*s*wi*(-A22*h5y + A26*h5*s*sina - A26*h5x + A44*h5y + A45*h5x)
            k0[pos+501] += -cosa*detJ*s*wi*(-A22*h5y*h6 + A26*h5*h6*s*sina - A26*h5x*h6 + A44*h5*h6y + A45*h5*h6x)
            k0[pos+502] += -cosa*detJ*s*wi*(-A22*h5y*h7 + A26*h5*h7*s*sina - A26*h5x*h7 + A44*h5*h7y + A45*h5*h7x)
            k0[pos+503] += -cosa*detJ*s*wi*(-A22*h5y*h8 + A26*h5*h8*s*sina - A26*h5x*h8 + A44*h5*h8y + A45*h5*h8x)
            k0[pos+504] += detJ*wi*(-A45*cosa*h1*h5*s + h1*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h1x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h1y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+505] += detJ*wi*(-A45*cosa*h2*h5*s + h2*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h2x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h2y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+506] += detJ*wi*(-A45*cosa*h3*h5*s + h3*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h3x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h3y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+507] += detJ*wi*(-A45*cosa*h4*h5*s + h4*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h4x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h4y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+508] += detJ*wi*(-A45*cosa*(h5*h5)*s + h5*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h5x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h5y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+509] += detJ*wi*(-A45*cosa*h5*h6*s + h6*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h6x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h6y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+510] += detJ*wi*(-A45*cosa*h5*h7*s + h7*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h7x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h7y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+511] += detJ*wi*(-A45*cosa*h5*h8*s + h8*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h8x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h8y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+512] += detJ*wi*(-A44*cosa*h1*h5*s + h1y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h1*s*sina + h1x))
            k0[pos+513] += detJ*wi*(-A44*cosa*h2*h5*s + h2y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h2*s*sina + h2x))
            k0[pos+514] += detJ*wi*(-A44*cosa*h3*h5*s + h3y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h3*s*sina + h3x))
            k0[pos+515] += detJ*wi*(-A44*cosa*h4*h5*s + h4y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h4*s*sina + h4x))
            k0[pos+516] += detJ*wi*(-A44*cosa*(h5*h5)*s + B22*(h5y*h5y) + 2*B26*h5y*(-h5*s*sina + h5x) + B66*(-h5*s*sina + h5x)**2)
            k0[pos+517] += detJ*wi*(-A44*cosa*h5*h6*s + h6y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h6*s*sina + h6x))
            k0[pos+518] += detJ*wi*(-A44*cosa*h5*h7*s + h7y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h7*s*sina + h7x))
            k0[pos+519] += detJ*wi*(-A44*cosa*h5*h8*s + h8y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h8*s*sina + h8x))
            k0[pos+520] += detJ*wi*(h1*s*sina*(A22*h6y + A26*(-h6*s*sina + h6x)) + h1x*(A12*h6y + A16*(-h6*s*sina + h6x)) + h1y*(A26*h6y + A66*(-h6*s*sina + h6x)))
            k0[pos+521] += detJ*wi*(h2*s*sina*(A22*h6y + A26*(-h6*s*sina + h6x)) + h2x*(A12*h6y + A16*(-h6*s*sina + h6x)) + h2y*(A26*h6y + A66*(-h6*s*sina + h6x)))
            k0[pos+522] += detJ*wi*(h3*s*sina*(A22*h6y + A26*(-h6*s*sina + h6x)) + h3x*(A12*h6y + A16*(-h6*s*sina + h6x)) + h3y*(A26*h6y + A66*(-h6*s*sina + h6x)))
            k0[pos+523] += detJ*wi*(h4*s*sina*(A22*h6y + A26*(-h6*s*sina + h6x)) + h4x*(A12*h6y + A16*(-h6*s*sina + h6x)) + h4y*(A26*h6y + A66*(-h6*s*sina + h6x)))
            k0[pos+524] += detJ*wi*(h5*s*sina*(A22*h6y + A26*(-h6*s*sina + h6x)) + h5x*(A12*h6y + A16*(-h6*s*sina + h6x)) + h5y*(A26*h6y + A66*(-h6*s*sina + h6x)))
            k0[pos+525] += detJ*wi*(h6*s*sina*(A22*h6y + A26*(-h6*s*sina + h6x)) + h6x*(A12*h6y + A16*(-h6*s*sina + h6x)) + h6y*(A26*h6y + A66*(-h6*s*sina + h6x)))
            k0[pos+526] += detJ*wi*(h7*s*sina*(A22*h6y + A26*(-h6*s*sina + h6x)) + h7x*(A12*h6y + A16*(-h6*s*sina + h6x)) + h7y*(A26*h6y + A66*(-h6*s*sina + h6x)))
            k0[pos+527] += detJ*wi*(h8*s*sina*(A22*h6y + A26*(-h6*s*sina + h6x)) + h8x*(A12*h6y + A16*(-h6*s*sina + h6x)) + h8y*(A26*h6y + A66*(-h6*s*sina + h6x)))
            k0[pos+528] += detJ*wi*(A44*(cosa*cosa)*h1*h6*(s*s) + h1y*(A22*h6y + A26*(-h6*s*sina + h6x)) + (A26*h6y + A66*(-h6*s*sina + h6x))*(-h1*s*sina + h1x))
            k0[pos+529] += detJ*wi*(A44*(cosa*cosa)*h2*h6*(s*s) + h2y*(A22*h6y + A26*(-h6*s*sina + h6x)) + (A26*h6y + A66*(-h6*s*sina + h6x))*(-h2*s*sina + h2x))
            k0[pos+530] += detJ*wi*(A44*(cosa*cosa)*h3*h6*(s*s) + h3y*(A22*h6y + A26*(-h6*s*sina + h6x)) + (A26*h6y + A66*(-h6*s*sina + h6x))*(-h3*s*sina + h3x))
            k0[pos+531] += detJ*wi*(A44*(cosa*cosa)*h4*h6*(s*s) + h4y*(A22*h6y + A26*(-h6*s*sina + h6x)) + (A26*h6y + A66*(-h6*s*sina + h6x))*(-h4*s*sina + h4x))
            k0[pos+532] += detJ*wi*(A44*(cosa*cosa)*h5*h6*(s*s) + h5y*(A22*h6y + A26*(-h6*s*sina + h6x)) + (A26*h6y + A66*(-h6*s*sina + h6x))*(-h5*s*sina + h5x))
            k0[pos+533] += detJ*wi*(A22*(h6y*h6y) + 2*A26*h6y*(-h6*s*sina + h6x) + A44*(cosa*cosa)*(h6*h6)*(s*s) + A66*(-h6*s*sina + h6x)**2)
            k0[pos+534] += detJ*wi*(A44*(cosa*cosa)*h6*h7*(s*s) + h7y*(A22*h6y + A26*(-h6*s*sina + h6x)) + (A26*h6y + A66*(-h6*s*sina + h6x))*(-h7*s*sina + h7x))
            k0[pos+535] += detJ*wi*(A44*(cosa*cosa)*h6*h8*(s*s) + h8y*(A22*h6y + A26*(-h6*s*sina + h6x)) + (A26*h6y + A66*(-h6*s*sina + h6x))*(-h8*s*sina + h8x))
            k0[pos+536] += -cosa*detJ*s*wi*(A44*h1y*h6 + A45*h1x*h6 - h1*(A22*h6y - A26*h6*s*sina + A26*h6x))
            k0[pos+537] += -cosa*detJ*s*wi*(A44*h2y*h6 + A45*h2x*h6 - h2*(A22*h6y - A26*h6*s*sina + A26*h6x))
            k0[pos+538] += -cosa*detJ*s*wi*(A44*h3y*h6 + A45*h3x*h6 - h3*(A22*h6y - A26*h6*s*sina + A26*h6x))
            k0[pos+539] += -cosa*detJ*s*wi*(A44*h4y*h6 + A45*h4x*h6 - h4*(A22*h6y - A26*h6*s*sina + A26*h6x))
            k0[pos+540] += -cosa*detJ*s*wi*(A44*h5y*h6 + A45*h5x*h6 - h5*(A22*h6y - A26*h6*s*sina + A26*h6x))
            k0[pos+541] += -cosa*detJ*h6*s*wi*(-A22*h6y + A26*h6*s*sina - A26*h6x + A44*h6y + A45*h6x)
            k0[pos+542] += -cosa*detJ*s*wi*(-A22*h6y*h7 + A26*h6*h7*s*sina - A26*h6x*h7 + A44*h6*h7y + A45*h6*h7x)
            k0[pos+543] += -cosa*detJ*s*wi*(-A22*h6y*h8 + A26*h6*h8*s*sina - A26*h6x*h8 + A44*h6*h8y + A45*h6*h8x)
            k0[pos+544] += detJ*wi*(-A45*cosa*h1*h6*s + h1*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h1x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h1y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+545] += detJ*wi*(-A45*cosa*h2*h6*s + h2*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h2x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h2y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+546] += detJ*wi*(-A45*cosa*h3*h6*s + h3*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h3x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h3y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+547] += detJ*wi*(-A45*cosa*h4*h6*s + h4*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h4x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h4y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+548] += detJ*wi*(-A45*cosa*h5*h6*s + h5*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h5x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h5y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+549] += detJ*wi*(-A45*cosa*(h6*h6)*s + h6*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h6x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h6y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+550] += detJ*wi*(-A45*cosa*h6*h7*s + h7*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h7x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h7y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+551] += detJ*wi*(-A45*cosa*h6*h8*s + h8*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h8x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h8y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+552] += detJ*wi*(-A44*cosa*h1*h6*s + h1y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h1*s*sina + h1x))
            k0[pos+553] += detJ*wi*(-A44*cosa*h2*h6*s + h2y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h2*s*sina + h2x))
            k0[pos+554] += detJ*wi*(-A44*cosa*h3*h6*s + h3y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h3*s*sina + h3x))
            k0[pos+555] += detJ*wi*(-A44*cosa*h4*h6*s + h4y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h4*s*sina + h4x))
            k0[pos+556] += detJ*wi*(-A44*cosa*h5*h6*s + h5y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h5*s*sina + h5x))
            k0[pos+557] += detJ*wi*(-A44*cosa*(h6*h6)*s + B22*(h6y*h6y) + 2*B26*h6y*(-h6*s*sina + h6x) + B66*(-h6*s*sina + h6x)**2)
            k0[pos+558] += detJ*wi*(-A44*cosa*h6*h7*s + h7y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h7*s*sina + h7x))
            k0[pos+559] += detJ*wi*(-A44*cosa*h6*h8*s + h8y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h8*s*sina + h8x))
            k0[pos+560] += detJ*wi*(h1*s*sina*(A22*h7y + A26*(-h7*s*sina + h7x)) + h1x*(A12*h7y + A16*(-h7*s*sina + h7x)) + h1y*(A26*h7y + A66*(-h7*s*sina + h7x)))
            k0[pos+561] += detJ*wi*(h2*s*sina*(A22*h7y + A26*(-h7*s*sina + h7x)) + h2x*(A12*h7y + A16*(-h7*s*sina + h7x)) + h2y*(A26*h7y + A66*(-h7*s*sina + h7x)))
            k0[pos+562] += detJ*wi*(h3*s*sina*(A22*h7y + A26*(-h7*s*sina + h7x)) + h3x*(A12*h7y + A16*(-h7*s*sina + h7x)) + h3y*(A26*h7y + A66*(-h7*s*sina + h7x)))
            k0[pos+563] += detJ*wi*(h4*s*sina*(A22*h7y + A26*(-h7*s*sina + h7x)) + h4x*(A12*h7y + A16*(-h7*s*sina + h7x)) + h4y*(A26*h7y + A66*(-h7*s*sina + h7x)))
            k0[pos+564] += detJ*wi*(h5*s*sina*(A22*h7y + A26*(-h7*s*sina + h7x)) + h5x*(A12*h7y + A16*(-h7*s*sina + h7x)) + h5y*(A26*h7y + A66*(-h7*s*sina + h7x)))
            k0[pos+565] += detJ*wi*(h6*s*sina*(A22*h7y + A26*(-h7*s*sina + h7x)) + h6x*(A12*h7y + A16*(-h7*s*sina + h7x)) + h6y*(A26*h7y + A66*(-h7*s*sina + h7x)))
            k0[pos+566] += detJ*wi*(h7*s*sina*(A22*h7y + A26*(-h7*s*sina + h7x)) + h7x*(A12*h7y + A16*(-h7*s*sina + h7x)) + h7y*(A26*h7y + A66*(-h7*s*sina + h7x)))
            k0[pos+567] += detJ*wi*(h8*s*sina*(A22*h7y + A26*(-h7*s*sina + h7x)) + h8x*(A12*h7y + A16*(-h7*s*sina + h7x)) + h8y*(A26*h7y + A66*(-h7*s*sina + h7x)))
            k0[pos+568] += detJ*wi*(A44*(cosa*cosa)*h1*h7*(s*s) + h1y*(A22*h7y + A26*(-h7*s*sina + h7x)) + (A26*h7y + A66*(-h7*s*sina + h7x))*(-h1*s*sina + h1x))
            k0[pos+569] += detJ*wi*(A44*(cosa*cosa)*h2*h7*(s*s) + h2y*(A22*h7y + A26*(-h7*s*sina + h7x)) + (A26*h7y + A66*(-h7*s*sina + h7x))*(-h2*s*sina + h2x))
            k0[pos+570] += detJ*wi*(A44*(cosa*cosa)*h3*h7*(s*s) + h3y*(A22*h7y + A26*(-h7*s*sina + h7x)) + (A26*h7y + A66*(-h7*s*sina + h7x))*(-h3*s*sina + h3x))
            k0[pos+571] += detJ*wi*(A44*(cosa*cosa)*h4*h7*(s*s) + h4y*(A22*h7y + A26*(-h7*s*sina + h7x)) + (A26*h7y + A66*(-h7*s*sina + h7x))*(-h4*s*sina + h4x))
            k0[pos+572] += detJ*wi*(A44*(cosa*cosa)*h5*h7*(s*s) + h5y*(A22*h7y + A26*(-h7*s*sina + h7x)) + (A26*h7y + A66*(-h7*s*sina + h7x))*(-h5*s*sina + h5x))
            k0[pos+573] += detJ*wi*(A44*(cosa*cosa)*h6*h7*(s*s) + h6y*(A22*h7y + A26*(-h7*s*sina + h7x)) + (A26*h7y + A66*(-h7*s*sina + h7x))*(-h6*s*sina + h6x))
            k0[pos+574] += detJ*wi*(A22*(h7y*h7y) + 2*A26*h7y*(-h7*s*sina + h7x) + A44*(cosa*cosa)*(h7*h7)*(s*s) + A66*(-h7*s*sina + h7x)**2)
            k0[pos+575] += detJ*wi*(A44*(cosa*cosa)*h7*h8*(s*s) + h8y*(A22*h7y + A26*(-h7*s*sina + h7x)) + (A26*h7y + A66*(-h7*s*sina + h7x))*(-h8*s*sina + h8x))
            k0[pos+576] += -cosa*detJ*s*wi*(A44*h1y*h7 + A45*h1x*h7 - h1*(A22*h7y - A26*h7*s*sina + A26*h7x))
            k0[pos+577] += -cosa*detJ*s*wi*(A44*h2y*h7 + A45*h2x*h7 - h2*(A22*h7y - A26*h7*s*sina + A26*h7x))
            k0[pos+578] += -cosa*detJ*s*wi*(A44*h3y*h7 + A45*h3x*h7 - h3*(A22*h7y - A26*h7*s*sina + A26*h7x))
            k0[pos+579] += -cosa*detJ*s*wi*(A44*h4y*h7 + A45*h4x*h7 - h4*(A22*h7y - A26*h7*s*sina + A26*h7x))
            k0[pos+580] += -cosa*detJ*s*wi*(A44*h5y*h7 + A45*h5x*h7 - h5*(A22*h7y - A26*h7*s*sina + A26*h7x))
            k0[pos+581] += -cosa*detJ*s*wi*(A44*h6y*h7 + A45*h6x*h7 - h6*(A22*h7y - A26*h7*s*sina + A26*h7x))
            k0[pos+582] += -cosa*detJ*h7*s*wi*(-A22*h7y + A26*h7*s*sina - A26*h7x + A44*h7y + A45*h7x)
            k0[pos+583] += -cosa*detJ*s*wi*(-A22*h7y*h8 + A26*h7*h8*s*sina - A26*h7x*h8 + A44*h7*h8y + A45*h7*h8x)
            k0[pos+584] += detJ*wi*(-A45*cosa*h1*h7*s + h1*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h1x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h1y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+585] += detJ*wi*(-A45*cosa*h2*h7*s + h2*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h2x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h2y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+586] += detJ*wi*(-A45*cosa*h3*h7*s + h3*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h3x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h3y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+587] += detJ*wi*(-A45*cosa*h4*h7*s + h4*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h4x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h4y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+588] += detJ*wi*(-A45*cosa*h5*h7*s + h5*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h5x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h5y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+589] += detJ*wi*(-A45*cosa*h6*h7*s + h6*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h6x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h6y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+590] += detJ*wi*(-A45*cosa*(h7*h7)*s + h7*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h7x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h7y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+591] += detJ*wi*(-A45*cosa*h7*h8*s + h8*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h8x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h8y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+592] += detJ*wi*(-A44*cosa*h1*h7*s + h1y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h1*s*sina + h1x))
            k0[pos+593] += detJ*wi*(-A44*cosa*h2*h7*s + h2y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h2*s*sina + h2x))
            k0[pos+594] += detJ*wi*(-A44*cosa*h3*h7*s + h3y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h3*s*sina + h3x))
            k0[pos+595] += detJ*wi*(-A44*cosa*h4*h7*s + h4y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h4*s*sina + h4x))
            k0[pos+596] += detJ*wi*(-A44*cosa*h5*h7*s + h5y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h5*s*sina + h5x))
            k0[pos+597] += detJ*wi*(-A44*cosa*h6*h7*s + h6y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h6*s*sina + h6x))
            k0[pos+598] += detJ*wi*(-A44*cosa*(h7*h7)*s + B22*(h7y*h7y) + 2*B26*h7y*(-h7*s*sina + h7x) + B66*(-h7*s*sina + h7x)**2)
            k0[pos+599] += detJ*wi*(-A44*cosa*h7*h8*s + h8y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h8*s*sina + h8x))
            k0[pos+600] += detJ*wi*(h1*s*sina*(A22*h8y + A26*(-h8*s*sina + h8x)) + h1x*(A12*h8y + A16*(-h8*s*sina + h8x)) + h1y*(A26*h8y + A66*(-h8*s*sina + h8x)))
            k0[pos+601] += detJ*wi*(h2*s*sina*(A22*h8y + A26*(-h8*s*sina + h8x)) + h2x*(A12*h8y + A16*(-h8*s*sina + h8x)) + h2y*(A26*h8y + A66*(-h8*s*sina + h8x)))
            k0[pos+602] += detJ*wi*(h3*s*sina*(A22*h8y + A26*(-h8*s*sina + h8x)) + h3x*(A12*h8y + A16*(-h8*s*sina + h8x)) + h3y*(A26*h8y + A66*(-h8*s*sina + h8x)))
            k0[pos+603] += detJ*wi*(h4*s*sina*(A22*h8y + A26*(-h8*s*sina + h8x)) + h4x*(A12*h8y + A16*(-h8*s*sina + h8x)) + h4y*(A26*h8y + A66*(-h8*s*sina + h8x)))
            k0[pos+604] += detJ*wi*(h5*s*sina*(A22*h8y + A26*(-h8*s*sina + h8x)) + h5x*(A12*h8y + A16*(-h8*s*sina + h8x)) + h5y*(A26*h8y + A66*(-h8*s*sina + h8x)))
            k0[pos+605] += detJ*wi*(h6*s*sina*(A22*h8y + A26*(-h8*s*sina + h8x)) + h6x*(A12*h8y + A16*(-h8*s*sina + h8x)) + h6y*(A26*h8y + A66*(-h8*s*sina + h8x)))
            k0[pos+606] += detJ*wi*(h7*s*sina*(A22*h8y + A26*(-h8*s*sina + h8x)) + h7x*(A12*h8y + A16*(-h8*s*sina + h8x)) + h7y*(A26*h8y + A66*(-h8*s*sina + h8x)))
            k0[pos+607] += detJ*wi*(h8*s*sina*(A22*h8y + A26*(-h8*s*sina + h8x)) + h8x*(A12*h8y + A16*(-h8*s*sina + h8x)) + h8y*(A26*h8y + A66*(-h8*s*sina + h8x)))
            k0[pos+608] += detJ*wi*(A44*(cosa*cosa)*h1*h8*(s*s) + h1y*(A22*h8y + A26*(-h8*s*sina + h8x)) + (A26*h8y + A66*(-h8*s*sina + h8x))*(-h1*s*sina + h1x))
            k0[pos+609] += detJ*wi*(A44*(cosa*cosa)*h2*h8*(s*s) + h2y*(A22*h8y + A26*(-h8*s*sina + h8x)) + (A26*h8y + A66*(-h8*s*sina + h8x))*(-h2*s*sina + h2x))
            k0[pos+610] += detJ*wi*(A44*(cosa*cosa)*h3*h8*(s*s) + h3y*(A22*h8y + A26*(-h8*s*sina + h8x)) + (A26*h8y + A66*(-h8*s*sina + h8x))*(-h3*s*sina + h3x))
            k0[pos+611] += detJ*wi*(A44*(cosa*cosa)*h4*h8*(s*s) + h4y*(A22*h8y + A26*(-h8*s*sina + h8x)) + (A26*h8y + A66*(-h8*s*sina + h8x))*(-h4*s*sina + h4x))
            k0[pos+612] += detJ*wi*(A44*(cosa*cosa)*h5*h8*(s*s) + h5y*(A22*h8y + A26*(-h8*s*sina + h8x)) + (A26*h8y + A66*(-h8*s*sina + h8x))*(-h5*s*sina + h5x))
            k0[pos+613] += detJ*wi*(A44*(cosa*cosa)*h6*h8*(s*s) + h6y*(A22*h8y + A26*(-h8*s*sina + h8x)) + (A26*h8y + A66*(-h8*s*sina + h8x))*(-h6*s*sina + h6x))
            k0[pos+614] += detJ*wi*(A44*(cosa*cosa)*h7*h8*(s*s) + h7y*(A22*h8y + A26*(-h8*s*sina + h8x)) + (A26*h8y + A66*(-h8*s*sina + h8x))*(-h7*s*sina + h7x))
            k0[pos+615] += detJ*wi*(A22*(h8y*h8y) + 2*A26*h8y*(-h8*s*sina + h8x) + A44*(cosa*cosa)*(h8*h8)*(s*s) + A66*(-h8*s*sina + h8x)**2)
            k0[pos+616] += -cosa*detJ*s*wi*(A44*h1y*h8 + A45*h1x*h8 - h1*(A22*h8y - A26*h8*s*sina + A26*h8x))
            k0[pos+617] += -cosa*detJ*s*wi*(A44*h2y*h8 + A45*h2x*h8 - h2*(A22*h8y - A26*h8*s*sina + A26*h8x))
            k0[pos+618] += -cosa*detJ*s*wi*(A44*h3y*h8 + A45*h3x*h8 - h3*(A22*h8y - A26*h8*s*sina + A26*h8x))
            k0[pos+619] += -cosa*detJ*s*wi*(A44*h4y*h8 + A45*h4x*h8 - h4*(A22*h8y - A26*h8*s*sina + A26*h8x))
            k0[pos+620] += -cosa*detJ*s*wi*(A44*h5y*h8 + A45*h5x*h8 - h5*(A22*h8y - A26*h8*s*sina + A26*h8x))
            k0[pos+621] += -cosa*detJ*s*wi*(A44*h6y*h8 + A45*h6x*h8 - h6*(A22*h8y - A26*h8*s*sina + A26*h8x))
            k0[pos+622] += -cosa*detJ*s*wi*(A44*h7y*h8 + A45*h7x*h8 - h7*(A22*h8y - A26*h8*s*sina + A26*h8x))
            k0[pos+623] += -cosa*detJ*h8*s*wi*(-A22*h8y + A26*h8*s*sina - A26*h8x + A44*h8y + A45*h8x)
            k0[pos+624] += detJ*wi*(-A45*cosa*h1*h8*s + h1*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h1x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h1y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+625] += detJ*wi*(-A45*cosa*h2*h8*s + h2*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h2x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h2y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+626] += detJ*wi*(-A45*cosa*h3*h8*s + h3*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h3x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h3y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+627] += detJ*wi*(-A45*cosa*h4*h8*s + h4*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h4x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h4y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+628] += detJ*wi*(-A45*cosa*h5*h8*s + h5*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h5x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h5y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+629] += detJ*wi*(-A45*cosa*h6*h8*s + h6*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h6x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h6y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+630] += detJ*wi*(-A45*cosa*h7*h8*s + h7*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h7x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h7y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+631] += detJ*wi*(-A45*cosa*(h8*h8)*s + h8*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h8x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h8y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+632] += detJ*wi*(-A44*cosa*h1*h8*s + h1y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h1*s*sina + h1x))
            k0[pos+633] += detJ*wi*(-A44*cosa*h2*h8*s + h2y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h2*s*sina + h2x))
            k0[pos+634] += detJ*wi*(-A44*cosa*h3*h8*s + h3y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h3*s*sina + h3x))
            k0[pos+635] += detJ*wi*(-A44*cosa*h4*h8*s + h4y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h4*s*sina + h4x))
            k0[pos+636] += detJ*wi*(-A44*cosa*h5*h8*s + h5y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h5*s*sina + h5x))
            k0[pos+637] += detJ*wi*(-A44*cosa*h6*h8*s + h6y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h6*s*sina + h6x))
            k0[pos+638] += detJ*wi*(-A44*cosa*h7*h8*s + h7y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h7*s*sina + h7x))
            k0[pos+639] += detJ*wi*(-A44*cosa*(h8*h8)*s + B22*(h8y*h8y) + 2*B26*h8y*(-h8*s*sina + h8x) + B66*(-h8*s*sina + h8x)**2)
            k0[pos+640] += cosa*detJ*h1*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+641] += cosa*detJ*h1*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+642] += cosa*detJ*h1*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+643] += cosa*detJ*h1*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+644] += cosa*detJ*h1*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+645] += cosa*detJ*h1*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+646] += cosa*detJ*h1*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+647] += cosa*detJ*h1*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+648] += -cosa*detJ*h1*s*wi*(-A22*h1y + A26*h1*s*sina - A26*h1x + A44*h1y + A45*h1x)
            k0[pos+649] += cosa*detJ*s*wi*(A22*h1*h2y + A26*h1*(-h2*s*sina + h2x) - h2*(A44*h1y + A45*h1x))
            k0[pos+650] += cosa*detJ*s*wi*(A22*h1*h3y + A26*h1*(-h3*s*sina + h3x) - h3*(A44*h1y + A45*h1x))
            k0[pos+651] += cosa*detJ*s*wi*(A22*h1*h4y + A26*h1*(-h4*s*sina + h4x) - h4*(A44*h1y + A45*h1x))
            k0[pos+652] += cosa*detJ*s*wi*(A22*h1*h5y + A26*h1*(-h5*s*sina + h5x) - h5*(A44*h1y + A45*h1x))
            k0[pos+653] += cosa*detJ*s*wi*(A22*h1*h6y + A26*h1*(-h6*s*sina + h6x) - h6*(A44*h1y + A45*h1x))
            k0[pos+654] += cosa*detJ*s*wi*(A22*h1*h7y + A26*h1*(-h7*s*sina + h7x) - h7*(A44*h1y + A45*h1x))
            k0[pos+655] += cosa*detJ*s*wi*(A22*h1*h8y + A26*h1*(-h8*s*sina + h8x) - h8*(A44*h1y + A45*h1x))
            k0[pos+656] += detJ*wi*(A22*(cosa*cosa)*(h1*h1)*(s*s) + A44*(h1y*h1y) + 2*A45*h1x*h1y + A55*(h1x*h1x))
            k0[pos+657] += detJ*wi*(A22*(cosa*cosa)*h1*h2*(s*s) + A44*h1y*h2y + A45*h1x*h2y + A45*h1y*h2x + A55*h1x*h2x)
            k0[pos+658] += detJ*wi*(A22*(cosa*cosa)*h1*h3*(s*s) + A44*h1y*h3y + A45*h1x*h3y + A45*h1y*h3x + A55*h1x*h3x)
            k0[pos+659] += detJ*wi*(A22*(cosa*cosa)*h1*h4*(s*s) + A44*h1y*h4y + A45*h1x*h4y + A45*h1y*h4x + A55*h1x*h4x)
            k0[pos+660] += detJ*wi*(A22*(cosa*cosa)*h1*h5*(s*s) + A44*h1y*h5y + A45*h1x*h5y + A45*h1y*h5x + A55*h1x*h5x)
            k0[pos+661] += detJ*wi*(A22*(cosa*cosa)*h1*h6*(s*s) + A44*h1y*h6y + A45*h1x*h6y + A45*h1y*h6x + A55*h1x*h6x)
            k0[pos+662] += detJ*wi*(A22*(cosa*cosa)*h1*h7*(s*s) + A44*h1y*h7y + A45*h1x*h7y + A45*h1y*h7x + A55*h1x*h7x)
            k0[pos+663] += detJ*wi*(A22*(cosa*cosa)*h1*h8*(s*s) + A44*h1y*h8y + A45*h1x*h8y + A45*h1y*h8x + A55*h1x*h8x)
            k0[pos+664] += detJ*h1*wi*(A45*h1y + A55*h1x + cosa*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+665] += detJ*wi*(A45*h1y*h2 + A55*h1x*h2 + cosa*h1*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+666] += detJ*wi*(A45*h1y*h3 + A55*h1x*h3 + cosa*h1*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+667] += detJ*wi*(A45*h1y*h4 + A55*h1x*h4 + cosa*h1*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+668] += detJ*wi*(A45*h1y*h5 + A55*h1x*h5 + cosa*h1*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+669] += detJ*wi*(A45*h1y*h6 + A55*h1x*h6 + cosa*h1*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+670] += detJ*wi*(A45*h1y*h7 + A55*h1x*h7 + cosa*h1*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+671] += detJ*wi*(A45*h1y*h8 + A55*h1x*h8 + cosa*h1*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+672] += detJ*h1*wi*(A44*h1y + A45*h1x + cosa*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+673] += detJ*wi*(A44*h1y*h2 + A45*h1x*h2 + cosa*h1*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+674] += detJ*wi*(A44*h1y*h3 + A45*h1x*h3 + cosa*h1*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+675] += detJ*wi*(A44*h1y*h4 + A45*h1x*h4 + cosa*h1*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+676] += detJ*wi*(A44*h1y*h5 + A45*h1x*h5 + cosa*h1*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+677] += detJ*wi*(A44*h1y*h6 + A45*h1x*h6 + cosa*h1*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+678] += detJ*wi*(A44*h1y*h7 + A45*h1x*h7 + cosa*h1*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+679] += detJ*wi*(A44*h1y*h8 + A45*h1x*h8 + cosa*h1*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+680] += cosa*detJ*h2*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+681] += cosa*detJ*h2*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+682] += cosa*detJ*h2*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+683] += cosa*detJ*h2*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+684] += cosa*detJ*h2*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+685] += cosa*detJ*h2*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+686] += cosa*detJ*h2*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+687] += cosa*detJ*h2*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+688] += cosa*detJ*s*wi*(A22*h1y*h2 + A26*h2*(-h1*s*sina + h1x) - h1*(A44*h2y + A45*h2x))
            k0[pos+689] += -cosa*detJ*h2*s*wi*(-A22*h2y + A26*h2*s*sina - A26*h2x + A44*h2y + A45*h2x)
            k0[pos+690] += cosa*detJ*s*wi*(A22*h2*h3y + A26*h2*(-h3*s*sina + h3x) - h3*(A44*h2y + A45*h2x))
            k0[pos+691] += cosa*detJ*s*wi*(A22*h2*h4y + A26*h2*(-h4*s*sina + h4x) - h4*(A44*h2y + A45*h2x))
            k0[pos+692] += cosa*detJ*s*wi*(A22*h2*h5y + A26*h2*(-h5*s*sina + h5x) - h5*(A44*h2y + A45*h2x))
            k0[pos+693] += cosa*detJ*s*wi*(A22*h2*h6y + A26*h2*(-h6*s*sina + h6x) - h6*(A44*h2y + A45*h2x))
            k0[pos+694] += cosa*detJ*s*wi*(A22*h2*h7y + A26*h2*(-h7*s*sina + h7x) - h7*(A44*h2y + A45*h2x))
            k0[pos+695] += cosa*detJ*s*wi*(A22*h2*h8y + A26*h2*(-h8*s*sina + h8x) - h8*(A44*h2y + A45*h2x))
            k0[pos+696] += detJ*wi*(A22*(cosa*cosa)*h1*h2*(s*s) + A44*h1y*h2y + A45*h1x*h2y + A45*h1y*h2x + A55*h1x*h2x)
            k0[pos+697] += detJ*wi*(A22*(cosa*cosa)*(h2*h2)*(s*s) + A44*(h2y*h2y) + 2*A45*h2x*h2y + A55*(h2x*h2x))
            k0[pos+698] += detJ*wi*(A22*(cosa*cosa)*h2*h3*(s*s) + A44*h2y*h3y + A45*h2x*h3y + A45*h2y*h3x + A55*h2x*h3x)
            k0[pos+699] += detJ*wi*(A22*(cosa*cosa)*h2*h4*(s*s) + A44*h2y*h4y + A45*h2x*h4y + A45*h2y*h4x + A55*h2x*h4x)
            k0[pos+700] += detJ*wi*(A22*(cosa*cosa)*h2*h5*(s*s) + A44*h2y*h5y + A45*h2x*h5y + A45*h2y*h5x + A55*h2x*h5x)
            k0[pos+701] += detJ*wi*(A22*(cosa*cosa)*h2*h6*(s*s) + A44*h2y*h6y + A45*h2x*h6y + A45*h2y*h6x + A55*h2x*h6x)
            k0[pos+702] += detJ*wi*(A22*(cosa*cosa)*h2*h7*(s*s) + A44*h2y*h7y + A45*h2x*h7y + A45*h2y*h7x + A55*h2x*h7x)
            k0[pos+703] += detJ*wi*(A22*(cosa*cosa)*h2*h8*(s*s) + A44*h2y*h8y + A45*h2x*h8y + A45*h2y*h8x + A55*h2x*h8x)
            k0[pos+704] += detJ*wi*(A45*h1*h2y + A55*h1*h2x + cosa*h2*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+705] += detJ*h2*wi*(A45*h2y + A55*h2x + cosa*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+706] += detJ*wi*(A45*h2y*h3 + A55*h2x*h3 + cosa*h2*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+707] += detJ*wi*(A45*h2y*h4 + A55*h2x*h4 + cosa*h2*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+708] += detJ*wi*(A45*h2y*h5 + A55*h2x*h5 + cosa*h2*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+709] += detJ*wi*(A45*h2y*h6 + A55*h2x*h6 + cosa*h2*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+710] += detJ*wi*(A45*h2y*h7 + A55*h2x*h7 + cosa*h2*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+711] += detJ*wi*(A45*h2y*h8 + A55*h2x*h8 + cosa*h2*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+712] += detJ*wi*(A44*h1*h2y + A45*h1*h2x + cosa*h2*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+713] += detJ*h2*wi*(A44*h2y + A45*h2x + cosa*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+714] += detJ*wi*(A44*h2y*h3 + A45*h2x*h3 + cosa*h2*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+715] += detJ*wi*(A44*h2y*h4 + A45*h2x*h4 + cosa*h2*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+716] += detJ*wi*(A44*h2y*h5 + A45*h2x*h5 + cosa*h2*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+717] += detJ*wi*(A44*h2y*h6 + A45*h2x*h6 + cosa*h2*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+718] += detJ*wi*(A44*h2y*h7 + A45*h2x*h7 + cosa*h2*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+719] += detJ*wi*(A44*h2y*h8 + A45*h2x*h8 + cosa*h2*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+720] += cosa*detJ*h3*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+721] += cosa*detJ*h3*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+722] += cosa*detJ*h3*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+723] += cosa*detJ*h3*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+724] += cosa*detJ*h3*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+725] += cosa*detJ*h3*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+726] += cosa*detJ*h3*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+727] += cosa*detJ*h3*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+728] += cosa*detJ*s*wi*(A22*h1y*h3 + A26*h3*(-h1*s*sina + h1x) - h1*(A44*h3y + A45*h3x))
            k0[pos+729] += cosa*detJ*s*wi*(A22*h2y*h3 + A26*h3*(-h2*s*sina + h2x) - h2*(A44*h3y + A45*h3x))
            k0[pos+730] += -cosa*detJ*h3*s*wi*(-A22*h3y + A26*h3*s*sina - A26*h3x + A44*h3y + A45*h3x)
            k0[pos+731] += cosa*detJ*s*wi*(A22*h3*h4y + A26*h3*(-h4*s*sina + h4x) - h4*(A44*h3y + A45*h3x))
            k0[pos+732] += cosa*detJ*s*wi*(A22*h3*h5y + A26*h3*(-h5*s*sina + h5x) - h5*(A44*h3y + A45*h3x))
            k0[pos+733] += cosa*detJ*s*wi*(A22*h3*h6y + A26*h3*(-h6*s*sina + h6x) - h6*(A44*h3y + A45*h3x))
            k0[pos+734] += cosa*detJ*s*wi*(A22*h3*h7y + A26*h3*(-h7*s*sina + h7x) - h7*(A44*h3y + A45*h3x))
            k0[pos+735] += cosa*detJ*s*wi*(A22*h3*h8y + A26*h3*(-h8*s*sina + h8x) - h8*(A44*h3y + A45*h3x))
            k0[pos+736] += detJ*wi*(A22*(cosa*cosa)*h1*h3*(s*s) + A44*h1y*h3y + A45*h1x*h3y + A45*h1y*h3x + A55*h1x*h3x)
            k0[pos+737] += detJ*wi*(A22*(cosa*cosa)*h2*h3*(s*s) + A44*h2y*h3y + A45*h2x*h3y + A45*h2y*h3x + A55*h2x*h3x)
            k0[pos+738] += detJ*wi*(A22*(cosa*cosa)*(h3*h3)*(s*s) + A44*(h3y*h3y) + 2*A45*h3x*h3y + A55*(h3x*h3x))
            k0[pos+739] += detJ*wi*(A22*(cosa*cosa)*h3*h4*(s*s) + A44*h3y*h4y + A45*h3x*h4y + A45*h3y*h4x + A55*h3x*h4x)
            k0[pos+740] += detJ*wi*(A22*(cosa*cosa)*h3*h5*(s*s) + A44*h3y*h5y + A45*h3x*h5y + A45*h3y*h5x + A55*h3x*h5x)
            k0[pos+741] += detJ*wi*(A22*(cosa*cosa)*h3*h6*(s*s) + A44*h3y*h6y + A45*h3x*h6y + A45*h3y*h6x + A55*h3x*h6x)
            k0[pos+742] += detJ*wi*(A22*(cosa*cosa)*h3*h7*(s*s) + A44*h3y*h7y + A45*h3x*h7y + A45*h3y*h7x + A55*h3x*h7x)
            k0[pos+743] += detJ*wi*(A22*(cosa*cosa)*h3*h8*(s*s) + A44*h3y*h8y + A45*h3x*h8y + A45*h3y*h8x + A55*h3x*h8x)
            k0[pos+744] += detJ*wi*(A45*h1*h3y + A55*h1*h3x + cosa*h3*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+745] += detJ*wi*(A45*h2*h3y + A55*h2*h3x + cosa*h3*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+746] += detJ*h3*wi*(A45*h3y + A55*h3x + cosa*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+747] += detJ*wi*(A45*h3y*h4 + A55*h3x*h4 + cosa*h3*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+748] += detJ*wi*(A45*h3y*h5 + A55*h3x*h5 + cosa*h3*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+749] += detJ*wi*(A45*h3y*h6 + A55*h3x*h6 + cosa*h3*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+750] += detJ*wi*(A45*h3y*h7 + A55*h3x*h7 + cosa*h3*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+751] += detJ*wi*(A45*h3y*h8 + A55*h3x*h8 + cosa*h3*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+752] += detJ*wi*(A44*h1*h3y + A45*h1*h3x + cosa*h3*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+753] += detJ*wi*(A44*h2*h3y + A45*h2*h3x + cosa*h3*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+754] += detJ*h3*wi*(A44*h3y + A45*h3x + cosa*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+755] += detJ*wi*(A44*h3y*h4 + A45*h3x*h4 + cosa*h3*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+756] += detJ*wi*(A44*h3y*h5 + A45*h3x*h5 + cosa*h3*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+757] += detJ*wi*(A44*h3y*h6 + A45*h3x*h6 + cosa*h3*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+758] += detJ*wi*(A44*h3y*h7 + A45*h3x*h7 + cosa*h3*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+759] += detJ*wi*(A44*h3y*h8 + A45*h3x*h8 + cosa*h3*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+760] += cosa*detJ*h4*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+761] += cosa*detJ*h4*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+762] += cosa*detJ*h4*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+763] += cosa*detJ*h4*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+764] += cosa*detJ*h4*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+765] += cosa*detJ*h4*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+766] += cosa*detJ*h4*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+767] += cosa*detJ*h4*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+768] += cosa*detJ*s*wi*(A22*h1y*h4 + A26*h4*(-h1*s*sina + h1x) - h1*(A44*h4y + A45*h4x))
            k0[pos+769] += cosa*detJ*s*wi*(A22*h2y*h4 + A26*h4*(-h2*s*sina + h2x) - h2*(A44*h4y + A45*h4x))
            k0[pos+770] += cosa*detJ*s*wi*(A22*h3y*h4 + A26*h4*(-h3*s*sina + h3x) - h3*(A44*h4y + A45*h4x))
            k0[pos+771] += -cosa*detJ*h4*s*wi*(-A22*h4y + A26*h4*s*sina - A26*h4x + A44*h4y + A45*h4x)
            k0[pos+772] += cosa*detJ*s*wi*(A22*h4*h5y + A26*h4*(-h5*s*sina + h5x) - h5*(A44*h4y + A45*h4x))
            k0[pos+773] += cosa*detJ*s*wi*(A22*h4*h6y + A26*h4*(-h6*s*sina + h6x) - h6*(A44*h4y + A45*h4x))
            k0[pos+774] += cosa*detJ*s*wi*(A22*h4*h7y + A26*h4*(-h7*s*sina + h7x) - h7*(A44*h4y + A45*h4x))
            k0[pos+775] += cosa*detJ*s*wi*(A22*h4*h8y + A26*h4*(-h8*s*sina + h8x) - h8*(A44*h4y + A45*h4x))
            k0[pos+776] += detJ*wi*(A22*(cosa*cosa)*h1*h4*(s*s) + A44*h1y*h4y + A45*h1x*h4y + A45*h1y*h4x + A55*h1x*h4x)
            k0[pos+777] += detJ*wi*(A22*(cosa*cosa)*h2*h4*(s*s) + A44*h2y*h4y + A45*h2x*h4y + A45*h2y*h4x + A55*h2x*h4x)
            k0[pos+778] += detJ*wi*(A22*(cosa*cosa)*h3*h4*(s*s) + A44*h3y*h4y + A45*h3x*h4y + A45*h3y*h4x + A55*h3x*h4x)
            k0[pos+779] += detJ*wi*(A22*(cosa*cosa)*(h4*h4)*(s*s) + A44*(h4y*h4y) + 2*A45*h4x*h4y + A55*(h4x*h4x))
            k0[pos+780] += detJ*wi*(A22*(cosa*cosa)*h4*h5*(s*s) + A44*h4y*h5y + A45*h4x*h5y + A45*h4y*h5x + A55*h4x*h5x)
            k0[pos+781] += detJ*wi*(A22*(cosa*cosa)*h4*h6*(s*s) + A44*h4y*h6y + A45*h4x*h6y + A45*h4y*h6x + A55*h4x*h6x)
            k0[pos+782] += detJ*wi*(A22*(cosa*cosa)*h4*h7*(s*s) + A44*h4y*h7y + A45*h4x*h7y + A45*h4y*h7x + A55*h4x*h7x)
            k0[pos+783] += detJ*wi*(A22*(cosa*cosa)*h4*h8*(s*s) + A44*h4y*h8y + A45*h4x*h8y + A45*h4y*h8x + A55*h4x*h8x)
            k0[pos+784] += detJ*wi*(A45*h1*h4y + A55*h1*h4x + cosa*h4*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+785] += detJ*wi*(A45*h2*h4y + A55*h2*h4x + cosa*h4*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+786] += detJ*wi*(A45*h3*h4y + A55*h3*h4x + cosa*h4*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+787] += detJ*h4*wi*(A45*h4y + A55*h4x + cosa*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+788] += detJ*wi*(A45*h4y*h5 + A55*h4x*h5 + cosa*h4*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+789] += detJ*wi*(A45*h4y*h6 + A55*h4x*h6 + cosa*h4*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+790] += detJ*wi*(A45*h4y*h7 + A55*h4x*h7 + cosa*h4*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+791] += detJ*wi*(A45*h4y*h8 + A55*h4x*h8 + cosa*h4*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+792] += detJ*wi*(A44*h1*h4y + A45*h1*h4x + cosa*h4*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+793] += detJ*wi*(A44*h2*h4y + A45*h2*h4x + cosa*h4*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+794] += detJ*wi*(A44*h3*h4y + A45*h3*h4x + cosa*h4*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+795] += detJ*h4*wi*(A44*h4y + A45*h4x + cosa*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+796] += detJ*wi*(A44*h4y*h5 + A45*h4x*h5 + cosa*h4*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+797] += detJ*wi*(A44*h4y*h6 + A45*h4x*h6 + cosa*h4*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+798] += detJ*wi*(A44*h4y*h7 + A45*h4x*h7 + cosa*h4*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+799] += detJ*wi*(A44*h4y*h8 + A45*h4x*h8 + cosa*h4*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+800] += cosa*detJ*h5*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+801] += cosa*detJ*h5*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+802] += cosa*detJ*h5*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+803] += cosa*detJ*h5*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+804] += cosa*detJ*h5*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+805] += cosa*detJ*h5*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+806] += cosa*detJ*h5*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+807] += cosa*detJ*h5*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+808] += cosa*detJ*s*wi*(A22*h1y*h5 + A26*h5*(-h1*s*sina + h1x) - h1*(A44*h5y + A45*h5x))
            k0[pos+809] += cosa*detJ*s*wi*(A22*h2y*h5 + A26*h5*(-h2*s*sina + h2x) - h2*(A44*h5y + A45*h5x))
            k0[pos+810] += cosa*detJ*s*wi*(A22*h3y*h5 + A26*h5*(-h3*s*sina + h3x) - h3*(A44*h5y + A45*h5x))
            k0[pos+811] += cosa*detJ*s*wi*(A22*h4y*h5 + A26*h5*(-h4*s*sina + h4x) - h4*(A44*h5y + A45*h5x))
            k0[pos+812] += -cosa*detJ*h5*s*wi*(-A22*h5y + A26*h5*s*sina - A26*h5x + A44*h5y + A45*h5x)
            k0[pos+813] += cosa*detJ*s*wi*(A22*h5*h6y + A26*h5*(-h6*s*sina + h6x) - h6*(A44*h5y + A45*h5x))
            k0[pos+814] += cosa*detJ*s*wi*(A22*h5*h7y + A26*h5*(-h7*s*sina + h7x) - h7*(A44*h5y + A45*h5x))
            k0[pos+815] += cosa*detJ*s*wi*(A22*h5*h8y + A26*h5*(-h8*s*sina + h8x) - h8*(A44*h5y + A45*h5x))
            k0[pos+816] += detJ*wi*(A22*(cosa*cosa)*h1*h5*(s*s) + A44*h1y*h5y + A45*h1x*h5y + A45*h1y*h5x + A55*h1x*h5x)
            k0[pos+817] += detJ*wi*(A22*(cosa*cosa)*h2*h5*(s*s) + A44*h2y*h5y + A45*h2x*h5y + A45*h2y*h5x + A55*h2x*h5x)
            k0[pos+818] += detJ*wi*(A22*(cosa*cosa)*h3*h5*(s*s) + A44*h3y*h5y + A45*h3x*h5y + A45*h3y*h5x + A55*h3x*h5x)
            k0[pos+819] += detJ*wi*(A22*(cosa*cosa)*h4*h5*(s*s) + A44*h4y*h5y + A45*h4x*h5y + A45*h4y*h5x + A55*h4x*h5x)
            k0[pos+820] += detJ*wi*(A22*(cosa*cosa)*(h5*h5)*(s*s) + A44*(h5y*h5y) + 2*A45*h5x*h5y + A55*(h5x*h5x))
            k0[pos+821] += detJ*wi*(A22*(cosa*cosa)*h5*h6*(s*s) + A44*h5y*h6y + A45*h5x*h6y + A45*h5y*h6x + A55*h5x*h6x)
            k0[pos+822] += detJ*wi*(A22*(cosa*cosa)*h5*h7*(s*s) + A44*h5y*h7y + A45*h5x*h7y + A45*h5y*h7x + A55*h5x*h7x)
            k0[pos+823] += detJ*wi*(A22*(cosa*cosa)*h5*h8*(s*s) + A44*h5y*h8y + A45*h5x*h8y + A45*h5y*h8x + A55*h5x*h8x)
            k0[pos+824] += detJ*wi*(A45*h1*h5y + A55*h1*h5x + cosa*h5*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+825] += detJ*wi*(A45*h2*h5y + A55*h2*h5x + cosa*h5*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+826] += detJ*wi*(A45*h3*h5y + A55*h3*h5x + cosa*h5*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+827] += detJ*wi*(A45*h4*h5y + A55*h4*h5x + cosa*h5*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+828] += detJ*h5*wi*(A45*h5y + A55*h5x + cosa*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+829] += detJ*wi*(A45*h5y*h6 + A55*h5x*h6 + cosa*h5*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+830] += detJ*wi*(A45*h5y*h7 + A55*h5x*h7 + cosa*h5*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+831] += detJ*wi*(A45*h5y*h8 + A55*h5x*h8 + cosa*h5*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+832] += detJ*wi*(A44*h1*h5y + A45*h1*h5x + cosa*h5*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+833] += detJ*wi*(A44*h2*h5y + A45*h2*h5x + cosa*h5*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+834] += detJ*wi*(A44*h3*h5y + A45*h3*h5x + cosa*h5*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+835] += detJ*wi*(A44*h4*h5y + A45*h4*h5x + cosa*h5*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+836] += detJ*h5*wi*(A44*h5y + A45*h5x + cosa*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+837] += detJ*wi*(A44*h5y*h6 + A45*h5x*h6 + cosa*h5*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+838] += detJ*wi*(A44*h5y*h7 + A45*h5x*h7 + cosa*h5*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+839] += detJ*wi*(A44*h5y*h8 + A45*h5x*h8 + cosa*h5*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+840] += cosa*detJ*h6*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+841] += cosa*detJ*h6*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+842] += cosa*detJ*h6*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+843] += cosa*detJ*h6*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+844] += cosa*detJ*h6*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+845] += cosa*detJ*h6*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+846] += cosa*detJ*h6*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+847] += cosa*detJ*h6*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+848] += cosa*detJ*s*wi*(A22*h1y*h6 + A26*h6*(-h1*s*sina + h1x) - h1*(A44*h6y + A45*h6x))
            k0[pos+849] += cosa*detJ*s*wi*(A22*h2y*h6 + A26*h6*(-h2*s*sina + h2x) - h2*(A44*h6y + A45*h6x))
            k0[pos+850] += cosa*detJ*s*wi*(A22*h3y*h6 + A26*h6*(-h3*s*sina + h3x) - h3*(A44*h6y + A45*h6x))
            k0[pos+851] += cosa*detJ*s*wi*(A22*h4y*h6 + A26*h6*(-h4*s*sina + h4x) - h4*(A44*h6y + A45*h6x))
            k0[pos+852] += cosa*detJ*s*wi*(A22*h5y*h6 + A26*h6*(-h5*s*sina + h5x) - h5*(A44*h6y + A45*h6x))
            k0[pos+853] += -cosa*detJ*h6*s*wi*(-A22*h6y + A26*h6*s*sina - A26*h6x + A44*h6y + A45*h6x)
            k0[pos+854] += cosa*detJ*s*wi*(A22*h6*h7y + A26*h6*(-h7*s*sina + h7x) - h7*(A44*h6y + A45*h6x))
            k0[pos+855] += cosa*detJ*s*wi*(A22*h6*h8y + A26*h6*(-h8*s*sina + h8x) - h8*(A44*h6y + A45*h6x))
            k0[pos+856] += detJ*wi*(A22*(cosa*cosa)*h1*h6*(s*s) + A44*h1y*h6y + A45*h1x*h6y + A45*h1y*h6x + A55*h1x*h6x)
            k0[pos+857] += detJ*wi*(A22*(cosa*cosa)*h2*h6*(s*s) + A44*h2y*h6y + A45*h2x*h6y + A45*h2y*h6x + A55*h2x*h6x)
            k0[pos+858] += detJ*wi*(A22*(cosa*cosa)*h3*h6*(s*s) + A44*h3y*h6y + A45*h3x*h6y + A45*h3y*h6x + A55*h3x*h6x)
            k0[pos+859] += detJ*wi*(A22*(cosa*cosa)*h4*h6*(s*s) + A44*h4y*h6y + A45*h4x*h6y + A45*h4y*h6x + A55*h4x*h6x)
            k0[pos+860] += detJ*wi*(A22*(cosa*cosa)*h5*h6*(s*s) + A44*h5y*h6y + A45*h5x*h6y + A45*h5y*h6x + A55*h5x*h6x)
            k0[pos+861] += detJ*wi*(A22*(cosa*cosa)*(h6*h6)*(s*s) + A44*(h6y*h6y) + 2*A45*h6x*h6y + A55*(h6x*h6x))
            k0[pos+862] += detJ*wi*(A22*(cosa*cosa)*h6*h7*(s*s) + A44*h6y*h7y + A45*h6x*h7y + A45*h6y*h7x + A55*h6x*h7x)
            k0[pos+863] += detJ*wi*(A22*(cosa*cosa)*h6*h8*(s*s) + A44*h6y*h8y + A45*h6x*h8y + A45*h6y*h8x + A55*h6x*h8x)
            k0[pos+864] += detJ*wi*(A45*h1*h6y + A55*h1*h6x + cosa*h6*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+865] += detJ*wi*(A45*h2*h6y + A55*h2*h6x + cosa*h6*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+866] += detJ*wi*(A45*h3*h6y + A55*h3*h6x + cosa*h6*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+867] += detJ*wi*(A45*h4*h6y + A55*h4*h6x + cosa*h6*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+868] += detJ*wi*(A45*h5*h6y + A55*h5*h6x + cosa*h6*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+869] += detJ*h6*wi*(A45*h6y + A55*h6x + cosa*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+870] += detJ*wi*(A45*h6y*h7 + A55*h6x*h7 + cosa*h6*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+871] += detJ*wi*(A45*h6y*h8 + A55*h6x*h8 + cosa*h6*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+872] += detJ*wi*(A44*h1*h6y + A45*h1*h6x + cosa*h6*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+873] += detJ*wi*(A44*h2*h6y + A45*h2*h6x + cosa*h6*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+874] += detJ*wi*(A44*h3*h6y + A45*h3*h6x + cosa*h6*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+875] += detJ*wi*(A44*h4*h6y + A45*h4*h6x + cosa*h6*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+876] += detJ*wi*(A44*h5*h6y + A45*h5*h6x + cosa*h6*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+877] += detJ*h6*wi*(A44*h6y + A45*h6x + cosa*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+878] += detJ*wi*(A44*h6y*h7 + A45*h6x*h7 + cosa*h6*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+879] += detJ*wi*(A44*h6y*h8 + A45*h6x*h8 + cosa*h6*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+880] += cosa*detJ*h7*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+881] += cosa*detJ*h7*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+882] += cosa*detJ*h7*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+883] += cosa*detJ*h7*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+884] += cosa*detJ*h7*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+885] += cosa*detJ*h7*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+886] += cosa*detJ*h7*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+887] += cosa*detJ*h7*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+888] += cosa*detJ*s*wi*(A22*h1y*h7 + A26*h7*(-h1*s*sina + h1x) - h1*(A44*h7y + A45*h7x))
            k0[pos+889] += cosa*detJ*s*wi*(A22*h2y*h7 + A26*h7*(-h2*s*sina + h2x) - h2*(A44*h7y + A45*h7x))
            k0[pos+890] += cosa*detJ*s*wi*(A22*h3y*h7 + A26*h7*(-h3*s*sina + h3x) - h3*(A44*h7y + A45*h7x))
            k0[pos+891] += cosa*detJ*s*wi*(A22*h4y*h7 + A26*h7*(-h4*s*sina + h4x) - h4*(A44*h7y + A45*h7x))
            k0[pos+892] += cosa*detJ*s*wi*(A22*h5y*h7 + A26*h7*(-h5*s*sina + h5x) - h5*(A44*h7y + A45*h7x))
            k0[pos+893] += cosa*detJ*s*wi*(A22*h6y*h7 + A26*h7*(-h6*s*sina + h6x) - h6*(A44*h7y + A45*h7x))
            k0[pos+894] += -cosa*detJ*h7*s*wi*(-A22*h7y + A26*h7*s*sina - A26*h7x + A44*h7y + A45*h7x)
            k0[pos+895] += cosa*detJ*s*wi*(A22*h7*h8y + A26*h7*(-h8*s*sina + h8x) - h8*(A44*h7y + A45*h7x))
            k0[pos+896] += detJ*wi*(A22*(cosa*cosa)*h1*h7*(s*s) + A44*h1y*h7y + A45*h1x*h7y + A45*h1y*h7x + A55*h1x*h7x)
            k0[pos+897] += detJ*wi*(A22*(cosa*cosa)*h2*h7*(s*s) + A44*h2y*h7y + A45*h2x*h7y + A45*h2y*h7x + A55*h2x*h7x)
            k0[pos+898] += detJ*wi*(A22*(cosa*cosa)*h3*h7*(s*s) + A44*h3y*h7y + A45*h3x*h7y + A45*h3y*h7x + A55*h3x*h7x)
            k0[pos+899] += detJ*wi*(A22*(cosa*cosa)*h4*h7*(s*s) + A44*h4y*h7y + A45*h4x*h7y + A45*h4y*h7x + A55*h4x*h7x)
            k0[pos+900] += detJ*wi*(A22*(cosa*cosa)*h5*h7*(s*s) + A44*h5y*h7y + A45*h5x*h7y + A45*h5y*h7x + A55*h5x*h7x)
            k0[pos+901] += detJ*wi*(A22*(cosa*cosa)*h6*h7*(s*s) + A44*h6y*h7y + A45*h6x*h7y + A45*h6y*h7x + A55*h6x*h7x)
            k0[pos+902] += detJ*wi*(A22*(cosa*cosa)*(h7*h7)*(s*s) + A44*(h7y*h7y) + 2*A45*h7x*h7y + A55*(h7x*h7x))
            k0[pos+903] += detJ*wi*(A22*(cosa*cosa)*h7*h8*(s*s) + A44*h7y*h8y + A45*h7x*h8y + A45*h7y*h8x + A55*h7x*h8x)
            k0[pos+904] += detJ*wi*(A45*h1*h7y + A55*h1*h7x + cosa*h7*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+905] += detJ*wi*(A45*h2*h7y + A55*h2*h7x + cosa*h7*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+906] += detJ*wi*(A45*h3*h7y + A55*h3*h7x + cosa*h7*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+907] += detJ*wi*(A45*h4*h7y + A55*h4*h7x + cosa*h7*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+908] += detJ*wi*(A45*h5*h7y + A55*h5*h7x + cosa*h7*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+909] += detJ*wi*(A45*h6*h7y + A55*h6*h7x + cosa*h7*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+910] += detJ*h7*wi*(A45*h7y + A55*h7x + cosa*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+911] += detJ*wi*(A45*h7y*h8 + A55*h7x*h8 + cosa*h7*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+912] += detJ*wi*(A44*h1*h7y + A45*h1*h7x + cosa*h7*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+913] += detJ*wi*(A44*h2*h7y + A45*h2*h7x + cosa*h7*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+914] += detJ*wi*(A44*h3*h7y + A45*h3*h7x + cosa*h7*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+915] += detJ*wi*(A44*h4*h7y + A45*h4*h7x + cosa*h7*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+916] += detJ*wi*(A44*h5*h7y + A45*h5*h7x + cosa*h7*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+917] += detJ*wi*(A44*h6*h7y + A45*h6*h7x + cosa*h7*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+918] += detJ*h7*wi*(A44*h7y + A45*h7x + cosa*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+919] += detJ*wi*(A44*h7y*h8 + A45*h7x*h8 + cosa*h7*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+920] += cosa*detJ*h8*s*wi*(A12*h1x + A22*h1*s*sina + A26*h1y)
            k0[pos+921] += cosa*detJ*h8*s*wi*(A12*h2x + A22*h2*s*sina + A26*h2y)
            k0[pos+922] += cosa*detJ*h8*s*wi*(A12*h3x + A22*h3*s*sina + A26*h3y)
            k0[pos+923] += cosa*detJ*h8*s*wi*(A12*h4x + A22*h4*s*sina + A26*h4y)
            k0[pos+924] += cosa*detJ*h8*s*wi*(A12*h5x + A22*h5*s*sina + A26*h5y)
            k0[pos+925] += cosa*detJ*h8*s*wi*(A12*h6x + A22*h6*s*sina + A26*h6y)
            k0[pos+926] += cosa*detJ*h8*s*wi*(A12*h7x + A22*h7*s*sina + A26*h7y)
            k0[pos+927] += cosa*detJ*h8*s*wi*(A12*h8x + A22*h8*s*sina + A26*h8y)
            k0[pos+928] += cosa*detJ*s*wi*(A22*h1y*h8 + A26*h8*(-h1*s*sina + h1x) - h1*(A44*h8y + A45*h8x))
            k0[pos+929] += cosa*detJ*s*wi*(A22*h2y*h8 + A26*h8*(-h2*s*sina + h2x) - h2*(A44*h8y + A45*h8x))
            k0[pos+930] += cosa*detJ*s*wi*(A22*h3y*h8 + A26*h8*(-h3*s*sina + h3x) - h3*(A44*h8y + A45*h8x))
            k0[pos+931] += cosa*detJ*s*wi*(A22*h4y*h8 + A26*h8*(-h4*s*sina + h4x) - h4*(A44*h8y + A45*h8x))
            k0[pos+932] += cosa*detJ*s*wi*(A22*h5y*h8 + A26*h8*(-h5*s*sina + h5x) - h5*(A44*h8y + A45*h8x))
            k0[pos+933] += cosa*detJ*s*wi*(A22*h6y*h8 + A26*h8*(-h6*s*sina + h6x) - h6*(A44*h8y + A45*h8x))
            k0[pos+934] += cosa*detJ*s*wi*(A22*h7y*h8 + A26*h8*(-h7*s*sina + h7x) - h7*(A44*h8y + A45*h8x))
            k0[pos+935] += -cosa*detJ*h8*s*wi*(-A22*h8y + A26*h8*s*sina - A26*h8x + A44*h8y + A45*h8x)
            k0[pos+936] += detJ*wi*(A22*(cosa*cosa)*h1*h8*(s*s) + A44*h1y*h8y + A45*h1x*h8y + A45*h1y*h8x + A55*h1x*h8x)
            k0[pos+937] += detJ*wi*(A22*(cosa*cosa)*h2*h8*(s*s) + A44*h2y*h8y + A45*h2x*h8y + A45*h2y*h8x + A55*h2x*h8x)
            k0[pos+938] += detJ*wi*(A22*(cosa*cosa)*h3*h8*(s*s) + A44*h3y*h8y + A45*h3x*h8y + A45*h3y*h8x + A55*h3x*h8x)
            k0[pos+939] += detJ*wi*(A22*(cosa*cosa)*h4*h8*(s*s) + A44*h4y*h8y + A45*h4x*h8y + A45*h4y*h8x + A55*h4x*h8x)
            k0[pos+940] += detJ*wi*(A22*(cosa*cosa)*h5*h8*(s*s) + A44*h5y*h8y + A45*h5x*h8y + A45*h5y*h8x + A55*h5x*h8x)
            k0[pos+941] += detJ*wi*(A22*(cosa*cosa)*h6*h8*(s*s) + A44*h6y*h8y + A45*h6x*h8y + A45*h6y*h8x + A55*h6x*h8x)
            k0[pos+942] += detJ*wi*(A22*(cosa*cosa)*h7*h8*(s*s) + A44*h7y*h8y + A45*h7x*h8y + A45*h7y*h8x + A55*h7x*h8x)
            k0[pos+943] += detJ*wi*(A22*(cosa*cosa)*(h8*h8)*(s*s) + A44*(h8y*h8y) + 2*A45*h8x*h8y + A55*(h8x*h8x))
            k0[pos+944] += detJ*wi*(A45*h1*h8y + A55*h1*h8x + cosa*h8*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+945] += detJ*wi*(A45*h2*h8y + A55*h2*h8x + cosa*h8*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+946] += detJ*wi*(A45*h3*h8y + A55*h3*h8x + cosa*h8*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+947] += detJ*wi*(A45*h4*h8y + A55*h4*h8x + cosa*h8*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+948] += detJ*wi*(A45*h5*h8y + A55*h5*h8x + cosa*h8*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+949] += detJ*wi*(A45*h6*h8y + A55*h6*h8x + cosa*h8*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+950] += detJ*wi*(A45*h7*h8y + A55*h7*h8x + cosa*h8*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+951] += detJ*h8*wi*(A45*h8y + A55*h8x + cosa*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+952] += detJ*wi*(A44*h1*h8y + A45*h1*h8x + cosa*h8*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+953] += detJ*wi*(A44*h2*h8y + A45*h2*h8x + cosa*h8*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+954] += detJ*wi*(A44*h3*h8y + A45*h3*h8x + cosa*h8*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+955] += detJ*wi*(A44*h4*h8y + A45*h4*h8x + cosa*h8*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+956] += detJ*wi*(A44*h5*h8y + A45*h5*h8x + cosa*h8*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+957] += detJ*wi*(A44*h6*h8y + A45*h6*h8x + cosa*h8*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+958] += detJ*wi*(A44*h7*h8y + A45*h7*h8x + cosa*h8*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+959] += detJ*h8*wi*(A44*h8y + A45*h8x + cosa*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+960] += detJ*wi*(B11*(h1x*h1x) + 2*B12*h1*h1x*s*sina + 2*B16*h1x*h1y + B22*(h1*h1)*(s*s)*(sina*sina) + 2*B26*h1*h1y*s*sina + B66*(h1y*h1y))
            k0[pos+961] += detJ*wi*(h2*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h2x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h2y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+962] += detJ*wi*(h3*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h3x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h3y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+963] += detJ*wi*(h4*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h4x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h4y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+964] += detJ*wi*(h5*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h5x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h5y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+965] += detJ*wi*(h6*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h6x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h6y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+966] += detJ*wi*(h7*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h7x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h7y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+967] += detJ*wi*(h8*s*sina*(B12*h1x + B22*h1*s*sina + B26*h1y) + h8x*(B11*h1x + B12*h1*s*sina + B16*h1y) + h8y*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+968] += detJ*wi*(-A45*cosa*(h1*h1)*s + h1y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h1*s*sina + h1x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+969] += detJ*wi*(-A45*cosa*h1*h2*s + h2y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h2*s*sina + h2x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+970] += detJ*wi*(-A45*cosa*h1*h3*s + h3y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h3*s*sina + h3x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+971] += detJ*wi*(-A45*cosa*h1*h4*s + h4y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h4*s*sina + h4x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+972] += detJ*wi*(-A45*cosa*h1*h5*s + h5y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h5*s*sina + h5x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+973] += detJ*wi*(-A45*cosa*h1*h6*s + h6y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h6*s*sina + h6x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+974] += detJ*wi*(-A45*cosa*h1*h7*s + h7y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h7*s*sina + h7x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+975] += detJ*wi*(-A45*cosa*h1*h8*s + h8y*(B12*h1x + B22*h1*s*sina + B26*h1y) + (-h8*s*sina + h8x)*(B16*h1x + B26*h1*s*sina + B66*h1y))
            k0[pos+976] += detJ*h1*wi*(A45*h1y + A55*h1x + cosa*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+977] += detJ*wi*(A45*h1*h2y + A55*h1*h2x + cosa*h2*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+978] += detJ*wi*(A45*h1*h3y + A55*h1*h3x + cosa*h3*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+979] += detJ*wi*(A45*h1*h4y + A55*h1*h4x + cosa*h4*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+980] += detJ*wi*(A45*h1*h5y + A55*h1*h5x + cosa*h5*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+981] += detJ*wi*(A45*h1*h6y + A55*h1*h6x + cosa*h6*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+982] += detJ*wi*(A45*h1*h7y + A55*h1*h7x + cosa*h7*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+983] += detJ*wi*(A45*h1*h8y + A55*h1*h8x + cosa*h8*s*(B12*h1x + B22*h1*s*sina + B26*h1y))
            k0[pos+984] += detJ*wi*(A55*(h1*h1) + D11*(h1x*h1x) + 2*D12*h1*h1x*s*sina + 2*D16*h1x*h1y + D22*(h1*h1)*(s*s)*(sina*sina) + 2*D26*h1*h1y*s*sina + D66*(h1y*h1y))
            k0[pos+985] += detJ*wi*(A55*h1*h2 + h2*s*sina*(D12*h1x + D22*h1*s*sina + D26*h1y) + h2x*(D11*h1x + D12*h1*s*sina + D16*h1y) + h2y*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+986] += detJ*wi*(A55*h1*h3 + h3*s*sina*(D12*h1x + D22*h1*s*sina + D26*h1y) + h3x*(D11*h1x + D12*h1*s*sina + D16*h1y) + h3y*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+987] += detJ*wi*(A55*h1*h4 + h4*s*sina*(D12*h1x + D22*h1*s*sina + D26*h1y) + h4x*(D11*h1x + D12*h1*s*sina + D16*h1y) + h4y*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+988] += detJ*wi*(A55*h1*h5 + h5*s*sina*(D12*h1x + D22*h1*s*sina + D26*h1y) + h5x*(D11*h1x + D12*h1*s*sina + D16*h1y) + h5y*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+989] += detJ*wi*(A55*h1*h6 + h6*s*sina*(D12*h1x + D22*h1*s*sina + D26*h1y) + h6x*(D11*h1x + D12*h1*s*sina + D16*h1y) + h6y*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+990] += detJ*wi*(A55*h1*h7 + h7*s*sina*(D12*h1x + D22*h1*s*sina + D26*h1y) + h7x*(D11*h1x + D12*h1*s*sina + D16*h1y) + h7y*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+991] += detJ*wi*(A55*h1*h8 + h8*s*sina*(D12*h1x + D22*h1*s*sina + D26*h1y) + h8x*(D11*h1x + D12*h1*s*sina + D16*h1y) + h8y*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+992] += detJ*wi*(A45*(h1*h1) + h1y*(D12*h1x + D22*h1*s*sina + D26*h1y) + (-h1*s*sina + h1x)*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+993] += detJ*wi*(A45*h1*h2 + h2y*(D12*h1x + D22*h1*s*sina + D26*h1y) + (-h2*s*sina + h2x)*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+994] += detJ*wi*(A45*h1*h3 + h3y*(D12*h1x + D22*h1*s*sina + D26*h1y) + (-h3*s*sina + h3x)*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+995] += detJ*wi*(A45*h1*h4 + h4y*(D12*h1x + D22*h1*s*sina + D26*h1y) + (-h4*s*sina + h4x)*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+996] += detJ*wi*(A45*h1*h5 + h5y*(D12*h1x + D22*h1*s*sina + D26*h1y) + (-h5*s*sina + h5x)*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+997] += detJ*wi*(A45*h1*h6 + h6y*(D12*h1x + D22*h1*s*sina + D26*h1y) + (-h6*s*sina + h6x)*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+998] += detJ*wi*(A45*h1*h7 + h7y*(D12*h1x + D22*h1*s*sina + D26*h1y) + (-h7*s*sina + h7x)*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+999] += detJ*wi*(A45*h1*h8 + h8y*(D12*h1x + D22*h1*s*sina + D26*h1y) + (-h8*s*sina + h8x)*(D16*h1x + D26*h1*s*sina + D66*h1y))
            k0[pos+1000] += detJ*wi*(h1*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h1x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h1y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1001] += detJ*wi*(B11*(h2x*h2x) + 2*B12*h2*h2x*s*sina + 2*B16*h2x*h2y + B22*(h2*h2)*(s*s)*(sina*sina) + 2*B26*h2*h2y*s*sina + B66*(h2y*h2y))
            k0[pos+1002] += detJ*wi*(h3*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h3x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h3y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1003] += detJ*wi*(h4*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h4x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h4y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1004] += detJ*wi*(h5*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h5x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h5y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1005] += detJ*wi*(h6*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h6x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h6y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1006] += detJ*wi*(h7*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h7x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h7y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1007] += detJ*wi*(h8*s*sina*(B12*h2x + B22*h2*s*sina + B26*h2y) + h8x*(B11*h2x + B12*h2*s*sina + B16*h2y) + h8y*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1008] += detJ*wi*(-A45*cosa*h1*h2*s + h1y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h1*s*sina + h1x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1009] += detJ*wi*(-A45*cosa*(h2*h2)*s + h2y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h2*s*sina + h2x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1010] += detJ*wi*(-A45*cosa*h2*h3*s + h3y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h3*s*sina + h3x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1011] += detJ*wi*(-A45*cosa*h2*h4*s + h4y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h4*s*sina + h4x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1012] += detJ*wi*(-A45*cosa*h2*h5*s + h5y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h5*s*sina + h5x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1013] += detJ*wi*(-A45*cosa*h2*h6*s + h6y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h6*s*sina + h6x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1014] += detJ*wi*(-A45*cosa*h2*h7*s + h7y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h7*s*sina + h7x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1015] += detJ*wi*(-A45*cosa*h2*h8*s + h8y*(B12*h2x + B22*h2*s*sina + B26*h2y) + (-h8*s*sina + h8x)*(B16*h2x + B26*h2*s*sina + B66*h2y))
            k0[pos+1016] += detJ*wi*(A45*h1y*h2 + A55*h1x*h2 + cosa*h1*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+1017] += detJ*h2*wi*(A45*h2y + A55*h2x + cosa*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+1018] += detJ*wi*(A45*h2*h3y + A55*h2*h3x + cosa*h3*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+1019] += detJ*wi*(A45*h2*h4y + A55*h2*h4x + cosa*h4*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+1020] += detJ*wi*(A45*h2*h5y + A55*h2*h5x + cosa*h5*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+1021] += detJ*wi*(A45*h2*h6y + A55*h2*h6x + cosa*h6*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+1022] += detJ*wi*(A45*h2*h7y + A55*h2*h7x + cosa*h7*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+1023] += detJ*wi*(A45*h2*h8y + A55*h2*h8x + cosa*h8*s*(B12*h2x + B22*h2*s*sina + B26*h2y))
            k0[pos+1024] += detJ*wi*(A55*h1*h2 + h1*s*sina*(D12*h2x + D22*h2*s*sina + D26*h2y) + h1x*(D11*h2x + D12*h2*s*sina + D16*h2y) + h1y*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1025] += detJ*wi*(A55*(h2*h2) + D11*(h2x*h2x) + 2*D12*h2*h2x*s*sina + 2*D16*h2x*h2y + D22*(h2*h2)*(s*s)*(sina*sina) + 2*D26*h2*h2y*s*sina + D66*(h2y*h2y))
            k0[pos+1026] += detJ*wi*(A55*h2*h3 + h3*s*sina*(D12*h2x + D22*h2*s*sina + D26*h2y) + h3x*(D11*h2x + D12*h2*s*sina + D16*h2y) + h3y*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1027] += detJ*wi*(A55*h2*h4 + h4*s*sina*(D12*h2x + D22*h2*s*sina + D26*h2y) + h4x*(D11*h2x + D12*h2*s*sina + D16*h2y) + h4y*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1028] += detJ*wi*(A55*h2*h5 + h5*s*sina*(D12*h2x + D22*h2*s*sina + D26*h2y) + h5x*(D11*h2x + D12*h2*s*sina + D16*h2y) + h5y*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1029] += detJ*wi*(A55*h2*h6 + h6*s*sina*(D12*h2x + D22*h2*s*sina + D26*h2y) + h6x*(D11*h2x + D12*h2*s*sina + D16*h2y) + h6y*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1030] += detJ*wi*(A55*h2*h7 + h7*s*sina*(D12*h2x + D22*h2*s*sina + D26*h2y) + h7x*(D11*h2x + D12*h2*s*sina + D16*h2y) + h7y*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1031] += detJ*wi*(A55*h2*h8 + h8*s*sina*(D12*h2x + D22*h2*s*sina + D26*h2y) + h8x*(D11*h2x + D12*h2*s*sina + D16*h2y) + h8y*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1032] += detJ*wi*(A45*h1*h2 + h1y*(D12*h2x + D22*h2*s*sina + D26*h2y) + (-h1*s*sina + h1x)*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1033] += detJ*wi*(A45*(h2*h2) + h2y*(D12*h2x + D22*h2*s*sina + D26*h2y) + (-h2*s*sina + h2x)*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1034] += detJ*wi*(A45*h2*h3 + h3y*(D12*h2x + D22*h2*s*sina + D26*h2y) + (-h3*s*sina + h3x)*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1035] += detJ*wi*(A45*h2*h4 + h4y*(D12*h2x + D22*h2*s*sina + D26*h2y) + (-h4*s*sina + h4x)*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1036] += detJ*wi*(A45*h2*h5 + h5y*(D12*h2x + D22*h2*s*sina + D26*h2y) + (-h5*s*sina + h5x)*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1037] += detJ*wi*(A45*h2*h6 + h6y*(D12*h2x + D22*h2*s*sina + D26*h2y) + (-h6*s*sina + h6x)*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1038] += detJ*wi*(A45*h2*h7 + h7y*(D12*h2x + D22*h2*s*sina + D26*h2y) + (-h7*s*sina + h7x)*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1039] += detJ*wi*(A45*h2*h8 + h8y*(D12*h2x + D22*h2*s*sina + D26*h2y) + (-h8*s*sina + h8x)*(D16*h2x + D26*h2*s*sina + D66*h2y))
            k0[pos+1040] += detJ*wi*(h1*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h1x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h1y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1041] += detJ*wi*(h2*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h2x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h2y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1042] += detJ*wi*(B11*(h3x*h3x) + 2*B12*h3*h3x*s*sina + 2*B16*h3x*h3y + B22*(h3*h3)*(s*s)*(sina*sina) + 2*B26*h3*h3y*s*sina + B66*(h3y*h3y))
            k0[pos+1043] += detJ*wi*(h4*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h4x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h4y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1044] += detJ*wi*(h5*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h5x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h5y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1045] += detJ*wi*(h6*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h6x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h6y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1046] += detJ*wi*(h7*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h7x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h7y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1047] += detJ*wi*(h8*s*sina*(B12*h3x + B22*h3*s*sina + B26*h3y) + h8x*(B11*h3x + B12*h3*s*sina + B16*h3y) + h8y*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1048] += detJ*wi*(-A45*cosa*h1*h3*s + h1y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h1*s*sina + h1x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1049] += detJ*wi*(-A45*cosa*h2*h3*s + h2y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h2*s*sina + h2x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1050] += detJ*wi*(-A45*cosa*(h3*h3)*s + h3y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h3*s*sina + h3x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1051] += detJ*wi*(-A45*cosa*h3*h4*s + h4y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h4*s*sina + h4x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1052] += detJ*wi*(-A45*cosa*h3*h5*s + h5y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h5*s*sina + h5x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1053] += detJ*wi*(-A45*cosa*h3*h6*s + h6y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h6*s*sina + h6x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1054] += detJ*wi*(-A45*cosa*h3*h7*s + h7y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h7*s*sina + h7x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1055] += detJ*wi*(-A45*cosa*h3*h8*s + h8y*(B12*h3x + B22*h3*s*sina + B26*h3y) + (-h8*s*sina + h8x)*(B16*h3x + B26*h3*s*sina + B66*h3y))
            k0[pos+1056] += detJ*wi*(A45*h1y*h3 + A55*h1x*h3 + cosa*h1*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+1057] += detJ*wi*(A45*h2y*h3 + A55*h2x*h3 + cosa*h2*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+1058] += detJ*h3*wi*(A45*h3y + A55*h3x + cosa*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+1059] += detJ*wi*(A45*h3*h4y + A55*h3*h4x + cosa*h4*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+1060] += detJ*wi*(A45*h3*h5y + A55*h3*h5x + cosa*h5*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+1061] += detJ*wi*(A45*h3*h6y + A55*h3*h6x + cosa*h6*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+1062] += detJ*wi*(A45*h3*h7y + A55*h3*h7x + cosa*h7*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+1063] += detJ*wi*(A45*h3*h8y + A55*h3*h8x + cosa*h8*s*(B12*h3x + B22*h3*s*sina + B26*h3y))
            k0[pos+1064] += detJ*wi*(A55*h1*h3 + h1*s*sina*(D12*h3x + D22*h3*s*sina + D26*h3y) + h1x*(D11*h3x + D12*h3*s*sina + D16*h3y) + h1y*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1065] += detJ*wi*(A55*h2*h3 + h2*s*sina*(D12*h3x + D22*h3*s*sina + D26*h3y) + h2x*(D11*h3x + D12*h3*s*sina + D16*h3y) + h2y*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1066] += detJ*wi*(A55*(h3*h3) + D11*(h3x*h3x) + 2*D12*h3*h3x*s*sina + 2*D16*h3x*h3y + D22*(h3*h3)*(s*s)*(sina*sina) + 2*D26*h3*h3y*s*sina + D66*(h3y*h3y))
            k0[pos+1067] += detJ*wi*(A55*h3*h4 + h4*s*sina*(D12*h3x + D22*h3*s*sina + D26*h3y) + h4x*(D11*h3x + D12*h3*s*sina + D16*h3y) + h4y*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1068] += detJ*wi*(A55*h3*h5 + h5*s*sina*(D12*h3x + D22*h3*s*sina + D26*h3y) + h5x*(D11*h3x + D12*h3*s*sina + D16*h3y) + h5y*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1069] += detJ*wi*(A55*h3*h6 + h6*s*sina*(D12*h3x + D22*h3*s*sina + D26*h3y) + h6x*(D11*h3x + D12*h3*s*sina + D16*h3y) + h6y*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1070] += detJ*wi*(A55*h3*h7 + h7*s*sina*(D12*h3x + D22*h3*s*sina + D26*h3y) + h7x*(D11*h3x + D12*h3*s*sina + D16*h3y) + h7y*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1071] += detJ*wi*(A55*h3*h8 + h8*s*sina*(D12*h3x + D22*h3*s*sina + D26*h3y) + h8x*(D11*h3x + D12*h3*s*sina + D16*h3y) + h8y*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1072] += detJ*wi*(A45*h1*h3 + h1y*(D12*h3x + D22*h3*s*sina + D26*h3y) + (-h1*s*sina + h1x)*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1073] += detJ*wi*(A45*h2*h3 + h2y*(D12*h3x + D22*h3*s*sina + D26*h3y) + (-h2*s*sina + h2x)*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1074] += detJ*wi*(A45*(h3*h3) + h3y*(D12*h3x + D22*h3*s*sina + D26*h3y) + (-h3*s*sina + h3x)*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1075] += detJ*wi*(A45*h3*h4 + h4y*(D12*h3x + D22*h3*s*sina + D26*h3y) + (-h4*s*sina + h4x)*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1076] += detJ*wi*(A45*h3*h5 + h5y*(D12*h3x + D22*h3*s*sina + D26*h3y) + (-h5*s*sina + h5x)*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1077] += detJ*wi*(A45*h3*h6 + h6y*(D12*h3x + D22*h3*s*sina + D26*h3y) + (-h6*s*sina + h6x)*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1078] += detJ*wi*(A45*h3*h7 + h7y*(D12*h3x + D22*h3*s*sina + D26*h3y) + (-h7*s*sina + h7x)*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1079] += detJ*wi*(A45*h3*h8 + h8y*(D12*h3x + D22*h3*s*sina + D26*h3y) + (-h8*s*sina + h8x)*(D16*h3x + D26*h3*s*sina + D66*h3y))
            k0[pos+1080] += detJ*wi*(h1*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h1x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h1y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1081] += detJ*wi*(h2*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h2x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h2y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1082] += detJ*wi*(h3*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h3x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h3y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1083] += detJ*wi*(B11*(h4x*h4x) + 2*B12*h4*h4x*s*sina + 2*B16*h4x*h4y + B22*(h4*h4)*(s*s)*(sina*sina) + 2*B26*h4*h4y*s*sina + B66*(h4y*h4y))
            k0[pos+1084] += detJ*wi*(h5*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h5x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h5y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1085] += detJ*wi*(h6*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h6x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h6y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1086] += detJ*wi*(h7*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h7x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h7y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1087] += detJ*wi*(h8*s*sina*(B12*h4x + B22*h4*s*sina + B26*h4y) + h8x*(B11*h4x + B12*h4*s*sina + B16*h4y) + h8y*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1088] += detJ*wi*(-A45*cosa*h1*h4*s + h1y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h1*s*sina + h1x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1089] += detJ*wi*(-A45*cosa*h2*h4*s + h2y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h2*s*sina + h2x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1090] += detJ*wi*(-A45*cosa*h3*h4*s + h3y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h3*s*sina + h3x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1091] += detJ*wi*(-A45*cosa*(h4*h4)*s + h4y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h4*s*sina + h4x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1092] += detJ*wi*(-A45*cosa*h4*h5*s + h5y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h5*s*sina + h5x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1093] += detJ*wi*(-A45*cosa*h4*h6*s + h6y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h6*s*sina + h6x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1094] += detJ*wi*(-A45*cosa*h4*h7*s + h7y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h7*s*sina + h7x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1095] += detJ*wi*(-A45*cosa*h4*h8*s + h8y*(B12*h4x + B22*h4*s*sina + B26*h4y) + (-h8*s*sina + h8x)*(B16*h4x + B26*h4*s*sina + B66*h4y))
            k0[pos+1096] += detJ*wi*(A45*h1y*h4 + A55*h1x*h4 + cosa*h1*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+1097] += detJ*wi*(A45*h2y*h4 + A55*h2x*h4 + cosa*h2*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+1098] += detJ*wi*(A45*h3y*h4 + A55*h3x*h4 + cosa*h3*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+1099] += detJ*h4*wi*(A45*h4y + A55*h4x + cosa*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+1100] += detJ*wi*(A45*h4*h5y + A55*h4*h5x + cosa*h5*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+1101] += detJ*wi*(A45*h4*h6y + A55*h4*h6x + cosa*h6*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+1102] += detJ*wi*(A45*h4*h7y + A55*h4*h7x + cosa*h7*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+1103] += detJ*wi*(A45*h4*h8y + A55*h4*h8x + cosa*h8*s*(B12*h4x + B22*h4*s*sina + B26*h4y))
            k0[pos+1104] += detJ*wi*(A55*h1*h4 + h1*s*sina*(D12*h4x + D22*h4*s*sina + D26*h4y) + h1x*(D11*h4x + D12*h4*s*sina + D16*h4y) + h1y*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1105] += detJ*wi*(A55*h2*h4 + h2*s*sina*(D12*h4x + D22*h4*s*sina + D26*h4y) + h2x*(D11*h4x + D12*h4*s*sina + D16*h4y) + h2y*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1106] += detJ*wi*(A55*h3*h4 + h3*s*sina*(D12*h4x + D22*h4*s*sina + D26*h4y) + h3x*(D11*h4x + D12*h4*s*sina + D16*h4y) + h3y*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1107] += detJ*wi*(A55*(h4*h4) + D11*(h4x*h4x) + 2*D12*h4*h4x*s*sina + 2*D16*h4x*h4y + D22*(h4*h4)*(s*s)*(sina*sina) + 2*D26*h4*h4y*s*sina + D66*(h4y*h4y))
            k0[pos+1108] += detJ*wi*(A55*h4*h5 + h5*s*sina*(D12*h4x + D22*h4*s*sina + D26*h4y) + h5x*(D11*h4x + D12*h4*s*sina + D16*h4y) + h5y*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1109] += detJ*wi*(A55*h4*h6 + h6*s*sina*(D12*h4x + D22*h4*s*sina + D26*h4y) + h6x*(D11*h4x + D12*h4*s*sina + D16*h4y) + h6y*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1110] += detJ*wi*(A55*h4*h7 + h7*s*sina*(D12*h4x + D22*h4*s*sina + D26*h4y) + h7x*(D11*h4x + D12*h4*s*sina + D16*h4y) + h7y*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1111] += detJ*wi*(A55*h4*h8 + h8*s*sina*(D12*h4x + D22*h4*s*sina + D26*h4y) + h8x*(D11*h4x + D12*h4*s*sina + D16*h4y) + h8y*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1112] += detJ*wi*(A45*h1*h4 + h1y*(D12*h4x + D22*h4*s*sina + D26*h4y) + (-h1*s*sina + h1x)*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1113] += detJ*wi*(A45*h2*h4 + h2y*(D12*h4x + D22*h4*s*sina + D26*h4y) + (-h2*s*sina + h2x)*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1114] += detJ*wi*(A45*h3*h4 + h3y*(D12*h4x + D22*h4*s*sina + D26*h4y) + (-h3*s*sina + h3x)*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1115] += detJ*wi*(A45*(h4*h4) + h4y*(D12*h4x + D22*h4*s*sina + D26*h4y) + (-h4*s*sina + h4x)*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1116] += detJ*wi*(A45*h4*h5 + h5y*(D12*h4x + D22*h4*s*sina + D26*h4y) + (-h5*s*sina + h5x)*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1117] += detJ*wi*(A45*h4*h6 + h6y*(D12*h4x + D22*h4*s*sina + D26*h4y) + (-h6*s*sina + h6x)*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1118] += detJ*wi*(A45*h4*h7 + h7y*(D12*h4x + D22*h4*s*sina + D26*h4y) + (-h7*s*sina + h7x)*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1119] += detJ*wi*(A45*h4*h8 + h8y*(D12*h4x + D22*h4*s*sina + D26*h4y) + (-h8*s*sina + h8x)*(D16*h4x + D26*h4*s*sina + D66*h4y))
            k0[pos+1120] += detJ*wi*(h1*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h1x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h1y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1121] += detJ*wi*(h2*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h2x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h2y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1122] += detJ*wi*(h3*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h3x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h3y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1123] += detJ*wi*(h4*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h4x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h4y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1124] += detJ*wi*(B11*(h5x*h5x) + 2*B12*h5*h5x*s*sina + 2*B16*h5x*h5y + B22*(h5*h5)*(s*s)*(sina*sina) + 2*B26*h5*h5y*s*sina + B66*(h5y*h5y))
            k0[pos+1125] += detJ*wi*(h6*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h6x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h6y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1126] += detJ*wi*(h7*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h7x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h7y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1127] += detJ*wi*(h8*s*sina*(B12*h5x + B22*h5*s*sina + B26*h5y) + h8x*(B11*h5x + B12*h5*s*sina + B16*h5y) + h8y*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1128] += detJ*wi*(-A45*cosa*h1*h5*s + h1y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h1*s*sina + h1x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1129] += detJ*wi*(-A45*cosa*h2*h5*s + h2y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h2*s*sina + h2x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1130] += detJ*wi*(-A45*cosa*h3*h5*s + h3y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h3*s*sina + h3x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1131] += detJ*wi*(-A45*cosa*h4*h5*s + h4y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h4*s*sina + h4x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1132] += detJ*wi*(-A45*cosa*(h5*h5)*s + h5y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h5*s*sina + h5x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1133] += detJ*wi*(-A45*cosa*h5*h6*s + h6y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h6*s*sina + h6x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1134] += detJ*wi*(-A45*cosa*h5*h7*s + h7y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h7*s*sina + h7x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1135] += detJ*wi*(-A45*cosa*h5*h8*s + h8y*(B12*h5x + B22*h5*s*sina + B26*h5y) + (-h8*s*sina + h8x)*(B16*h5x + B26*h5*s*sina + B66*h5y))
            k0[pos+1136] += detJ*wi*(A45*h1y*h5 + A55*h1x*h5 + cosa*h1*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+1137] += detJ*wi*(A45*h2y*h5 + A55*h2x*h5 + cosa*h2*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+1138] += detJ*wi*(A45*h3y*h5 + A55*h3x*h5 + cosa*h3*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+1139] += detJ*wi*(A45*h4y*h5 + A55*h4x*h5 + cosa*h4*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+1140] += detJ*h5*wi*(A45*h5y + A55*h5x + cosa*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+1141] += detJ*wi*(A45*h5*h6y + A55*h5*h6x + cosa*h6*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+1142] += detJ*wi*(A45*h5*h7y + A55*h5*h7x + cosa*h7*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+1143] += detJ*wi*(A45*h5*h8y + A55*h5*h8x + cosa*h8*s*(B12*h5x + B22*h5*s*sina + B26*h5y))
            k0[pos+1144] += detJ*wi*(A55*h1*h5 + h1*s*sina*(D12*h5x + D22*h5*s*sina + D26*h5y) + h1x*(D11*h5x + D12*h5*s*sina + D16*h5y) + h1y*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1145] += detJ*wi*(A55*h2*h5 + h2*s*sina*(D12*h5x + D22*h5*s*sina + D26*h5y) + h2x*(D11*h5x + D12*h5*s*sina + D16*h5y) + h2y*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1146] += detJ*wi*(A55*h3*h5 + h3*s*sina*(D12*h5x + D22*h5*s*sina + D26*h5y) + h3x*(D11*h5x + D12*h5*s*sina + D16*h5y) + h3y*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1147] += detJ*wi*(A55*h4*h5 + h4*s*sina*(D12*h5x + D22*h5*s*sina + D26*h5y) + h4x*(D11*h5x + D12*h5*s*sina + D16*h5y) + h4y*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1148] += detJ*wi*(A55*(h5*h5) + D11*(h5x*h5x) + 2*D12*h5*h5x*s*sina + 2*D16*h5x*h5y + D22*(h5*h5)*(s*s)*(sina*sina) + 2*D26*h5*h5y*s*sina + D66*(h5y*h5y))
            k0[pos+1149] += detJ*wi*(A55*h5*h6 + h6*s*sina*(D12*h5x + D22*h5*s*sina + D26*h5y) + h6x*(D11*h5x + D12*h5*s*sina + D16*h5y) + h6y*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1150] += detJ*wi*(A55*h5*h7 + h7*s*sina*(D12*h5x + D22*h5*s*sina + D26*h5y) + h7x*(D11*h5x + D12*h5*s*sina + D16*h5y) + h7y*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1151] += detJ*wi*(A55*h5*h8 + h8*s*sina*(D12*h5x + D22*h5*s*sina + D26*h5y) + h8x*(D11*h5x + D12*h5*s*sina + D16*h5y) + h8y*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1152] += detJ*wi*(A45*h1*h5 + h1y*(D12*h5x + D22*h5*s*sina + D26*h5y) + (-h1*s*sina + h1x)*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1153] += detJ*wi*(A45*h2*h5 + h2y*(D12*h5x + D22*h5*s*sina + D26*h5y) + (-h2*s*sina + h2x)*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1154] += detJ*wi*(A45*h3*h5 + h3y*(D12*h5x + D22*h5*s*sina + D26*h5y) + (-h3*s*sina + h3x)*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1155] += detJ*wi*(A45*h4*h5 + h4y*(D12*h5x + D22*h5*s*sina + D26*h5y) + (-h4*s*sina + h4x)*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1156] += detJ*wi*(A45*(h5*h5) + h5y*(D12*h5x + D22*h5*s*sina + D26*h5y) + (-h5*s*sina + h5x)*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1157] += detJ*wi*(A45*h5*h6 + h6y*(D12*h5x + D22*h5*s*sina + D26*h5y) + (-h6*s*sina + h6x)*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1158] += detJ*wi*(A45*h5*h7 + h7y*(D12*h5x + D22*h5*s*sina + D26*h5y) + (-h7*s*sina + h7x)*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1159] += detJ*wi*(A45*h5*h8 + h8y*(D12*h5x + D22*h5*s*sina + D26*h5y) + (-h8*s*sina + h8x)*(D16*h5x + D26*h5*s*sina + D66*h5y))
            k0[pos+1160] += detJ*wi*(h1*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h1x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h1y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1161] += detJ*wi*(h2*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h2x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h2y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1162] += detJ*wi*(h3*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h3x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h3y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1163] += detJ*wi*(h4*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h4x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h4y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1164] += detJ*wi*(h5*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h5x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h5y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1165] += detJ*wi*(B11*(h6x*h6x) + 2*B12*h6*h6x*s*sina + 2*B16*h6x*h6y + B22*(h6*h6)*(s*s)*(sina*sina) + 2*B26*h6*h6y*s*sina + B66*(h6y*h6y))
            k0[pos+1166] += detJ*wi*(h7*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h7x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h7y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1167] += detJ*wi*(h8*s*sina*(B12*h6x + B22*h6*s*sina + B26*h6y) + h8x*(B11*h6x + B12*h6*s*sina + B16*h6y) + h8y*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1168] += detJ*wi*(-A45*cosa*h1*h6*s + h1y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h1*s*sina + h1x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1169] += detJ*wi*(-A45*cosa*h2*h6*s + h2y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h2*s*sina + h2x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1170] += detJ*wi*(-A45*cosa*h3*h6*s + h3y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h3*s*sina + h3x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1171] += detJ*wi*(-A45*cosa*h4*h6*s + h4y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h4*s*sina + h4x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1172] += detJ*wi*(-A45*cosa*h5*h6*s + h5y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h5*s*sina + h5x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1173] += detJ*wi*(-A45*cosa*(h6*h6)*s + h6y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h6*s*sina + h6x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1174] += detJ*wi*(-A45*cosa*h6*h7*s + h7y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h7*s*sina + h7x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1175] += detJ*wi*(-A45*cosa*h6*h8*s + h8y*(B12*h6x + B22*h6*s*sina + B26*h6y) + (-h8*s*sina + h8x)*(B16*h6x + B26*h6*s*sina + B66*h6y))
            k0[pos+1176] += detJ*wi*(A45*h1y*h6 + A55*h1x*h6 + cosa*h1*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+1177] += detJ*wi*(A45*h2y*h6 + A55*h2x*h6 + cosa*h2*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+1178] += detJ*wi*(A45*h3y*h6 + A55*h3x*h6 + cosa*h3*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+1179] += detJ*wi*(A45*h4y*h6 + A55*h4x*h6 + cosa*h4*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+1180] += detJ*wi*(A45*h5y*h6 + A55*h5x*h6 + cosa*h5*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+1181] += detJ*h6*wi*(A45*h6y + A55*h6x + cosa*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+1182] += detJ*wi*(A45*h6*h7y + A55*h6*h7x + cosa*h7*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+1183] += detJ*wi*(A45*h6*h8y + A55*h6*h8x + cosa*h8*s*(B12*h6x + B22*h6*s*sina + B26*h6y))
            k0[pos+1184] += detJ*wi*(A55*h1*h6 + h1*s*sina*(D12*h6x + D22*h6*s*sina + D26*h6y) + h1x*(D11*h6x + D12*h6*s*sina + D16*h6y) + h1y*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1185] += detJ*wi*(A55*h2*h6 + h2*s*sina*(D12*h6x + D22*h6*s*sina + D26*h6y) + h2x*(D11*h6x + D12*h6*s*sina + D16*h6y) + h2y*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1186] += detJ*wi*(A55*h3*h6 + h3*s*sina*(D12*h6x + D22*h6*s*sina + D26*h6y) + h3x*(D11*h6x + D12*h6*s*sina + D16*h6y) + h3y*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1187] += detJ*wi*(A55*h4*h6 + h4*s*sina*(D12*h6x + D22*h6*s*sina + D26*h6y) + h4x*(D11*h6x + D12*h6*s*sina + D16*h6y) + h4y*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1188] += detJ*wi*(A55*h5*h6 + h5*s*sina*(D12*h6x + D22*h6*s*sina + D26*h6y) + h5x*(D11*h6x + D12*h6*s*sina + D16*h6y) + h5y*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1189] += detJ*wi*(A55*(h6*h6) + D11*(h6x*h6x) + 2*D12*h6*h6x*s*sina + 2*D16*h6x*h6y + D22*(h6*h6)*(s*s)*(sina*sina) + 2*D26*h6*h6y*s*sina + D66*(h6y*h6y))
            k0[pos+1190] += detJ*wi*(A55*h6*h7 + h7*s*sina*(D12*h6x + D22*h6*s*sina + D26*h6y) + h7x*(D11*h6x + D12*h6*s*sina + D16*h6y) + h7y*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1191] += detJ*wi*(A55*h6*h8 + h8*s*sina*(D12*h6x + D22*h6*s*sina + D26*h6y) + h8x*(D11*h6x + D12*h6*s*sina + D16*h6y) + h8y*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1192] += detJ*wi*(A45*h1*h6 + h1y*(D12*h6x + D22*h6*s*sina + D26*h6y) + (-h1*s*sina + h1x)*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1193] += detJ*wi*(A45*h2*h6 + h2y*(D12*h6x + D22*h6*s*sina + D26*h6y) + (-h2*s*sina + h2x)*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1194] += detJ*wi*(A45*h3*h6 + h3y*(D12*h6x + D22*h6*s*sina + D26*h6y) + (-h3*s*sina + h3x)*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1195] += detJ*wi*(A45*h4*h6 + h4y*(D12*h6x + D22*h6*s*sina + D26*h6y) + (-h4*s*sina + h4x)*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1196] += detJ*wi*(A45*h5*h6 + h5y*(D12*h6x + D22*h6*s*sina + D26*h6y) + (-h5*s*sina + h5x)*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1197] += detJ*wi*(A45*(h6*h6) + h6y*(D12*h6x + D22*h6*s*sina + D26*h6y) + (-h6*s*sina + h6x)*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1198] += detJ*wi*(A45*h6*h7 + h7y*(D12*h6x + D22*h6*s*sina + D26*h6y) + (-h7*s*sina + h7x)*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1199] += detJ*wi*(A45*h6*h8 + h8y*(D12*h6x + D22*h6*s*sina + D26*h6y) + (-h8*s*sina + h8x)*(D16*h6x + D26*h6*s*sina + D66*h6y))
            k0[pos+1200] += detJ*wi*(h1*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h1x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h1y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1201] += detJ*wi*(h2*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h2x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h2y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1202] += detJ*wi*(h3*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h3x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h3y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1203] += detJ*wi*(h4*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h4x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h4y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1204] += detJ*wi*(h5*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h5x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h5y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1205] += detJ*wi*(h6*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h6x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h6y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1206] += detJ*wi*(B11*(h7x*h7x) + 2*B12*h7*h7x*s*sina + 2*B16*h7x*h7y + B22*(h7*h7)*(s*s)*(sina*sina) + 2*B26*h7*h7y*s*sina + B66*(h7y*h7y))
            k0[pos+1207] += detJ*wi*(h8*s*sina*(B12*h7x + B22*h7*s*sina + B26*h7y) + h8x*(B11*h7x + B12*h7*s*sina + B16*h7y) + h8y*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1208] += detJ*wi*(-A45*cosa*h1*h7*s + h1y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h1*s*sina + h1x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1209] += detJ*wi*(-A45*cosa*h2*h7*s + h2y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h2*s*sina + h2x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1210] += detJ*wi*(-A45*cosa*h3*h7*s + h3y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h3*s*sina + h3x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1211] += detJ*wi*(-A45*cosa*h4*h7*s + h4y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h4*s*sina + h4x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1212] += detJ*wi*(-A45*cosa*h5*h7*s + h5y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h5*s*sina + h5x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1213] += detJ*wi*(-A45*cosa*h6*h7*s + h6y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h6*s*sina + h6x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1214] += detJ*wi*(-A45*cosa*(h7*h7)*s + h7y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h7*s*sina + h7x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1215] += detJ*wi*(-A45*cosa*h7*h8*s + h8y*(B12*h7x + B22*h7*s*sina + B26*h7y) + (-h8*s*sina + h8x)*(B16*h7x + B26*h7*s*sina + B66*h7y))
            k0[pos+1216] += detJ*wi*(A45*h1y*h7 + A55*h1x*h7 + cosa*h1*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+1217] += detJ*wi*(A45*h2y*h7 + A55*h2x*h7 + cosa*h2*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+1218] += detJ*wi*(A45*h3y*h7 + A55*h3x*h7 + cosa*h3*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+1219] += detJ*wi*(A45*h4y*h7 + A55*h4x*h7 + cosa*h4*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+1220] += detJ*wi*(A45*h5y*h7 + A55*h5x*h7 + cosa*h5*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+1221] += detJ*wi*(A45*h6y*h7 + A55*h6x*h7 + cosa*h6*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+1222] += detJ*h7*wi*(A45*h7y + A55*h7x + cosa*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+1223] += detJ*wi*(A45*h7*h8y + A55*h7*h8x + cosa*h8*s*(B12*h7x + B22*h7*s*sina + B26*h7y))
            k0[pos+1224] += detJ*wi*(A55*h1*h7 + h1*s*sina*(D12*h7x + D22*h7*s*sina + D26*h7y) + h1x*(D11*h7x + D12*h7*s*sina + D16*h7y) + h1y*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1225] += detJ*wi*(A55*h2*h7 + h2*s*sina*(D12*h7x + D22*h7*s*sina + D26*h7y) + h2x*(D11*h7x + D12*h7*s*sina + D16*h7y) + h2y*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1226] += detJ*wi*(A55*h3*h7 + h3*s*sina*(D12*h7x + D22*h7*s*sina + D26*h7y) + h3x*(D11*h7x + D12*h7*s*sina + D16*h7y) + h3y*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1227] += detJ*wi*(A55*h4*h7 + h4*s*sina*(D12*h7x + D22*h7*s*sina + D26*h7y) + h4x*(D11*h7x + D12*h7*s*sina + D16*h7y) + h4y*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1228] += detJ*wi*(A55*h5*h7 + h5*s*sina*(D12*h7x + D22*h7*s*sina + D26*h7y) + h5x*(D11*h7x + D12*h7*s*sina + D16*h7y) + h5y*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1229] += detJ*wi*(A55*h6*h7 + h6*s*sina*(D12*h7x + D22*h7*s*sina + D26*h7y) + h6x*(D11*h7x + D12*h7*s*sina + D16*h7y) + h6y*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1230] += detJ*wi*(A55*(h7*h7) + D11*(h7x*h7x) + 2*D12*h7*h7x*s*sina + 2*D16*h7x*h7y + D22*(h7*h7)*(s*s)*(sina*sina) + 2*D26*h7*h7y*s*sina + D66*(h7y*h7y))
            k0[pos+1231] += detJ*wi*(A55*h7*h8 + h8*s*sina*(D12*h7x + D22*h7*s*sina + D26*h7y) + h8x*(D11*h7x + D12*h7*s*sina + D16*h7y) + h8y*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1232] += detJ*wi*(A45*h1*h7 + h1y*(D12*h7x + D22*h7*s*sina + D26*h7y) + (-h1*s*sina + h1x)*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1233] += detJ*wi*(A45*h2*h7 + h2y*(D12*h7x + D22*h7*s*sina + D26*h7y) + (-h2*s*sina + h2x)*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1234] += detJ*wi*(A45*h3*h7 + h3y*(D12*h7x + D22*h7*s*sina + D26*h7y) + (-h3*s*sina + h3x)*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1235] += detJ*wi*(A45*h4*h7 + h4y*(D12*h7x + D22*h7*s*sina + D26*h7y) + (-h4*s*sina + h4x)*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1236] += detJ*wi*(A45*h5*h7 + h5y*(D12*h7x + D22*h7*s*sina + D26*h7y) + (-h5*s*sina + h5x)*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1237] += detJ*wi*(A45*h6*h7 + h6y*(D12*h7x + D22*h7*s*sina + D26*h7y) + (-h6*s*sina + h6x)*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1238] += detJ*wi*(A45*(h7*h7) + h7y*(D12*h7x + D22*h7*s*sina + D26*h7y) + (-h7*s*sina + h7x)*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1239] += detJ*wi*(A45*h7*h8 + h8y*(D12*h7x + D22*h7*s*sina + D26*h7y) + (-h8*s*sina + h8x)*(D16*h7x + D26*h7*s*sina + D66*h7y))
            k0[pos+1240] += detJ*wi*(h1*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h1x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h1y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1241] += detJ*wi*(h2*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h2x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h2y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1242] += detJ*wi*(h3*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h3x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h3y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1243] += detJ*wi*(h4*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h4x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h4y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1244] += detJ*wi*(h5*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h5x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h5y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1245] += detJ*wi*(h6*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h6x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h6y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1246] += detJ*wi*(h7*s*sina*(B12*h8x + B22*h8*s*sina + B26*h8y) + h7x*(B11*h8x + B12*h8*s*sina + B16*h8y) + h7y*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1247] += detJ*wi*(B11*(h8x*h8x) + 2*B12*h8*h8x*s*sina + 2*B16*h8x*h8y + B22*(h8*h8)*(s*s)*(sina*sina) + 2*B26*h8*h8y*s*sina + B66*(h8y*h8y))
            k0[pos+1248] += detJ*wi*(-A45*cosa*h1*h8*s + h1y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h1*s*sina + h1x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1249] += detJ*wi*(-A45*cosa*h2*h8*s + h2y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h2*s*sina + h2x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1250] += detJ*wi*(-A45*cosa*h3*h8*s + h3y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h3*s*sina + h3x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1251] += detJ*wi*(-A45*cosa*h4*h8*s + h4y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h4*s*sina + h4x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1252] += detJ*wi*(-A45*cosa*h5*h8*s + h5y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h5*s*sina + h5x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1253] += detJ*wi*(-A45*cosa*h6*h8*s + h6y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h6*s*sina + h6x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1254] += detJ*wi*(-A45*cosa*h7*h8*s + h7y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h7*s*sina + h7x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1255] += detJ*wi*(-A45*cosa*(h8*h8)*s + h8y*(B12*h8x + B22*h8*s*sina + B26*h8y) + (-h8*s*sina + h8x)*(B16*h8x + B26*h8*s*sina + B66*h8y))
            k0[pos+1256] += detJ*wi*(A45*h1y*h8 + A55*h1x*h8 + cosa*h1*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+1257] += detJ*wi*(A45*h2y*h8 + A55*h2x*h8 + cosa*h2*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+1258] += detJ*wi*(A45*h3y*h8 + A55*h3x*h8 + cosa*h3*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+1259] += detJ*wi*(A45*h4y*h8 + A55*h4x*h8 + cosa*h4*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+1260] += detJ*wi*(A45*h5y*h8 + A55*h5x*h8 + cosa*h5*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+1261] += detJ*wi*(A45*h6y*h8 + A55*h6x*h8 + cosa*h6*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+1262] += detJ*wi*(A45*h7y*h8 + A55*h7x*h8 + cosa*h7*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+1263] += detJ*h8*wi*(A45*h8y + A55*h8x + cosa*s*(B12*h8x + B22*h8*s*sina + B26*h8y))
            k0[pos+1264] += detJ*wi*(A55*h1*h8 + h1*s*sina*(D12*h8x + D22*h8*s*sina + D26*h8y) + h1x*(D11*h8x + D12*h8*s*sina + D16*h8y) + h1y*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1265] += detJ*wi*(A55*h2*h8 + h2*s*sina*(D12*h8x + D22*h8*s*sina + D26*h8y) + h2x*(D11*h8x + D12*h8*s*sina + D16*h8y) + h2y*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1266] += detJ*wi*(A55*h3*h8 + h3*s*sina*(D12*h8x + D22*h8*s*sina + D26*h8y) + h3x*(D11*h8x + D12*h8*s*sina + D16*h8y) + h3y*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1267] += detJ*wi*(A55*h4*h8 + h4*s*sina*(D12*h8x + D22*h8*s*sina + D26*h8y) + h4x*(D11*h8x + D12*h8*s*sina + D16*h8y) + h4y*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1268] += detJ*wi*(A55*h5*h8 + h5*s*sina*(D12*h8x + D22*h8*s*sina + D26*h8y) + h5x*(D11*h8x + D12*h8*s*sina + D16*h8y) + h5y*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1269] += detJ*wi*(A55*h6*h8 + h6*s*sina*(D12*h8x + D22*h8*s*sina + D26*h8y) + h6x*(D11*h8x + D12*h8*s*sina + D16*h8y) + h6y*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1270] += detJ*wi*(A55*h7*h8 + h7*s*sina*(D12*h8x + D22*h8*s*sina + D26*h8y) + h7x*(D11*h8x + D12*h8*s*sina + D16*h8y) + h7y*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1271] += detJ*wi*(A55*(h8*h8) + D11*(h8x*h8x) + 2*D12*h8*h8x*s*sina + 2*D16*h8x*h8y + D22*(h8*h8)*(s*s)*(sina*sina) + 2*D26*h8*h8y*s*sina + D66*(h8y*h8y))
            k0[pos+1272] += detJ*wi*(A45*h1*h8 + h1y*(D12*h8x + D22*h8*s*sina + D26*h8y) + (-h1*s*sina + h1x)*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1273] += detJ*wi*(A45*h2*h8 + h2y*(D12*h8x + D22*h8*s*sina + D26*h8y) + (-h2*s*sina + h2x)*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1274] += detJ*wi*(A45*h3*h8 + h3y*(D12*h8x + D22*h8*s*sina + D26*h8y) + (-h3*s*sina + h3x)*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1275] += detJ*wi*(A45*h4*h8 + h4y*(D12*h8x + D22*h8*s*sina + D26*h8y) + (-h4*s*sina + h4x)*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1276] += detJ*wi*(A45*h5*h8 + h5y*(D12*h8x + D22*h8*s*sina + D26*h8y) + (-h5*s*sina + h5x)*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1277] += detJ*wi*(A45*h6*h8 + h6y*(D12*h8x + D22*h8*s*sina + D26*h8y) + (-h6*s*sina + h6x)*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1278] += detJ*wi*(A45*h7*h8 + h7y*(D12*h8x + D22*h8*s*sina + D26*h8y) + (-h7*s*sina + h7x)*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1279] += detJ*wi*(A45*(h8*h8) + h8y*(D12*h8x + D22*h8*s*sina + D26*h8y) + (-h8*s*sina + h8x)*(D16*h8x + D26*h8*s*sina + D66*h8y))
            k0[pos+1280] += detJ*wi*(h1*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h1x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h1y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+1281] += detJ*wi*(h2*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h2x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h2y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+1282] += detJ*wi*(h3*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h3x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h3y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+1283] += detJ*wi*(h4*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h4x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h4y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+1284] += detJ*wi*(h5*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h5x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h5y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+1285] += detJ*wi*(h6*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h6x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h6y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+1286] += detJ*wi*(h7*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h7x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h7y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+1287] += detJ*wi*(h8*s*sina*(B22*h1y + B26*(-h1*s*sina + h1x)) + h8x*(B12*h1y + B16*(-h1*s*sina + h1x)) + h8y*(B26*h1y + B66*(-h1*s*sina + h1x)))
            k0[pos+1288] += detJ*wi*(-A44*cosa*(h1*h1)*s + B22*(h1y*h1y) + 2*B26*h1y*(-h1*s*sina + h1x) + B66*(-h1*s*sina + h1x)**2)
            k0[pos+1289] += detJ*wi*(-A44*cosa*h1*h2*s + h2y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h2*s*sina + h2x))
            k0[pos+1290] += detJ*wi*(-A44*cosa*h1*h3*s + h3y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h3*s*sina + h3x))
            k0[pos+1291] += detJ*wi*(-A44*cosa*h1*h4*s + h4y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h4*s*sina + h4x))
            k0[pos+1292] += detJ*wi*(-A44*cosa*h1*h5*s + h5y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h5*s*sina + h5x))
            k0[pos+1293] += detJ*wi*(-A44*cosa*h1*h6*s + h6y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h6*s*sina + h6x))
            k0[pos+1294] += detJ*wi*(-A44*cosa*h1*h7*s + h7y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h7*s*sina + h7x))
            k0[pos+1295] += detJ*wi*(-A44*cosa*h1*h8*s + h8y*(B22*h1y + B26*(-h1*s*sina + h1x)) + (B26*h1y + B66*(-h1*s*sina + h1x))*(-h8*s*sina + h8x))
            k0[pos+1296] += detJ*h1*wi*(A44*h1y + A45*h1x + cosa*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+1297] += detJ*wi*(A44*h1*h2y + A45*h1*h2x + cosa*h2*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+1298] += detJ*wi*(A44*h1*h3y + A45*h1*h3x + cosa*h3*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+1299] += detJ*wi*(A44*h1*h4y + A45*h1*h4x + cosa*h4*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+1300] += detJ*wi*(A44*h1*h5y + A45*h1*h5x + cosa*h5*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+1301] += detJ*wi*(A44*h1*h6y + A45*h1*h6x + cosa*h6*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+1302] += detJ*wi*(A44*h1*h7y + A45*h1*h7x + cosa*h7*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+1303] += detJ*wi*(A44*h1*h8y + A45*h1*h8x + cosa*h8*s*(B22*h1y - B26*h1*s*sina + B26*h1x))
            k0[pos+1304] += detJ*wi*(A45*(h1*h1) + h1*s*sina*(D22*h1y + D26*(-h1*s*sina + h1x)) + h1x*(D12*h1y + D16*(-h1*s*sina + h1x)) + h1y*(D26*h1y + D66*(-h1*s*sina + h1x)))
            k0[pos+1305] += detJ*wi*(A45*h1*h2 + h2*s*sina*(D22*h1y + D26*(-h1*s*sina + h1x)) + h2x*(D12*h1y + D16*(-h1*s*sina + h1x)) + h2y*(D26*h1y + D66*(-h1*s*sina + h1x)))
            k0[pos+1306] += detJ*wi*(A45*h1*h3 + h3*s*sina*(D22*h1y + D26*(-h1*s*sina + h1x)) + h3x*(D12*h1y + D16*(-h1*s*sina + h1x)) + h3y*(D26*h1y + D66*(-h1*s*sina + h1x)))
            k0[pos+1307] += detJ*wi*(A45*h1*h4 + h4*s*sina*(D22*h1y + D26*(-h1*s*sina + h1x)) + h4x*(D12*h1y + D16*(-h1*s*sina + h1x)) + h4y*(D26*h1y + D66*(-h1*s*sina + h1x)))
            k0[pos+1308] += detJ*wi*(A45*h1*h5 + h5*s*sina*(D22*h1y + D26*(-h1*s*sina + h1x)) + h5x*(D12*h1y + D16*(-h1*s*sina + h1x)) + h5y*(D26*h1y + D66*(-h1*s*sina + h1x)))
            k0[pos+1309] += detJ*wi*(A45*h1*h6 + h6*s*sina*(D22*h1y + D26*(-h1*s*sina + h1x)) + h6x*(D12*h1y + D16*(-h1*s*sina + h1x)) + h6y*(D26*h1y + D66*(-h1*s*sina + h1x)))
            k0[pos+1310] += detJ*wi*(A45*h1*h7 + h7*s*sina*(D22*h1y + D26*(-h1*s*sina + h1x)) + h7x*(D12*h1y + D16*(-h1*s*sina + h1x)) + h7y*(D26*h1y + D66*(-h1*s*sina + h1x)))
            k0[pos+1311] += detJ*wi*(A45*h1*h8 + h8*s*sina*(D22*h1y + D26*(-h1*s*sina + h1x)) + h8x*(D12*h1y + D16*(-h1*s*sina + h1x)) + h8y*(D26*h1y + D66*(-h1*s*sina + h1x)))
            k0[pos+1312] += detJ*wi*(A44*(h1*h1) + D66*(-h1*s*sina + h1x)**2 + h1y*(D22*h1y - 2*D26*h1*s*sina + 2*D26*h1x))
            k0[pos+1313] += detJ*wi*(A44*h1*h2 + h2y*(D22*h1y + D26*(-h1*s*sina + h1x)) + (D26*h1y + D66*(-h1*s*sina + h1x))*(-h2*s*sina + h2x))
            k0[pos+1314] += detJ*wi*(A44*h1*h3 + h3y*(D22*h1y + D26*(-h1*s*sina + h1x)) + (D26*h1y + D66*(-h1*s*sina + h1x))*(-h3*s*sina + h3x))
            k0[pos+1315] += detJ*wi*(A44*h1*h4 + h4y*(D22*h1y + D26*(-h1*s*sina + h1x)) + (D26*h1y + D66*(-h1*s*sina + h1x))*(-h4*s*sina + h4x))
            k0[pos+1316] += detJ*wi*(A44*h1*h5 + h5y*(D22*h1y + D26*(-h1*s*sina + h1x)) + (D26*h1y + D66*(-h1*s*sina + h1x))*(-h5*s*sina + h5x))
            k0[pos+1317] += detJ*wi*(A44*h1*h6 + h6y*(D22*h1y + D26*(-h1*s*sina + h1x)) + (D26*h1y + D66*(-h1*s*sina + h1x))*(-h6*s*sina + h6x))
            k0[pos+1318] += detJ*wi*(A44*h1*h7 + h7y*(D22*h1y + D26*(-h1*s*sina + h1x)) + (D26*h1y + D66*(-h1*s*sina + h1x))*(-h7*s*sina + h7x))
            k0[pos+1319] += detJ*wi*(A44*h1*h8 + h8y*(D22*h1y + D26*(-h1*s*sina + h1x)) + (D26*h1y + D66*(-h1*s*sina + h1x))*(-h8*s*sina + h8x))
            k0[pos+1320] += detJ*wi*(h1*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h1x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h1y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+1321] += detJ*wi*(h2*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h2x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h2y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+1322] += detJ*wi*(h3*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h3x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h3y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+1323] += detJ*wi*(h4*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h4x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h4y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+1324] += detJ*wi*(h5*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h5x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h5y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+1325] += detJ*wi*(h6*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h6x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h6y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+1326] += detJ*wi*(h7*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h7x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h7y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+1327] += detJ*wi*(h8*s*sina*(B22*h2y + B26*(-h2*s*sina + h2x)) + h8x*(B12*h2y + B16*(-h2*s*sina + h2x)) + h8y*(B26*h2y + B66*(-h2*s*sina + h2x)))
            k0[pos+1328] += detJ*wi*(-A44*cosa*h1*h2*s + h1y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h1*s*sina + h1x))
            k0[pos+1329] += detJ*wi*(-A44*cosa*(h2*h2)*s + B22*(h2y*h2y) + 2*B26*h2y*(-h2*s*sina + h2x) + B66*(-h2*s*sina + h2x)**2)
            k0[pos+1330] += detJ*wi*(-A44*cosa*h2*h3*s + h3y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h3*s*sina + h3x))
            k0[pos+1331] += detJ*wi*(-A44*cosa*h2*h4*s + h4y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h4*s*sina + h4x))
            k0[pos+1332] += detJ*wi*(-A44*cosa*h2*h5*s + h5y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h5*s*sina + h5x))
            k0[pos+1333] += detJ*wi*(-A44*cosa*h2*h6*s + h6y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h6*s*sina + h6x))
            k0[pos+1334] += detJ*wi*(-A44*cosa*h2*h7*s + h7y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h7*s*sina + h7x))
            k0[pos+1335] += detJ*wi*(-A44*cosa*h2*h8*s + h8y*(B22*h2y + B26*(-h2*s*sina + h2x)) + (B26*h2y + B66*(-h2*s*sina + h2x))*(-h8*s*sina + h8x))
            k0[pos+1336] += detJ*wi*(A44*h1y*h2 + A45*h1x*h2 + cosa*h1*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+1337] += detJ*h2*wi*(A44*h2y + A45*h2x + cosa*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+1338] += detJ*wi*(A44*h2*h3y + A45*h2*h3x + cosa*h3*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+1339] += detJ*wi*(A44*h2*h4y + A45*h2*h4x + cosa*h4*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+1340] += detJ*wi*(A44*h2*h5y + A45*h2*h5x + cosa*h5*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+1341] += detJ*wi*(A44*h2*h6y + A45*h2*h6x + cosa*h6*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+1342] += detJ*wi*(A44*h2*h7y + A45*h2*h7x + cosa*h7*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+1343] += detJ*wi*(A44*h2*h8y + A45*h2*h8x + cosa*h8*s*(B22*h2y - B26*h2*s*sina + B26*h2x))
            k0[pos+1344] += detJ*wi*(A45*h1*h2 + h1*s*sina*(D22*h2y + D26*(-h2*s*sina + h2x)) + h1x*(D12*h2y + D16*(-h2*s*sina + h2x)) + h1y*(D26*h2y + D66*(-h2*s*sina + h2x)))
            k0[pos+1345] += detJ*wi*(A45*(h2*h2) + h2*s*sina*(D22*h2y + D26*(-h2*s*sina + h2x)) + h2x*(D12*h2y + D16*(-h2*s*sina + h2x)) + h2y*(D26*h2y + D66*(-h2*s*sina + h2x)))
            k0[pos+1346] += detJ*wi*(A45*h2*h3 + h3*s*sina*(D22*h2y + D26*(-h2*s*sina + h2x)) + h3x*(D12*h2y + D16*(-h2*s*sina + h2x)) + h3y*(D26*h2y + D66*(-h2*s*sina + h2x)))
            k0[pos+1347] += detJ*wi*(A45*h2*h4 + h4*s*sina*(D22*h2y + D26*(-h2*s*sina + h2x)) + h4x*(D12*h2y + D16*(-h2*s*sina + h2x)) + h4y*(D26*h2y + D66*(-h2*s*sina + h2x)))
            k0[pos+1348] += detJ*wi*(A45*h2*h5 + h5*s*sina*(D22*h2y + D26*(-h2*s*sina + h2x)) + h5x*(D12*h2y + D16*(-h2*s*sina + h2x)) + h5y*(D26*h2y + D66*(-h2*s*sina + h2x)))
            k0[pos+1349] += detJ*wi*(A45*h2*h6 + h6*s*sina*(D22*h2y + D26*(-h2*s*sina + h2x)) + h6x*(D12*h2y + D16*(-h2*s*sina + h2x)) + h6y*(D26*h2y + D66*(-h2*s*sina + h2x)))
            k0[pos+1350] += detJ*wi*(A45*h2*h7 + h7*s*sina*(D22*h2y + D26*(-h2*s*sina + h2x)) + h7x*(D12*h2y + D16*(-h2*s*sina + h2x)) + h7y*(D26*h2y + D66*(-h2*s*sina + h2x)))
            k0[pos+1351] += detJ*wi*(A45*h2*h8 + h8*s*sina*(D22*h2y + D26*(-h2*s*sina + h2x)) + h8x*(D12*h2y + D16*(-h2*s*sina + h2x)) + h8y*(D26*h2y + D66*(-h2*s*sina + h2x)))
            k0[pos+1352] += detJ*wi*(A44*h1*h2 + h1y*(D22*h2y + D26*(-h2*s*sina + h2x)) + (D26*h2y + D66*(-h2*s*sina + h2x))*(-h1*s*sina + h1x))
            k0[pos+1353] += detJ*wi*(A44*(h2*h2) + D66*(-h2*s*sina + h2x)**2 + h2y*(D22*h2y - 2*D26*h2*s*sina + 2*D26*h2x))
            k0[pos+1354] += detJ*wi*(A44*h2*h3 + h3y*(D22*h2y + D26*(-h2*s*sina + h2x)) + (D26*h2y + D66*(-h2*s*sina + h2x))*(-h3*s*sina + h3x))
            k0[pos+1355] += detJ*wi*(A44*h2*h4 + h4y*(D22*h2y + D26*(-h2*s*sina + h2x)) + (D26*h2y + D66*(-h2*s*sina + h2x))*(-h4*s*sina + h4x))
            k0[pos+1356] += detJ*wi*(A44*h2*h5 + h5y*(D22*h2y + D26*(-h2*s*sina + h2x)) + (D26*h2y + D66*(-h2*s*sina + h2x))*(-h5*s*sina + h5x))
            k0[pos+1357] += detJ*wi*(A44*h2*h6 + h6y*(D22*h2y + D26*(-h2*s*sina + h2x)) + (D26*h2y + D66*(-h2*s*sina + h2x))*(-h6*s*sina + h6x))
            k0[pos+1358] += detJ*wi*(A44*h2*h7 + h7y*(D22*h2y + D26*(-h2*s*sina + h2x)) + (D26*h2y + D66*(-h2*s*sina + h2x))*(-h7*s*sina + h7x))
            k0[pos+1359] += detJ*wi*(A44*h2*h8 + h8y*(D22*h2y + D26*(-h2*s*sina + h2x)) + (D26*h2y + D66*(-h2*s*sina + h2x))*(-h8*s*sina + h8x))
            k0[pos+1360] += detJ*wi*(h1*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h1x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h1y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+1361] += detJ*wi*(h2*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h2x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h2y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+1362] += detJ*wi*(h3*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h3x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h3y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+1363] += detJ*wi*(h4*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h4x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h4y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+1364] += detJ*wi*(h5*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h5x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h5y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+1365] += detJ*wi*(h6*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h6x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h6y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+1366] += detJ*wi*(h7*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h7x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h7y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+1367] += detJ*wi*(h8*s*sina*(B22*h3y + B26*(-h3*s*sina + h3x)) + h8x*(B12*h3y + B16*(-h3*s*sina + h3x)) + h8y*(B26*h3y + B66*(-h3*s*sina + h3x)))
            k0[pos+1368] += detJ*wi*(-A44*cosa*h1*h3*s + h1y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h1*s*sina + h1x))
            k0[pos+1369] += detJ*wi*(-A44*cosa*h2*h3*s + h2y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h2*s*sina + h2x))
            k0[pos+1370] += detJ*wi*(-A44*cosa*(h3*h3)*s + B22*(h3y*h3y) + 2*B26*h3y*(-h3*s*sina + h3x) + B66*(-h3*s*sina + h3x)**2)
            k0[pos+1371] += detJ*wi*(-A44*cosa*h3*h4*s + h4y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h4*s*sina + h4x))
            k0[pos+1372] += detJ*wi*(-A44*cosa*h3*h5*s + h5y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h5*s*sina + h5x))
            k0[pos+1373] += detJ*wi*(-A44*cosa*h3*h6*s + h6y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h6*s*sina + h6x))
            k0[pos+1374] += detJ*wi*(-A44*cosa*h3*h7*s + h7y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h7*s*sina + h7x))
            k0[pos+1375] += detJ*wi*(-A44*cosa*h3*h8*s + h8y*(B22*h3y + B26*(-h3*s*sina + h3x)) + (B26*h3y + B66*(-h3*s*sina + h3x))*(-h8*s*sina + h8x))
            k0[pos+1376] += detJ*wi*(A44*h1y*h3 + A45*h1x*h3 + cosa*h1*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+1377] += detJ*wi*(A44*h2y*h3 + A45*h2x*h3 + cosa*h2*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+1378] += detJ*h3*wi*(A44*h3y + A45*h3x + cosa*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+1379] += detJ*wi*(A44*h3*h4y + A45*h3*h4x + cosa*h4*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+1380] += detJ*wi*(A44*h3*h5y + A45*h3*h5x + cosa*h5*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+1381] += detJ*wi*(A44*h3*h6y + A45*h3*h6x + cosa*h6*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+1382] += detJ*wi*(A44*h3*h7y + A45*h3*h7x + cosa*h7*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+1383] += detJ*wi*(A44*h3*h8y + A45*h3*h8x + cosa*h8*s*(B22*h3y - B26*h3*s*sina + B26*h3x))
            k0[pos+1384] += detJ*wi*(A45*h1*h3 + h1*s*sina*(D22*h3y + D26*(-h3*s*sina + h3x)) + h1x*(D12*h3y + D16*(-h3*s*sina + h3x)) + h1y*(D26*h3y + D66*(-h3*s*sina + h3x)))
            k0[pos+1385] += detJ*wi*(A45*h2*h3 + h2*s*sina*(D22*h3y + D26*(-h3*s*sina + h3x)) + h2x*(D12*h3y + D16*(-h3*s*sina + h3x)) + h2y*(D26*h3y + D66*(-h3*s*sina + h3x)))
            k0[pos+1386] += detJ*wi*(A45*(h3*h3) + h3*s*sina*(D22*h3y + D26*(-h3*s*sina + h3x)) + h3x*(D12*h3y + D16*(-h3*s*sina + h3x)) + h3y*(D26*h3y + D66*(-h3*s*sina + h3x)))
            k0[pos+1387] += detJ*wi*(A45*h3*h4 + h4*s*sina*(D22*h3y + D26*(-h3*s*sina + h3x)) + h4x*(D12*h3y + D16*(-h3*s*sina + h3x)) + h4y*(D26*h3y + D66*(-h3*s*sina + h3x)))
            k0[pos+1388] += detJ*wi*(A45*h3*h5 + h5*s*sina*(D22*h3y + D26*(-h3*s*sina + h3x)) + h5x*(D12*h3y + D16*(-h3*s*sina + h3x)) + h5y*(D26*h3y + D66*(-h3*s*sina + h3x)))
            k0[pos+1389] += detJ*wi*(A45*h3*h6 + h6*s*sina*(D22*h3y + D26*(-h3*s*sina + h3x)) + h6x*(D12*h3y + D16*(-h3*s*sina + h3x)) + h6y*(D26*h3y + D66*(-h3*s*sina + h3x)))
            k0[pos+1390] += detJ*wi*(A45*h3*h7 + h7*s*sina*(D22*h3y + D26*(-h3*s*sina + h3x)) + h7x*(D12*h3y + D16*(-h3*s*sina + h3x)) + h7y*(D26*h3y + D66*(-h3*s*sina + h3x)))
            k0[pos+1391] += detJ*wi*(A45*h3*h8 + h8*s*sina*(D22*h3y + D26*(-h3*s*sina + h3x)) + h8x*(D12*h3y + D16*(-h3*s*sina + h3x)) + h8y*(D26*h3y + D66*(-h3*s*sina + h3x)))
            k0[pos+1392] += detJ*wi*(A44*h1*h3 + h1y*(D22*h3y + D26*(-h3*s*sina + h3x)) + (D26*h3y + D66*(-h3*s*sina + h3x))*(-h1*s*sina + h1x))
            k0[pos+1393] += detJ*wi*(A44*h2*h3 + h2y*(D22*h3y + D26*(-h3*s*sina + h3x)) + (D26*h3y + D66*(-h3*s*sina + h3x))*(-h2*s*sina + h2x))
            k0[pos+1394] += detJ*wi*(A44*(h3*h3) + D66*(-h3*s*sina + h3x)**2 + h3y*(D22*h3y - 2*D26*h3*s*sina + 2*D26*h3x))
            k0[pos+1395] += detJ*wi*(A44*h3*h4 + h4y*(D22*h3y + D26*(-h3*s*sina + h3x)) + (D26*h3y + D66*(-h3*s*sina + h3x))*(-h4*s*sina + h4x))
            k0[pos+1396] += detJ*wi*(A44*h3*h5 + h5y*(D22*h3y + D26*(-h3*s*sina + h3x)) + (D26*h3y + D66*(-h3*s*sina + h3x))*(-h5*s*sina + h5x))
            k0[pos+1397] += detJ*wi*(A44*h3*h6 + h6y*(D22*h3y + D26*(-h3*s*sina + h3x)) + (D26*h3y + D66*(-h3*s*sina + h3x))*(-h6*s*sina + h6x))
            k0[pos+1398] += detJ*wi*(A44*h3*h7 + h7y*(D22*h3y + D26*(-h3*s*sina + h3x)) + (D26*h3y + D66*(-h3*s*sina + h3x))*(-h7*s*sina + h7x))
            k0[pos+1399] += detJ*wi*(A44*h3*h8 + h8y*(D22*h3y + D26*(-h3*s*sina + h3x)) + (D26*h3y + D66*(-h3*s*sina + h3x))*(-h8*s*sina + h8x))
            k0[pos+1400] += detJ*wi*(h1*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h1x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h1y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+1401] += detJ*wi*(h2*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h2x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h2y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+1402] += detJ*wi*(h3*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h3x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h3y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+1403] += detJ*wi*(h4*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h4x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h4y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+1404] += detJ*wi*(h5*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h5x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h5y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+1405] += detJ*wi*(h6*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h6x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h6y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+1406] += detJ*wi*(h7*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h7x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h7y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+1407] += detJ*wi*(h8*s*sina*(B22*h4y + B26*(-h4*s*sina + h4x)) + h8x*(B12*h4y + B16*(-h4*s*sina + h4x)) + h8y*(B26*h4y + B66*(-h4*s*sina + h4x)))
            k0[pos+1408] += detJ*wi*(-A44*cosa*h1*h4*s + h1y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h1*s*sina + h1x))
            k0[pos+1409] += detJ*wi*(-A44*cosa*h2*h4*s + h2y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h2*s*sina + h2x))
            k0[pos+1410] += detJ*wi*(-A44*cosa*h3*h4*s + h3y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h3*s*sina + h3x))
            k0[pos+1411] += detJ*wi*(-A44*cosa*(h4*h4)*s + B22*(h4y*h4y) + 2*B26*h4y*(-h4*s*sina + h4x) + B66*(-h4*s*sina + h4x)**2)
            k0[pos+1412] += detJ*wi*(-A44*cosa*h4*h5*s + h5y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h5*s*sina + h5x))
            k0[pos+1413] += detJ*wi*(-A44*cosa*h4*h6*s + h6y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h6*s*sina + h6x))
            k0[pos+1414] += detJ*wi*(-A44*cosa*h4*h7*s + h7y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h7*s*sina + h7x))
            k0[pos+1415] += detJ*wi*(-A44*cosa*h4*h8*s + h8y*(B22*h4y + B26*(-h4*s*sina + h4x)) + (B26*h4y + B66*(-h4*s*sina + h4x))*(-h8*s*sina + h8x))
            k0[pos+1416] += detJ*wi*(A44*h1y*h4 + A45*h1x*h4 + cosa*h1*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+1417] += detJ*wi*(A44*h2y*h4 + A45*h2x*h4 + cosa*h2*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+1418] += detJ*wi*(A44*h3y*h4 + A45*h3x*h4 + cosa*h3*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+1419] += detJ*h4*wi*(A44*h4y + A45*h4x + cosa*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+1420] += detJ*wi*(A44*h4*h5y + A45*h4*h5x + cosa*h5*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+1421] += detJ*wi*(A44*h4*h6y + A45*h4*h6x + cosa*h6*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+1422] += detJ*wi*(A44*h4*h7y + A45*h4*h7x + cosa*h7*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+1423] += detJ*wi*(A44*h4*h8y + A45*h4*h8x + cosa*h8*s*(B22*h4y - B26*h4*s*sina + B26*h4x))
            k0[pos+1424] += detJ*wi*(A45*h1*h4 + h1*s*sina*(D22*h4y + D26*(-h4*s*sina + h4x)) + h1x*(D12*h4y + D16*(-h4*s*sina + h4x)) + h1y*(D26*h4y + D66*(-h4*s*sina + h4x)))
            k0[pos+1425] += detJ*wi*(A45*h2*h4 + h2*s*sina*(D22*h4y + D26*(-h4*s*sina + h4x)) + h2x*(D12*h4y + D16*(-h4*s*sina + h4x)) + h2y*(D26*h4y + D66*(-h4*s*sina + h4x)))
            k0[pos+1426] += detJ*wi*(A45*h3*h4 + h3*s*sina*(D22*h4y + D26*(-h4*s*sina + h4x)) + h3x*(D12*h4y + D16*(-h4*s*sina + h4x)) + h3y*(D26*h4y + D66*(-h4*s*sina + h4x)))
            k0[pos+1427] += detJ*wi*(A45*(h4*h4) + h4*s*sina*(D22*h4y + D26*(-h4*s*sina + h4x)) + h4x*(D12*h4y + D16*(-h4*s*sina + h4x)) + h4y*(D26*h4y + D66*(-h4*s*sina + h4x)))
            k0[pos+1428] += detJ*wi*(A45*h4*h5 + h5*s*sina*(D22*h4y + D26*(-h4*s*sina + h4x)) + h5x*(D12*h4y + D16*(-h4*s*sina + h4x)) + h5y*(D26*h4y + D66*(-h4*s*sina + h4x)))
            k0[pos+1429] += detJ*wi*(A45*h4*h6 + h6*s*sina*(D22*h4y + D26*(-h4*s*sina + h4x)) + h6x*(D12*h4y + D16*(-h4*s*sina + h4x)) + h6y*(D26*h4y + D66*(-h4*s*sina + h4x)))
            k0[pos+1430] += detJ*wi*(A45*h4*h7 + h7*s*sina*(D22*h4y + D26*(-h4*s*sina + h4x)) + h7x*(D12*h4y + D16*(-h4*s*sina + h4x)) + h7y*(D26*h4y + D66*(-h4*s*sina + h4x)))
            k0[pos+1431] += detJ*wi*(A45*h4*h8 + h8*s*sina*(D22*h4y + D26*(-h4*s*sina + h4x)) + h8x*(D12*h4y + D16*(-h4*s*sina + h4x)) + h8y*(D26*h4y + D66*(-h4*s*sina + h4x)))
            k0[pos+1432] += detJ*wi*(A44*h1*h4 + h1y*(D22*h4y + D26*(-h4*s*sina + h4x)) + (D26*h4y + D66*(-h4*s*sina + h4x))*(-h1*s*sina + h1x))
            k0[pos+1433] += detJ*wi*(A44*h2*h4 + h2y*(D22*h4y + D26*(-h4*s*sina + h4x)) + (D26*h4y + D66*(-h4*s*sina + h4x))*(-h2*s*sina + h2x))
            k0[pos+1434] += detJ*wi*(A44*h3*h4 + h3y*(D22*h4y + D26*(-h4*s*sina + h4x)) + (D26*h4y + D66*(-h4*s*sina + h4x))*(-h3*s*sina + h3x))
            k0[pos+1435] += detJ*wi*(A44*(h4*h4) + D66*(-h4*s*sina + h4x)**2 + h4y*(D22*h4y - 2*D26*h4*s*sina + 2*D26*h4x))
            k0[pos+1436] += detJ*wi*(A44*h4*h5 + h5y*(D22*h4y + D26*(-h4*s*sina + h4x)) + (D26*h4y + D66*(-h4*s*sina + h4x))*(-h5*s*sina + h5x))
            k0[pos+1437] += detJ*wi*(A44*h4*h6 + h6y*(D22*h4y + D26*(-h4*s*sina + h4x)) + (D26*h4y + D66*(-h4*s*sina + h4x))*(-h6*s*sina + h6x))
            k0[pos+1438] += detJ*wi*(A44*h4*h7 + h7y*(D22*h4y + D26*(-h4*s*sina + h4x)) + (D26*h4y + D66*(-h4*s*sina + h4x))*(-h7*s*sina + h7x))
            k0[pos+1439] += detJ*wi*(A44*h4*h8 + h8y*(D22*h4y + D26*(-h4*s*sina + h4x)) + (D26*h4y + D66*(-h4*s*sina + h4x))*(-h8*s*sina + h8x))
            k0[pos+1440] += detJ*wi*(h1*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h1x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h1y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+1441] += detJ*wi*(h2*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h2x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h2y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+1442] += detJ*wi*(h3*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h3x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h3y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+1443] += detJ*wi*(h4*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h4x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h4y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+1444] += detJ*wi*(h5*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h5x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h5y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+1445] += detJ*wi*(h6*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h6x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h6y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+1446] += detJ*wi*(h7*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h7x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h7y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+1447] += detJ*wi*(h8*s*sina*(B22*h5y + B26*(-h5*s*sina + h5x)) + h8x*(B12*h5y + B16*(-h5*s*sina + h5x)) + h8y*(B26*h5y + B66*(-h5*s*sina + h5x)))
            k0[pos+1448] += detJ*wi*(-A44*cosa*h1*h5*s + h1y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h1*s*sina + h1x))
            k0[pos+1449] += detJ*wi*(-A44*cosa*h2*h5*s + h2y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h2*s*sina + h2x))
            k0[pos+1450] += detJ*wi*(-A44*cosa*h3*h5*s + h3y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h3*s*sina + h3x))
            k0[pos+1451] += detJ*wi*(-A44*cosa*h4*h5*s + h4y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h4*s*sina + h4x))
            k0[pos+1452] += detJ*wi*(-A44*cosa*(h5*h5)*s + B22*(h5y*h5y) + 2*B26*h5y*(-h5*s*sina + h5x) + B66*(-h5*s*sina + h5x)**2)
            k0[pos+1453] += detJ*wi*(-A44*cosa*h5*h6*s + h6y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h6*s*sina + h6x))
            k0[pos+1454] += detJ*wi*(-A44*cosa*h5*h7*s + h7y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h7*s*sina + h7x))
            k0[pos+1455] += detJ*wi*(-A44*cosa*h5*h8*s + h8y*(B22*h5y + B26*(-h5*s*sina + h5x)) + (B26*h5y + B66*(-h5*s*sina + h5x))*(-h8*s*sina + h8x))
            k0[pos+1456] += detJ*wi*(A44*h1y*h5 + A45*h1x*h5 + cosa*h1*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+1457] += detJ*wi*(A44*h2y*h5 + A45*h2x*h5 + cosa*h2*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+1458] += detJ*wi*(A44*h3y*h5 + A45*h3x*h5 + cosa*h3*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+1459] += detJ*wi*(A44*h4y*h5 + A45*h4x*h5 + cosa*h4*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+1460] += detJ*h5*wi*(A44*h5y + A45*h5x + cosa*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+1461] += detJ*wi*(A44*h5*h6y + A45*h5*h6x + cosa*h6*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+1462] += detJ*wi*(A44*h5*h7y + A45*h5*h7x + cosa*h7*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+1463] += detJ*wi*(A44*h5*h8y + A45*h5*h8x + cosa*h8*s*(B22*h5y - B26*h5*s*sina + B26*h5x))
            k0[pos+1464] += detJ*wi*(A45*h1*h5 + h1*s*sina*(D22*h5y + D26*(-h5*s*sina + h5x)) + h1x*(D12*h5y + D16*(-h5*s*sina + h5x)) + h1y*(D26*h5y + D66*(-h5*s*sina + h5x)))
            k0[pos+1465] += detJ*wi*(A45*h2*h5 + h2*s*sina*(D22*h5y + D26*(-h5*s*sina + h5x)) + h2x*(D12*h5y + D16*(-h5*s*sina + h5x)) + h2y*(D26*h5y + D66*(-h5*s*sina + h5x)))
            k0[pos+1466] += detJ*wi*(A45*h3*h5 + h3*s*sina*(D22*h5y + D26*(-h5*s*sina + h5x)) + h3x*(D12*h5y + D16*(-h5*s*sina + h5x)) + h3y*(D26*h5y + D66*(-h5*s*sina + h5x)))
            k0[pos+1467] += detJ*wi*(A45*h4*h5 + h4*s*sina*(D22*h5y + D26*(-h5*s*sina + h5x)) + h4x*(D12*h5y + D16*(-h5*s*sina + h5x)) + h4y*(D26*h5y + D66*(-h5*s*sina + h5x)))
            k0[pos+1468] += detJ*wi*(A45*(h5*h5) + h5*s*sina*(D22*h5y + D26*(-h5*s*sina + h5x)) + h5x*(D12*h5y + D16*(-h5*s*sina + h5x)) + h5y*(D26*h5y + D66*(-h5*s*sina + h5x)))
            k0[pos+1469] += detJ*wi*(A45*h5*h6 + h6*s*sina*(D22*h5y + D26*(-h5*s*sina + h5x)) + h6x*(D12*h5y + D16*(-h5*s*sina + h5x)) + h6y*(D26*h5y + D66*(-h5*s*sina + h5x)))
            k0[pos+1470] += detJ*wi*(A45*h5*h7 + h7*s*sina*(D22*h5y + D26*(-h5*s*sina + h5x)) + h7x*(D12*h5y + D16*(-h5*s*sina + h5x)) + h7y*(D26*h5y + D66*(-h5*s*sina + h5x)))
            k0[pos+1471] += detJ*wi*(A45*h5*h8 + h8*s*sina*(D22*h5y + D26*(-h5*s*sina + h5x)) + h8x*(D12*h5y + D16*(-h5*s*sina + h5x)) + h8y*(D26*h5y + D66*(-h5*s*sina + h5x)))
            k0[pos+1472] += detJ*wi*(A44*h1*h5 + h1y*(D22*h5y + D26*(-h5*s*sina + h5x)) + (D26*h5y + D66*(-h5*s*sina + h5x))*(-h1*s*sina + h1x))
            k0[pos+1473] += detJ*wi*(A44*h2*h5 + h2y*(D22*h5y + D26*(-h5*s*sina + h5x)) + (D26*h5y + D66*(-h5*s*sina + h5x))*(-h2*s*sina + h2x))
            k0[pos+1474] += detJ*wi*(A44*h3*h5 + h3y*(D22*h5y + D26*(-h5*s*sina + h5x)) + (D26*h5y + D66*(-h5*s*sina + h5x))*(-h3*s*sina + h3x))
            k0[pos+1475] += detJ*wi*(A44*h4*h5 + h4y*(D22*h5y + D26*(-h5*s*sina + h5x)) + (D26*h5y + D66*(-h5*s*sina + h5x))*(-h4*s*sina + h4x))
            k0[pos+1476] += detJ*wi*(A44*(h5*h5) + D66*(-h5*s*sina + h5x)**2 + h5y*(D22*h5y - 2*D26*h5*s*sina + 2*D26*h5x))
            k0[pos+1477] += detJ*wi*(A44*h5*h6 + h6y*(D22*h5y + D26*(-h5*s*sina + h5x)) + (D26*h5y + D66*(-h5*s*sina + h5x))*(-h6*s*sina + h6x))
            k0[pos+1478] += detJ*wi*(A44*h5*h7 + h7y*(D22*h5y + D26*(-h5*s*sina + h5x)) + (D26*h5y + D66*(-h5*s*sina + h5x))*(-h7*s*sina + h7x))
            k0[pos+1479] += detJ*wi*(A44*h5*h8 + h8y*(D22*h5y + D26*(-h5*s*sina + h5x)) + (D26*h5y + D66*(-h5*s*sina + h5x))*(-h8*s*sina + h8x))
            k0[pos+1480] += detJ*wi*(h1*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h1x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h1y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+1481] += detJ*wi*(h2*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h2x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h2y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+1482] += detJ*wi*(h3*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h3x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h3y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+1483] += detJ*wi*(h4*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h4x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h4y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+1484] += detJ*wi*(h5*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h5x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h5y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+1485] += detJ*wi*(h6*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h6x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h6y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+1486] += detJ*wi*(h7*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h7x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h7y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+1487] += detJ*wi*(h8*s*sina*(B22*h6y + B26*(-h6*s*sina + h6x)) + h8x*(B12*h6y + B16*(-h6*s*sina + h6x)) + h8y*(B26*h6y + B66*(-h6*s*sina + h6x)))
            k0[pos+1488] += detJ*wi*(-A44*cosa*h1*h6*s + h1y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h1*s*sina + h1x))
            k0[pos+1489] += detJ*wi*(-A44*cosa*h2*h6*s + h2y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h2*s*sina + h2x))
            k0[pos+1490] += detJ*wi*(-A44*cosa*h3*h6*s + h3y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h3*s*sina + h3x))
            k0[pos+1491] += detJ*wi*(-A44*cosa*h4*h6*s + h4y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h4*s*sina + h4x))
            k0[pos+1492] += detJ*wi*(-A44*cosa*h5*h6*s + h5y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h5*s*sina + h5x))
            k0[pos+1493] += detJ*wi*(-A44*cosa*(h6*h6)*s + B22*(h6y*h6y) + 2*B26*h6y*(-h6*s*sina + h6x) + B66*(-h6*s*sina + h6x)**2)
            k0[pos+1494] += detJ*wi*(-A44*cosa*h6*h7*s + h7y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h7*s*sina + h7x))
            k0[pos+1495] += detJ*wi*(-A44*cosa*h6*h8*s + h8y*(B22*h6y + B26*(-h6*s*sina + h6x)) + (B26*h6y + B66*(-h6*s*sina + h6x))*(-h8*s*sina + h8x))
            k0[pos+1496] += detJ*wi*(A44*h1y*h6 + A45*h1x*h6 + cosa*h1*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+1497] += detJ*wi*(A44*h2y*h6 + A45*h2x*h6 + cosa*h2*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+1498] += detJ*wi*(A44*h3y*h6 + A45*h3x*h6 + cosa*h3*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+1499] += detJ*wi*(A44*h4y*h6 + A45*h4x*h6 + cosa*h4*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+1500] += detJ*wi*(A44*h5y*h6 + A45*h5x*h6 + cosa*h5*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+1501] += detJ*h6*wi*(A44*h6y + A45*h6x + cosa*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+1502] += detJ*wi*(A44*h6*h7y + A45*h6*h7x + cosa*h7*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+1503] += detJ*wi*(A44*h6*h8y + A45*h6*h8x + cosa*h8*s*(B22*h6y - B26*h6*s*sina + B26*h6x))
            k0[pos+1504] += detJ*wi*(A45*h1*h6 + h1*s*sina*(D22*h6y + D26*(-h6*s*sina + h6x)) + h1x*(D12*h6y + D16*(-h6*s*sina + h6x)) + h1y*(D26*h6y + D66*(-h6*s*sina + h6x)))
            k0[pos+1505] += detJ*wi*(A45*h2*h6 + h2*s*sina*(D22*h6y + D26*(-h6*s*sina + h6x)) + h2x*(D12*h6y + D16*(-h6*s*sina + h6x)) + h2y*(D26*h6y + D66*(-h6*s*sina + h6x)))
            k0[pos+1506] += detJ*wi*(A45*h3*h6 + h3*s*sina*(D22*h6y + D26*(-h6*s*sina + h6x)) + h3x*(D12*h6y + D16*(-h6*s*sina + h6x)) + h3y*(D26*h6y + D66*(-h6*s*sina + h6x)))
            k0[pos+1507] += detJ*wi*(A45*h4*h6 + h4*s*sina*(D22*h6y + D26*(-h6*s*sina + h6x)) + h4x*(D12*h6y + D16*(-h6*s*sina + h6x)) + h4y*(D26*h6y + D66*(-h6*s*sina + h6x)))
            k0[pos+1508] += detJ*wi*(A45*h5*h6 + h5*s*sina*(D22*h6y + D26*(-h6*s*sina + h6x)) + h5x*(D12*h6y + D16*(-h6*s*sina + h6x)) + h5y*(D26*h6y + D66*(-h6*s*sina + h6x)))
            k0[pos+1509] += detJ*wi*(A45*(h6*h6) + h6*s*sina*(D22*h6y + D26*(-h6*s*sina + h6x)) + h6x*(D12*h6y + D16*(-h6*s*sina + h6x)) + h6y*(D26*h6y + D66*(-h6*s*sina + h6x)))
            k0[pos+1510] += detJ*wi*(A45*h6*h7 + h7*s*sina*(D22*h6y + D26*(-h6*s*sina + h6x)) + h7x*(D12*h6y + D16*(-h6*s*sina + h6x)) + h7y*(D26*h6y + D66*(-h6*s*sina + h6x)))
            k0[pos+1511] += detJ*wi*(A45*h6*h8 + h8*s*sina*(D22*h6y + D26*(-h6*s*sina + h6x)) + h8x*(D12*h6y + D16*(-h6*s*sina + h6x)) + h8y*(D26*h6y + D66*(-h6*s*sina + h6x)))
            k0[pos+1512] += detJ*wi*(A44*h1*h6 + h1y*(D22*h6y + D26*(-h6*s*sina + h6x)) + (D26*h6y + D66*(-h6*s*sina + h6x))*(-h1*s*sina + h1x))
            k0[pos+1513] += detJ*wi*(A44*h2*h6 + h2y*(D22*h6y + D26*(-h6*s*sina + h6x)) + (D26*h6y + D66*(-h6*s*sina + h6x))*(-h2*s*sina + h2x))
            k0[pos+1514] += detJ*wi*(A44*h3*h6 + h3y*(D22*h6y + D26*(-h6*s*sina + h6x)) + (D26*h6y + D66*(-h6*s*sina + h6x))*(-h3*s*sina + h3x))
            k0[pos+1515] += detJ*wi*(A44*h4*h6 + h4y*(D22*h6y + D26*(-h6*s*sina + h6x)) + (D26*h6y + D66*(-h6*s*sina + h6x))*(-h4*s*sina + h4x))
            k0[pos+1516] += detJ*wi*(A44*h5*h6 + h5y*(D22*h6y + D26*(-h6*s*sina + h6x)) + (D26*h6y + D66*(-h6*s*sina + h6x))*(-h5*s*sina + h5x))
            k0[pos+1517] += detJ*wi*(A44*(h6*h6) + D66*(-h6*s*sina + h6x)**2 + h6y*(D22*h6y - 2*D26*h6*s*sina + 2*D26*h6x))
            k0[pos+1518] += detJ*wi*(A44*h6*h7 + h7y*(D22*h6y + D26*(-h6*s*sina + h6x)) + (D26*h6y + D66*(-h6*s*sina + h6x))*(-h7*s*sina + h7x))
            k0[pos+1519] += detJ*wi*(A44*h6*h8 + h8y*(D22*h6y + D26*(-h6*s*sina + h6x)) + (D26*h6y + D66*(-h6*s*sina + h6x))*(-h8*s*sina + h8x))
            k0[pos+1520] += detJ*wi*(h1*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h1x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h1y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+1521] += detJ*wi*(h2*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h2x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h2y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+1522] += detJ*wi*(h3*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h3x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h3y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+1523] += detJ*wi*(h4*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h4x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h4y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+1524] += detJ*wi*(h5*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h5x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h5y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+1525] += detJ*wi*(h6*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h6x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h6y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+1526] += detJ*wi*(h7*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h7x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h7y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+1527] += detJ*wi*(h8*s*sina*(B22*h7y + B26*(-h7*s*sina + h7x)) + h8x*(B12*h7y + B16*(-h7*s*sina + h7x)) + h8y*(B26*h7y + B66*(-h7*s*sina + h7x)))
            k0[pos+1528] += detJ*wi*(-A44*cosa*h1*h7*s + h1y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h1*s*sina + h1x))
            k0[pos+1529] += detJ*wi*(-A44*cosa*h2*h7*s + h2y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h2*s*sina + h2x))
            k0[pos+1530] += detJ*wi*(-A44*cosa*h3*h7*s + h3y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h3*s*sina + h3x))
            k0[pos+1531] += detJ*wi*(-A44*cosa*h4*h7*s + h4y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h4*s*sina + h4x))
            k0[pos+1532] += detJ*wi*(-A44*cosa*h5*h7*s + h5y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h5*s*sina + h5x))
            k0[pos+1533] += detJ*wi*(-A44*cosa*h6*h7*s + h6y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h6*s*sina + h6x))
            k0[pos+1534] += detJ*wi*(-A44*cosa*(h7*h7)*s + B22*(h7y*h7y) + 2*B26*h7y*(-h7*s*sina + h7x) + B66*(-h7*s*sina + h7x)**2)
            k0[pos+1535] += detJ*wi*(-A44*cosa*h7*h8*s + h8y*(B22*h7y + B26*(-h7*s*sina + h7x)) + (B26*h7y + B66*(-h7*s*sina + h7x))*(-h8*s*sina + h8x))
            k0[pos+1536] += detJ*wi*(A44*h1y*h7 + A45*h1x*h7 + cosa*h1*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+1537] += detJ*wi*(A44*h2y*h7 + A45*h2x*h7 + cosa*h2*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+1538] += detJ*wi*(A44*h3y*h7 + A45*h3x*h7 + cosa*h3*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+1539] += detJ*wi*(A44*h4y*h7 + A45*h4x*h7 + cosa*h4*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+1540] += detJ*wi*(A44*h5y*h7 + A45*h5x*h7 + cosa*h5*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+1541] += detJ*wi*(A44*h6y*h7 + A45*h6x*h7 + cosa*h6*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+1542] += detJ*h7*wi*(A44*h7y + A45*h7x + cosa*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+1543] += detJ*wi*(A44*h7*h8y + A45*h7*h8x + cosa*h8*s*(B22*h7y - B26*h7*s*sina + B26*h7x))
            k0[pos+1544] += detJ*wi*(A45*h1*h7 + h1*s*sina*(D22*h7y + D26*(-h7*s*sina + h7x)) + h1x*(D12*h7y + D16*(-h7*s*sina + h7x)) + h1y*(D26*h7y + D66*(-h7*s*sina + h7x)))
            k0[pos+1545] += detJ*wi*(A45*h2*h7 + h2*s*sina*(D22*h7y + D26*(-h7*s*sina + h7x)) + h2x*(D12*h7y + D16*(-h7*s*sina + h7x)) + h2y*(D26*h7y + D66*(-h7*s*sina + h7x)))
            k0[pos+1546] += detJ*wi*(A45*h3*h7 + h3*s*sina*(D22*h7y + D26*(-h7*s*sina + h7x)) + h3x*(D12*h7y + D16*(-h7*s*sina + h7x)) + h3y*(D26*h7y + D66*(-h7*s*sina + h7x)))
            k0[pos+1547] += detJ*wi*(A45*h4*h7 + h4*s*sina*(D22*h7y + D26*(-h7*s*sina + h7x)) + h4x*(D12*h7y + D16*(-h7*s*sina + h7x)) + h4y*(D26*h7y + D66*(-h7*s*sina + h7x)))
            k0[pos+1548] += detJ*wi*(A45*h5*h7 + h5*s*sina*(D22*h7y + D26*(-h7*s*sina + h7x)) + h5x*(D12*h7y + D16*(-h7*s*sina + h7x)) + h5y*(D26*h7y + D66*(-h7*s*sina + h7x)))
            k0[pos+1549] += detJ*wi*(A45*h6*h7 + h6*s*sina*(D22*h7y + D26*(-h7*s*sina + h7x)) + h6x*(D12*h7y + D16*(-h7*s*sina + h7x)) + h6y*(D26*h7y + D66*(-h7*s*sina + h7x)))
            k0[pos+1550] += detJ*wi*(A45*(h7*h7) + h7*s*sina*(D22*h7y + D26*(-h7*s*sina + h7x)) + h7x*(D12*h7y + D16*(-h7*s*sina + h7x)) + h7y*(D26*h7y + D66*(-h7*s*sina + h7x)))
            k0[pos+1551] += detJ*wi*(A45*h7*h8 + h8*s*sina*(D22*h7y + D26*(-h7*s*sina + h7x)) + h8x*(D12*h7y + D16*(-h7*s*sina + h7x)) + h8y*(D26*h7y + D66*(-h7*s*sina + h7x)))
            k0[pos+1552] += detJ*wi*(A44*h1*h7 + h1y*(D22*h7y + D26*(-h7*s*sina + h7x)) + (D26*h7y + D66*(-h7*s*sina + h7x))*(-h1*s*sina + h1x))
            k0[pos+1553] += detJ*wi*(A44*h2*h7 + h2y*(D22*h7y + D26*(-h7*s*sina + h7x)) + (D26*h7y + D66*(-h7*s*sina + h7x))*(-h2*s*sina + h2x))
            k0[pos+1554] += detJ*wi*(A44*h3*h7 + h3y*(D22*h7y + D26*(-h7*s*sina + h7x)) + (D26*h7y + D66*(-h7*s*sina + h7x))*(-h3*s*sina + h3x))
            k0[pos+1555] += detJ*wi*(A44*h4*h7 + h4y*(D22*h7y + D26*(-h7*s*sina + h7x)) + (D26*h7y + D66*(-h7*s*sina + h7x))*(-h4*s*sina + h4x))
            k0[pos+1556] += detJ*wi*(A44*h5*h7 + h5y*(D22*h7y + D26*(-h7*s*sina + h7x)) + (D26*h7y + D66*(-h7*s*sina + h7x))*(-h5*s*sina + h5x))
            k0[pos+1557] += detJ*wi*(A44*h6*h7 + h6y*(D22*h7y + D26*(-h7*s*sina + h7x)) + (D26*h7y + D66*(-h7*s*sina + h7x))*(-h6*s*sina + h6x))
            k0[pos+1558] += detJ*wi*(A44*(h7*h7) + D66*(-h7*s*sina + h7x)**2 + h7y*(D22*h7y - 2*D26*h7*s*sina + 2*D26*h7x))
            k0[pos+1559] += detJ*wi*(A44*h7*h8 + h8y*(D22*h7y + D26*(-h7*s*sina + h7x)) + (D26*h7y + D66*(-h7*s*sina + h7x))*(-h8*s*sina + h8x))
            k0[pos+1560] += detJ*wi*(h1*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h1x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h1y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+1561] += detJ*wi*(h2*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h2x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h2y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+1562] += detJ*wi*(h3*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h3x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h3y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+1563] += detJ*wi*(h4*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h4x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h4y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+1564] += detJ*wi*(h5*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h5x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h5y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+1565] += detJ*wi*(h6*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h6x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h6y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+1566] += detJ*wi*(h7*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h7x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h7y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+1567] += detJ*wi*(h8*s*sina*(B22*h8y + B26*(-h8*s*sina + h8x)) + h8x*(B12*h8y + B16*(-h8*s*sina + h8x)) + h8y*(B26*h8y + B66*(-h8*s*sina + h8x)))
            k0[pos+1568] += detJ*wi*(-A44*cosa*h1*h8*s + h1y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h1*s*sina + h1x))
            k0[pos+1569] += detJ*wi*(-A44*cosa*h2*h8*s + h2y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h2*s*sina + h2x))
            k0[pos+1570] += detJ*wi*(-A44*cosa*h3*h8*s + h3y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h3*s*sina + h3x))
            k0[pos+1571] += detJ*wi*(-A44*cosa*h4*h8*s + h4y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h4*s*sina + h4x))
            k0[pos+1572] += detJ*wi*(-A44*cosa*h5*h8*s + h5y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h5*s*sina + h5x))
            k0[pos+1573] += detJ*wi*(-A44*cosa*h6*h8*s + h6y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h6*s*sina + h6x))
            k0[pos+1574] += detJ*wi*(-A44*cosa*h7*h8*s + h7y*(B22*h8y + B26*(-h8*s*sina + h8x)) + (B26*h8y + B66*(-h8*s*sina + h8x))*(-h7*s*sina + h7x))
            k0[pos+1575] += detJ*wi*(-A44*cosa*(h8*h8)*s + B22*(h8y*h8y) + 2*B26*h8y*(-h8*s*sina + h8x) + B66*(-h8*s*sina + h8x)**2)
            k0[pos+1576] += detJ*wi*(A44*h1y*h8 + A45*h1x*h8 + cosa*h1*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+1577] += detJ*wi*(A44*h2y*h8 + A45*h2x*h8 + cosa*h2*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+1578] += detJ*wi*(A44*h3y*h8 + A45*h3x*h8 + cosa*h3*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+1579] += detJ*wi*(A44*h4y*h8 + A45*h4x*h8 + cosa*h4*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+1580] += detJ*wi*(A44*h5y*h8 + A45*h5x*h8 + cosa*h5*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+1581] += detJ*wi*(A44*h6y*h8 + A45*h6x*h8 + cosa*h6*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+1582] += detJ*wi*(A44*h7y*h8 + A45*h7x*h8 + cosa*h7*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+1583] += detJ*h8*wi*(A44*h8y + A45*h8x + cosa*s*(B22*h8y - B26*h8*s*sina + B26*h8x))
            k0[pos+1584] += detJ*wi*(A45*h1*h8 + h1*s*sina*(D22*h8y + D26*(-h8*s*sina + h8x)) + h1x*(D12*h8y + D16*(-h8*s*sina + h8x)) + h1y*(D26*h8y + D66*(-h8*s*sina + h8x)))
            k0[pos+1585] += detJ*wi*(A45*h2*h8 + h2*s*sina*(D22*h8y + D26*(-h8*s*sina + h8x)) + h2x*(D12*h8y + D16*(-h8*s*sina + h8x)) + h2y*(D26*h8y + D66*(-h8*s*sina + h8x)))
            k0[pos+1586] += detJ*wi*(A45*h3*h8 + h3*s*sina*(D22*h8y + D26*(-h8*s*sina + h8x)) + h3x*(D12*h8y + D16*(-h8*s*sina + h8x)) + h3y*(D26*h8y + D66*(-h8*s*sina + h8x)))
            k0[pos+1587] += detJ*wi*(A45*h4*h8 + h4*s*sina*(D22*h8y + D26*(-h8*s*sina + h8x)) + h4x*(D12*h8y + D16*(-h8*s*sina + h8x)) + h4y*(D26*h8y + D66*(-h8*s*sina + h8x)))
            k0[pos+1588] += detJ*wi*(A45*h5*h8 + h5*s*sina*(D22*h8y + D26*(-h8*s*sina + h8x)) + h5x*(D12*h8y + D16*(-h8*s*sina + h8x)) + h5y*(D26*h8y + D66*(-h8*s*sina + h8x)))
            k0[pos+1589] += detJ*wi*(A45*h6*h8 + h6*s*sina*(D22*h8y + D26*(-h8*s*sina + h8x)) + h6x*(D12*h8y + D16*(-h8*s*sina + h8x)) + h6y*(D26*h8y + D66*(-h8*s*sina + h8x)))
            k0[pos+1590] += detJ*wi*(A45*h7*h8 + h7*s*sina*(D22*h8y + D26*(-h8*s*sina + h8x)) + h7x*(D12*h8y + D16*(-h8*s*sina + h8x)) + h7y*(D26*h8y + D66*(-h8*s*sina + h8x)))
            k0[pos+1591] += detJ*wi*(A45*(h8*h8) + h8*s*sina*(D22*h8y + D26*(-h8*s*sina + h8x)) + h8x*(D12*h8y + D16*(-h8*s*sina + h8x)) + h8y*(D26*h8y + D66*(-h8*s*sina + h8x)))
            k0[pos+1592] += detJ*wi*(A44*h1*h8 + h1y*(D22*h8y + D26*(-h8*s*sina + h8x)) + (D26*h8y + D66*(-h8*s*sina + h8x))*(-h1*s*sina + h1x))
            k0[pos+1593] += detJ*wi*(A44*h2*h8 + h2y*(D22*h8y + D26*(-h8*s*sina + h8x)) + (D26*h8y + D66*(-h8*s*sina + h8x))*(-h2*s*sina + h2x))
            k0[pos+1594] += detJ*wi*(A44*h3*h8 + h3y*(D22*h8y + D26*(-h8*s*sina + h8x)) + (D26*h8y + D66*(-h8*s*sina + h8x))*(-h3*s*sina + h3x))
            k0[pos+1595] += detJ*wi*(A44*h4*h8 + h4y*(D22*h8y + D26*(-h8*s*sina + h8x)) + (D26*h8y + D66*(-h8*s*sina + h8x))*(-h4*s*sina + h4x))
            k0[pos+1596] += detJ*wi*(A44*h5*h8 + h5y*(D22*h8y + D26*(-h8*s*sina + h8x)) + (D26*h8y + D66*(-h8*s*sina + h8x))*(-h5*s*sina + h5x))
            k0[pos+1597] += detJ*wi*(A44*h6*h8 + h6y*(D22*h8y + D26*(-h8*s*sina + h8x)) + (D26*h8y + D66*(-h8*s*sina + h8x))*(-h6*s*sina + h6x))
            k0[pos+1598] += detJ*wi*(A44*h7*h8 + h7y*(D22*h8y + D26*(-h8*s*sina + h8x)) + (D26*h8y + D66*(-h8*s*sina + h8x))*(-h7*s*sina + h7x))
            k0[pos+1599] += detJ*wi*(A44*(h8*h8) + D66*(-h8*s*sina + h8x)**2 + h8y*(D22*h8y - 2*D26*h8*s*sina + 2*D26*h8x))

    free(invJ)
    free(detJs)
